#ifndef MPINT_H
#define MPINT_H

#include "mp.h"

class mp_real;
#include "mpreal.h"
#include "mpcomplex.h"

/**  
 * This class represents multi-precision integer. 
 *
 * @author Herman Harjono
 * @version Sep 21, 1998.
 */ 
class mp_int:public mp
{

#include "mpint_friends.h"

 public:
  
  /**
   *  Creates a mp_int (initialized to zero). 
   *  @param precision The precision level of the object, in digits. 
   *  @see mp#mpipl
   */
  mp_int(const mpprecision& precision=mpipl):mp(precision){}
  
  /**
   *  Creates a mp_int from int. 
   *  @param precision The precision level of the object, in digits. 
   *  @see mp#mpipl
   */
  mp_int(int ia, const mpprecision& precision=mpipl):mp(ia, precision){}
  
  /**
   *  Creates a mp_int from float / double. 
   *  @param precision The precision level of the object, in digits. 
   *  The default is assigned the value of mpipl.
   *  @see mp#mpipl
   */
  mp_int(double, const mpprecision& precision=mpipl);
  
  /**
   *  Creates a mp_int from char*.
   *  @param precision The precision level of the object, in digits. 
   *  The default is assigned the value of mpipl.
   *  @see mp#mpipl
   */
  mp_int(char *, const mpprecision& precision=mpipl);
  
  /**
   *  Copy constructor. Calls mp's copy constructor.
   */
  mp_int(const mp_int& in) :mp(in){};
  
  /**
   *  Assignment operator.
   */
  mp_int& operator=(const int& ja);
  
  /**
   *  Assignment operator.
   */
  mp_int& operator=(const double&);
  
  /**
   *  Assignment operator.
   */
  mp_int& operator=(const char *);
  
  /**
   *  Assignment operator.
   */
  mp_int& operator=(const mp_int& ja);
  
  /**
   *  Assignment operator. Truncates the mp_real number.
   */
  mp_int& operator=(const mp_real&);
  
  /**
   *  Assignment operator. Gets the real part of the mp_complex number.
   */
  mp_int& operator=(const mp_complex&);
  
  /**
   *  Preincrement operator.
   */
  mp_int operator++();
  
  /**
   *  Postincrement operator.
   */
  mp_int operator++(int);
  
  /**
   *  Predecrement operator.
   */
  mp_int operator--();
  
  /**
   *  Postdecrement operator.
   */
  mp_int operator--(int);
  
  /**
   *  Calls operator+ and operator=
   */
  inline mp_int& operator+=(const mp_int& ja)
    {*this = (*this)+ja; return *this;}
  
  /**
   *  Calls operator- and operator=
   */
  inline mp_int& operator-=(const mp_int& ja)
    {*this = (*this)-ja; return *this;}
  
  /**
   *  Calls operator* and operator=
   */
  inline mp_int& operator*=(const mp_int& ja)
    {*this = (*this)*ja; return *this;}
  
  /**
   *  Calls operator/ and operator=
   */
  inline mp_int& operator/=(const mp_int& ja)
    {*this = (*this)/ja; return *this;}
  
  /**
   *  Calls operator% and operator=
   */
  inline mp_int& operator%=(const mp_int& ja)
    {*this = (*this)%ja; return *this;}
  
   
};

#endif
