#ifndef MP_REAL_H
#define MP_REAL_H

#include "mp.h"
class mp_int;
#include "mpint.h"

/**  
 * This class represents multi-precision real.
 * <I>
 * In order to use the library, this is the only header file
 * that needs to be included.
 * </I>
 *
 * @author Herman Harjono
 * @version Sep 21, 1998.
 */ 
class mp_real:public mp
{
 public:
  
  /**
   *  Creates a mp_real (initialized to zero). 
   *  @param size The precision level of the object, in words. 
   */
  mp_real(const mpsize& size):mp(size){}
  
  /**
   *  Creates a mp_real (initialized to zero). 
   *  @param precision The precision level of the object, in digits. 
   *  @see mp#mpipl
   */
  mp_real(const mpprecision& precision=mpipl):mp(precision){}
  
  /**
   *  Creates a mp_real from int. 
   *  @param precision The precision level of the object, in digits.
   *  @see mp#mpipl
   */
  mp_real(int ia, const mpprecision& precision=mpipl):mp(ia, precision){} 
  
  /**
   *  Creates a mp_real from float / double. 
   *  @param precision The precision level of the object, in digits. 
   *  @see mp#mpipl
   */
  mp_real(double da, const mpprecision& precision=mpipl)
    :mp(da, precision){}
  
  /**
   *  Creates a mp_real from char*. 
   *  @param precision The precision level of the object, in digits. 
   *  The default is assigned the value of mpipl.
   *  @see mp#mpipl
   */
  mp_real(char *ca, const mpprecision& precision=mpipl)
    :mp(ca, precision){}
  
  /**
   *  Creates a mp_real from mp_int. Inherits the mp_int object's precision.
   *  Calls mp's copy constructor.
   */
  mp_real(const mp_int& in):mp((mp&)in){};
  
  /**
   *  Copy constructor. Inherits the parameter's precision.
   *  Calls mp's copy constructor.
   */
  mp_real(const mp_real& in) :mp((mp&)in){};
  
  /**
   *  Assignment operator.
   */
  mp_real& operator=(const int&);
  
  /**
   *  Assignment operator.
   */
  mp_real& operator=(const double&);
  
  /**
   *  Assignment operator.
   */
  mp_real& operator=(const mp_real&);
  
  /**
   *  Assignment operator.
   */
  mp_real& operator=(const mp_int&);
  
  /**
   *  Assignment operator.
   */
  mp_real& operator=(const mp_complex&);
  
  /**
   *  Assignment operator.
  */
  mp_real& operator=(const char*);
  
   
  /**
   *  Preincrement operator.
   */
  mp_real operator++();
  
  /**
   *  Postincrement operator.
   */
  mp_real operator++(int);
  
  /**
   *  Predecrement operator.
   */
  mp_real operator--();
  
  /**
   *  Postdecrement operator.
   */
  mp_real operator--(int);
	
  /**
   *  Calls operator+ and operator=
   */
  inline mp_real& operator+=(const mp_real& ja)
    {*this = (*this)+ja; return *this;}
  
  /**
   *  Calls operator- and operator=
   */
  inline mp_real& operator-=(const mp_real& ja)
    {*this = (*this)-ja; return *this;}
	
  /**
   *  Calls operator* and operator=
   */
  inline mp_real& operator*=(const mp_real& ja)
    {*this = (*this)*ja; return *this;}
  
  /**
  *  Calls operator/ and operator=
  */
  inline mp_real& operator/=(const mp_real& ja)
    {*this = (*this)/ja; return *this;} 
  
   
};

#endif

