#include "mp.h"

void mp::mpcbrx (const mp& a, mp& b, int mpnw)
{
  int k;
  const double cl2= 1.4426950408889633;
  double t1;
  const int nit = 3;
  mpsize mpnw3(mpnw+3);
  mp f(mpsize(6)), sk0(mpnw3), sk1(mpnw3), sk2(mpnw3);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 6) 
    debug("MPCBRX I ", a);
#endif
  
  
  int na = MIN (a.nw, mpnw);
  int ncr = int(pow(2, MPMCR)); // *cast*
  
  if (na == 0) 
  {
    zero(b);
#if DEBUG == 2
    if (MPIDB >= 6) 
      debug("MPCBRX O ", b);
#endif
    return;
  }
  if (a.sign < 0.0) 
  {
    if (MPKER[14] != 0) 
    {
      cerr << "*** MPCBRX: Argument is negative.\n";
      MPIER = 14;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  //  Check if precision level is too low to justify the advanced routine.
  
  if (mpnw <= ncr) 
  {
    mpcbrt (a, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 6) 
      debug("MPCBRX O ", b);
#endif
    return;
  }
  int nws = mpnw;
  
  //  Determine the least integer MQ such that 2 ^ MQ >= MPNW.
  
  t1 = mpnw;
  int mq = int(cl2 * log (t1) + 1.0 - mprxx); // *Cast*
  
  //  Compute A^2 outside of the iteration loop.
  
  mpsqx (a, sk0, mpnw);
  
  //  Compute the initial approximation of A ^ (-2/3).
  
  mpnw = ncr + 1;
  mpcbrt (a, sk1, mpnw);
  mpdiv (sk1, a, b, mpnw);
  f.nw=1; f.sign=1; f.exponent=0; f[0]=1; f[1]=0;
  int iq = 0, nw1, nw2;
  
  //  Perform the Newton-Raphson iteration described above with a dynamically
  //  changing precision level MPNW (powers of two).
  
  for(k = MPMCR + 1; k<=mq - 1; k++)
  {
    nw1 = mpnw; 
    mpnw=MIN (2 * mpnw - 2, nws) + 1;
    nw2 = mpnw; 
    
    bool cont=true;
    while (cont)
    {
      mpsqx (b, sk1, mpnw);
      mpmulx (b, sk1, sk2, mpnw);
      mpmulx (sk0, sk2, sk1, mpnw);
      mpsub (f, sk1, sk2, mpnw);
      
      mpnw = nw1;
      
      mpmulx (b, sk2, sk1, mpnw);
      mpdivd (sk1, mp_dpe(3.0), sk2, mpnw);
      
      mpnw = nw2;
      
      mpadd (b, sk2, sk1, mpnw);
      mpeq (sk1, b, mpnw);
      
      if (k == mq - nit && iq == 0) 
        iq = 1;
      
      else
        cont=false;
    }
  }
  
  //  Perform last iteration using Karp's trick.
  
  mpmulx(a, b, sk0, mpnw);
  nw1 = mpnw; 
  mpnw = MIN (2 * mpnw - 2, nws) + 1;
  
  nw2 = mpnw;
  
  mpsqx (sk0, sk1, mpnw);
  mpmulx (sk0, sk1, sk2, mpnw);
  mpsub (a, sk2, sk1, mpnw);
  
  mpnw = nw1;
  
  mpmulx (sk1, b, sk2, mpnw);
  mpdivd (sk2, mp_dpe(3.0), sk1, mpnw);
  
  mpnw = nw2;
  
  mpadd (sk0, sk1, sk2, mpnw);
  mpeq (sk2, b, mpnw);  
  
  mproun (b, nws);
  
#if DEBUG == 2
  if (MPIDB >= 6) 
    debug("MPCBRX O ", b);
#endif
  
}

void mp::mpdivx (const mp& a, const mp& b, mp& c, int mpnw)
{
  int k;
  const double cl2 = 1.4426950408889633;
  double t1;
  int nit = 3;
  
  mpsize mpnw3(mpnw+3);
  mp f(mpsize(6)), sk0(mpnw3), sk1(mpnw3), sk2(mpnw3);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(c);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 7) 
  {
    debug("MPDIVX I ", a); 
    debug("MPDIVX I ", b);
  }
#endif
  
  int na = MIN (a.nw, mpnw);
  int nb = MIN (b.nw, mpnw);
  int ncr = int(pow(2, MPMCR)); // *cast*
  
  //  Check if dividend is zero.
  
  if (na == 0) 
  {
    zero(c);
#if DEBUG == 2
    if (MPIDB >= 7) 
      debug("MPDIVX O ", c);
#endif
    return;
  }
  
  //  Check if divisor is zero. 
  if (nb == 0)  
  {
    if (MPKER[33] != 0) 
    {
      cerr << "*** MPDIVX: Divisor is zero.\n";
      MPIER = 33;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  //  Check if precision level of divisor is too low to justify the advanced
  //  routine.
  if (nb <= ncr) 
  {
    mpdiv (a, b, c, mpnw);
#if DEBUG == 2
    if (MPIDB >= 7) 
      debug("MPDIVX O ", c);
#endif
    return;
  }
  int nws = mpnw;
  
  //  Determine the least integer MQ such that 2 ^ MQ >= MPNW.
  
  t1 = mpnw;
  int mq = int(cl2 * log (t1) + 1.0 - mprxx); // *cast*
  
  //  Compute the initial approximation of 1 / B to a precision of NCR words.
  mpnw = ncr + 1;
  
  f.nw=1; f.sign=1; f.exponent=0; f[0]=1; f[1]=0;
  mpdiv(f,b,c, mpnw);
  int iq = 0;
  int nw1, nw2;
  
  //  Perform the Newton-Raphson iterations described above.
  for(k = MPMCR + 1; k<= mq - 1; k++)
  {
    nw1 = mpnw;
    mpnw = MIN (2 * mpnw - 2, nws) + 1;
    nw2 = mpnw;
    bool cont = true;
    while (cont)
    {
      mpmulx (b, c, sk0, mpnw);
      mpsub (f, sk0, sk1, mpnw);
      mpnw = nw1;
      mpmulx (c, sk1, sk0, mpnw);
      mpnw = nw2;
      mpadd (c, sk0, sk1, mpnw);
      mpeq (sk1, c, mpnw);
      if (k == mq - nit && iq == 0) 
        iq = 1;
      else
        cont=false;
    }
  }
  
  //  Perform last iteration using Karp's trick.
  mpmulx (a, c, sk0, mpnw);
  nw1 = mpnw;
  mpnw = MIN (2 * mpnw - 2, nws) + 1;
  nw2 = mpnw;
  mpmulx (sk0, b, sk1, mpnw);
  mpsub (a, sk1, sk2, mpnw);
  mpnw = nw1;
  mpmulx (sk2, c, sk1, mpnw);
  mpnw = nw2;
  mpadd (sk0, sk1, sk2, mpnw);
  mpeq (sk2, c, mpnw);
  
  //  Restore original precision level.
  
  mproun (c, nws);
  
#if DEBUG == 2
  if (MPIDB >= 7) 
    debug("MPDIVX O ", c);
#endif
  
}

void mp::mpmulx(const mp& a, const mp&b, mp& c, int mpnw)
{
  double t1, t2, t3, t4;
  int i;
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(c);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 8)  
  {
    debug("MPMULX I ", a);
    debug("MPMULX I ", b);
  }
#endif
  
  int na = MIN (a.nw, mpnw);
  int nb = MIN (b.nw, mpnw);
  int ncr =int( pow(2, MPMCR)); // *cast*
  
  if (na == 0 || nb == 0) 
  {
    //  One of the inputs is zero -- result is zero.
    zero(c);
#if DEBUG == 2
    if (MPIDB >= 8) 
      debug("MPMULX O ", c);
#endif
    return;
  }
  
  //  Check if precision level of one of the arguments is too low to justify the
  //  advanced routine.
  
  if (na <= ncr || nb <= ncr) 
  {
    mpmul (a, b, c, mpnw);
#if DEBUG == 2
    if (MPIDB >= 8) 
      debug("MPMULX O ", c);
#endif
    return;
  }
  ;
  double *d1 = new double[2*mpnw+4], *d2 = new double[2*mpnw+4], 
    *d3 = new double[4*mpnw+8];
  
  //  Place the input data in A and B into the scratch arrays DD1 and DD2.
  //  This code also splits the input data into half-sized words.
  
  //!dir$ ivdep
  int i2;
  for (i = 0; i< na; i++)
  {
    i2 = 2 * i;
    t1 = a[i];
    t2 = int (mprbx * t1);
    d1[i2] = t2;
    d1[i2+1] = t1 - mpbbx * t2;
  }
  
  for(i = 2 * na; i<2 * nb; i++)
    d1[i] = 0.0;
  
  // !dir$ ivdep
  for(i = 0; i< nb; i++)
  {
    i2 = 2 * i;
    t1 = b[i];
    t2 = int (mprbx * t1);
    d2[i2] = t2;
    d2[i2+1] = t1 - mpbbx * t2;
  }
  
  for (i = 2 * nb; i< 2 * na; i++)
    d2[i] = 0.0;
  
  int nn = 2 * MAX (na, nb);
  int nx = int(sqrt (3.0 * nn) + mprxx); // *cast*
  
  mplconv (2, nn, nx, d1, d2, d3);
  
  //  Recombine words and release carries.
  
  int nc = MIN (na + nb, mpnw);
  int nc1 = MIN (mpnw + 1, na + nb - 1);
  d1[0] = SIGN (nc, a.sign * b.sign);
  d1[1] = a.exponent + b.exponent + 1;
  d1[2] = d3[0];
  d1[nc+2] = 0.0;
  d1[nc+3] = 0.0;
  
  //!dir$ ivdep
  for(i = 1; i<= nc1; i++)
  {
    i2 = 2 * i;
    t3 = d3[i2-1];
    t4 = d3[i2];
    t1 = int (mprdx * t3);
    t2 = t3 - mpbdx * t1;
    t3 = int (mprdx * t4);
    t4 -= mpbdx * t3;
    d1[i+2] = mpbbx * t2 + t4;
    d1[i+1] += mpbbx * t1 + t3;
  }
  //  Fix up the result.
  
  mpnorm (d1, c, mpnw);
  
#if DEBUG == 2
  if (MPIDB >= 8) 
    debug("MPMULX O ", c);
#endif
  delete d1; delete d2; delete d3;
}

void mp::mpnpwx (const mp& a, int n, mp& b, int mpnw)
{
  int j;
  const double cl2 = 1.4426950408889633; 
  double t1;
  mpsize mpnw2(mpnw+2);
  mp f1(mpsize(6)), sk0(mpnw2), sk1(mpnw2);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 6) 
    debug("MPNPWX I ", a);
#endif
  
  int ncr = int(pow(2, MPMCR));
  int na = MIN (a.nw, mpnw);
  
  //  Check if precision level of A is too low to justify the advanced routine.
  if (na <= ncr && n >= 0 && n <= 4) 
  {
    mpnpwr (a, n, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 6) 
      debug("MPNPWX O ", b);
#endif
    return;
  }
  
  if (na == 0) 
  {
    if (n >= 0) 
    {
      zero(b);
#if DEBUG == 2
      if (MPIDB >= 6) 
        debug("MPNPWX O ", b);
#endif
      return;
    } 
    else 
    {
      if (MPKER[58] != 0) 
      {
        cerr << "*** MPNPWX: Argument is zero and N is negative or zero.\n";
        MPIER = 58;
        if (MPKER[MPIER] == 2)  mpabrt();
      }
      return;
    }
  }
  
  int nn = abs (n);
  f1.sign=1;
  f1.nw=1;
  f1.exponent=0;
  f1[0]=1.;
  f1[1]=0.;
  
  bool skip=false;
  switch(nn)
  {
  case 0:
    mpeq (f1, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 6) 
      debug("MPNPWX O ", b);
#endif
    return;
    
  case 1:
    mpeq (a, b, mpnw);
    skip=true;
    break;
    
  case 2:
    mpsqx (a, b, mpnw);
    skip=true;
    break;
  }
  
  if(!skip)
  {
    //  Determine the least integer MN such that 2 ^ MN > NN.
    t1 = nn;
    int mn = int(cl2 * log (t1) + 1.0 + mprxx); // cast
    mpeq (f1, b, mpnw);
    mpeq (a, sk0, mpnw);
    int kn = nn;    
    
    //  Compute B ^ N using the binary rule for exponentiation.
    for (j = 1; j<= mn; j++)
    {
      int kk = kn / 2;
      if (kn != 2 * kk) 
      {
        mpmulx (b, sk0, sk1, mpnw);
        mpeq (sk1, b, mpnw);
      }
      
      kn = kk;
      if (j < mn) 
      {
        mpsqx (sk0, sk1, mpnw);
        mpeq (sk1, sk0, mpnw);
      }
    }
  }
  
  //  Compute reciprocal if N is negative.
  
  if (n < 0) 
  {
    mpdivx (f1, b, sk0, mpnw);
    mpeq (sk0, b, mpnw);
  }
  
#if DEBUG == 2
  if (MPIDB >= 6) 
    debug("MPNPWX O ", b);
#endif
}

void mp::mpnrtx (const mp& a, int n, mp& b, int mpnw)
{
  int k;
  double t2, tn;
  const double cl2 = 1.4426950408889633;
  const int nit = 3, n30 = int(pow(2,30)); // *cast*
  
  mpsize mpnw3(mpnw+3);
  mp f1(mpsize(6)), f2(mpsize(6)), 
    sk0(mpnw3), sk1(mpnw3), sk2(mpnw3), sk3(mpnw3);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 6) 
    debug("MPNRTX I ", a);
#endif  
  
  int ncr = int(pow(2, MPMCR)); // *cast*
  
  int na = MIN (a.nw, mpnw);
  
  if (na == 0) 
  {
    zero(b);
#if DEBUG == 2
    if (MPIDB >= 6) 
      debug("MPNRTX O ", b);
#endif
    return;
  }
  if (a.sign < 0) 
  {
    if (MPKER[61] != 0) 
    {
      cerr << "*** MPNRTX: Argument is negative.\n";
      MPIER = 61;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  if (n <= 0 || n > n30) 
  {
    if (MPKER[62] != 0) 
    {
      cerr << "*** MPNRT: Improper value of N: " << n << endl;
      MPIER = 62;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  //  Check if precision level is too low to justify the advanced routine.
  if (mpnw <= ncr) 
  {
    mpnrt (a, n, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 6) 
      debug("MPNRTX O ", b);
#endif
    return; 
  }
  
  //  If N = 1, 2 or 3,  MPEQ, MPSQRX or MPCBRX.  These are faster.
  switch (n)
  {
  case 1:
    mpeq (a, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 6) 
      debug("MPNRT O ", b);
#endif
    return;
  case 2:
    mpsqrx (a, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 6) 
      debug("MPNRT O ", b);
#endif
    return;
  case 3:
    mpcbrx (a, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 6) 
      debug("MPNRT O ", b);
#endif
    return;
  }
  
  int nws = mpnw;
  f1.nw=1;
  f1.sign=1;
  f1.exponent=0;
  f1[0]=1;
  f1[1]=0;
  
  //  Determine the least integer MQ such that 2 ^ MQ >= MPNW.
  mp_dpe t1;
  t1.a = mpnw;
  int mq = int(cl2 * log (t1.a) + 1.0 - mprxx); // *cast*
  
  //  Check how close A is to 1.
  mpsub (a, f1, sk0, mpnw);
  if (sk0.nw == 0) 
  {
    mpeq (f1, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 7) 
      debug("MPNRT O ", b);
#endif
    return;
  }
  
  mpmdc (sk0, t1);
  int n2 = int(cl2 * log (fabs (t1.a))); // *cast*
  t1.a *=  pow(0.5, n2);
  t1.n += n2;
  if (t1.n <= -30) 
  {
    t2 = n;
    n2 = int(cl2 * log (t2) + 1.0 + mprxx); // *cast*
    int n3 = - mpnbt * mpnw / t1.n;
    if (n3 < 1.25 * n2) 
    {
      //  A is so close to 1 that it is cheaper to use the binomial series.
      mpdivd (sk0, mp_dpe(t2), sk1, mpnw);
      mpadd (f1, sk1, sk2, mpnw);
      k = 0;
      do 
      {
        k++;
        t1.a = 1 - k * n;
        t2 = (k + 1) * n;
        mpmuld (sk1, mp_dpe(t1.a), sk3, mpnw);
        mpdivd (sk3, mp_dpe(t2), sk1, mpnw);
        mpmulx (sk0, sk1, sk3, mpnw);
        mpeq (sk3, sk1, mpnw);
        mpadd (sk1, sk2, sk3, mpnw);
        mpeq (sk3, sk2, mpnw);
      }
      while (sk1.nw != 0 && sk1.exponent >= - mpnw);        
      
      mpeq (sk2, b, mpnw);
      mproun (b, nws);
      
#if DEBUG == 2
      if (MPIDB >= 7) 
        debug("MPNRT O ", b);
#endif
      return;
    }
  }
  
  //  Compute the initial approximation of A ^ (-1/N).
  mpnw = ncr + 1;
  mpnrt (a, n, sk0, mpnw);
  mpdiv (f1, sk0, b, mpnw);
  
  tn = n;
  mp_dpe dpn(n, 0);
  mpdmc (dpn, f2);
  int iq = 0;
  
  //  Perform the Newton-Raphson iteration described above with a dynamically
  //  changing precision level MPNW (powers of two).  
  int nw1, nw2;
  for (k = MPMCR+1; k<= mq;k++)
  {
    nw1 = mpnw;
    
    mpnw = MIN (2 * mpnw - 2, nws) + 1;
    nw2 = mpnw;
    bool loop=true;
    while (loop)
    {
      
      mpnpwx (b, n, sk0, mpnw);
      mpmulx (a, sk0, sk1, mpnw);
      mpsub (f1, sk1, sk0, mpnw);
      mpnw = nw1;
      mpmulx (b, sk0, sk1, mpnw);
      mpdivd (sk1, mp_dpe(tn), sk0, mpnw);
      mpnw = nw2;
      mpadd (b, sk0, sk1, mpnw);
      mpeq (sk1, b, mpnw);
      if (k == mq - nit && iq == 0) 
        iq = 1;
      else loop=false;
    }
  }
  
  //  Take the reciprocal to give final result.
  mpdivx (f1, b, sk0, mpnw);
  mpeq (sk0, b, mpnw);
  
  mproun (b, nws);
  
#if DEBUG == 2
  if (MPIDB >= 6) 
    debug("MPNRTX O ", b);
#endif
}

void mp::mpsqrx (const mp& a, mp& b, int mpnw)
{
  int k;
  const double cl2 = 1.4426950408889633;
  const int nit = 3;
  mpsize mpnw3(mpnw+3);
  mp f(mpsize(6)), sk0(mpnw3), sk1(mpnw3), sk2(mpnw3);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2 
  if (MPIDB >= 6) 
    debug("MPSQRX I ", a);
#endif
  
  int na = MIN (a.nw, mpnw);
  int ncr = int(pow(2, MPMCR)); // *cast*
  
  if (na == 0) 
  {
    zero(b);
#if DEBUG == 2
    if (MPIDB >= 6) 
      debug("MPSQRX O ", b);
#endif
    return;
  }
  
  if (a.sign < 0.0) 
  {
    if (MPKER[71] != 0) 
    {
      cerr << "*** MPSQRX: Argument is negative.\n";
      MPIER = 71;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  //  Check if precision level is too low to justify the advanced routine.  
  if (mpnw <= ncr) 
  {
    mpsqrt (a, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 6) 
      debug("MPSQRX O ", b);
#endif
    return;
  }
  int nws = mpnw;
  
  //  Determine the least integer MQ such that 2 ^ MQ >= MPNW.
  double t1=mpnw;
  int mq = int(cl2 * log (t1) + 1.0 - mprxx); // *cast* 
  
  //  Compute the initial approximation of 1 / Sqrt(A).
  
  mpnw = ncr + 1;
  mpsqrt (a, sk0, mpnw);
  mpdiv (sk0, a, b, mpnw);
  
  f.nw=1;
  f.sign=1;
  f.exponent=0;
  f[0]=1.;
  f[1]=0.;
  int iq = 0;
  
  //  Perform the Newton-Raphson iteration described above with a dynamically
  //  changing precision level MPNW (one greater than powers of two).
  
  int nw1, nw2;
  for(k = MPMCR + 1; k<= mq - 1; k++)
  {
    nw1 = mpnw;
    mpnw = MIN (2 * mpnw - 2, nws) + 1;
    nw2 = mpnw;
    bool stop=false;
    while(!stop)
    {   
      mpsqx (b, sk0, mpnw);
      mpmulx (a, sk0, sk1, mpnw);
      mpsub (f, sk1, sk0, mpnw);
      mpnw = nw1;
      mpmulx (b, sk0, sk1, mpnw);
      mpmuld (sk1, mp_dpe(0.50), sk0, mpnw);
      mpnw = nw2;
      mpadd (b, sk0, sk1, mpnw);
      mpeq (sk1, b, mpnw);
      if (k == mq - nit && iq == 0) 
        iq = 1;
      else
        stop=true;
    }
  }
  //  Perform last iteration using Karp's trick.
  
  mpmulx (a, b, sk0, mpnw);
  nw1 = mpnw;
  mpnw = MIN (2 * mpnw - 2, nws) + 1;
  
  nw2 = mpnw;
  
  mpsqx (sk0, sk1, mpnw);
  mpsub (a, sk1, sk2, mpnw);
  mpnw = nw1;
  mpmulx (sk2, b, sk1, mpnw);
  mpmuld (sk1, mp_dpe(0.50), sk2, mpnw);
  mpnw = nw2;
  mpadd (sk0, sk2, sk1, mpnw);
  mpeq (sk1, b, mpnw);
  
  mproun (b, nws);
  
#if DEBUG == 2
  if (MPIDB >= 6) 
    debug("MPSQRX O ", b);
#endif
}

void mp::mpsqx(const mp& a, mp& b, int mpnw)
{
  double t1, t2, t3, t4;
  int i;
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 8) 
    debug("MPSQX I ", a);
#endif
  
  int na = MIN (a.nw, mpnw);
  int ncr = int(pow(2, MPMCR));
  
  if (na == 0) 
  {
    zero(b);
#if DEBUG == 2
    if (MPIDB >= 8) 
      debug("MPSQX O ", b);
#endif
    return;
  }
  
  //  Check if precision level of the argument is too low to justify the
  //  advanced routine.
  
  if (na <= ncr) 
  {
    mpmul (a, a, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 8) 
      debug("MPSQX O ", b);
#endif
    return;
  }
  
  double *d1 = new double[2*mpnw+4], *d2 = new double[4*mpnw+8];
  
  //  Place the input data in A into the scratch array DD1.
  //  This code also splits the input data into half-sized words.
  
  //!dir$ ivdep
  int i2;
  for(i = 0; i<na; i++)
  {
    i2 = 2 * i ;
    t1 = a[i];
    t2 = int (mprbx * t1);
    d1[i2] = t2;
    d1[i2+1] = t1 - mpbbx * t2;
  }
  
  int nn = 2 * na;
  int nx = int(sqrt (3.0 * nn) + mprxx); // *cast*
  mplconv (1, nn, nx, d1, d1, d2);
  
  //  Recombine words and release carries.
  
  int nc = MIN (2 * na, mpnw);
  int nc1 = MIN (mpnw + 1, 2 * na - 1);
  d1[0] = nc;
  d1[1] = 2 * a.exponent + 1;
  d1[2] = d2[0];
  d1[nc+2] = 0.0;
  d1[nc+3] = 0.0;
  
  //!dir$ ivdep
  for(i = 1; i<= nc1; i++)
  {
    i2 = 2 * i;
    t3 = d2[i2-1];
    t4 = d2[i2];
    t1 = int (mprdx * t3);
    t2 = t3 - mpbdx * t1;
    t3 = int (mprdx * t4);
    t4 -= mpbdx * t3;
    d1[i+2] = mpbbx * t2 + t4;
    d1[i+1] += mpbbx * t1 + t3;
  }
  
  //  Fix up the result.
  mpnorm (d1, b, mpnw);
  
#if DEBUG == 2
  if (MPIDB >= 8) 
    debug("MPSQX O ", b);
#endif
  delete [] d1; delete [] d2;
}

