#include "mpcomplex.h"

void mp_complex::mpcadd (const mp_complex& a, const mp_complex& b, 
                         mp_complex& c, int mpnw)
{
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(c);
    return;
  }
#if DEBUG == 2
  if (MPIDB >= 9)
    cerr << "MPCADD\n";
#endif
  
  mp::mpadd (*a.r, *b.r, *c.r, mpnw);
  mp::mpadd (*a.i, *b.i, *c.i, mpnw);
}

void mp_complex::mpcdiv (const mp_complex& a, const mp_complex& b, 
                         mp_complex& c, int mpnw)                        
{
  mpsize mpnw2(mpnw+2);
  mp f(mpsize(6)), sk0(mpnw2), sk1(mpnw2), sk2(mpnw2), sk3(mpnw2), sk4(mpnw2); 
  mp *ar=a.r, *ai=a.i, *br=b.r, *bi=b.i;
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(c);
    return;
  }
#if DEBUG == 2
  if (MPIDB >= 7) 
  {
    debug("MPCDIV I ", a);
    debug("MPCDIV I ", b);
  }
#endif
  
  if (b.r->nw  == 0. && b.i->nw == 0.) 
  {
    if (MPKER[16] != 0) 
    {
      cerr << "*** MPCDIV: Divisor is zero.\n";
      MPIER = 16;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  f.nw=1; f.exponent=0; f.sign=1; f[0]=1; f[1]=0;
  
  mp::mpmul (*ar, *br, sk0, mpnw);
  mp::mpmul (*ai, *bi, sk1, mpnw);
  mp::mpadd (sk0, sk1, sk2, mpnw);
  mp::mpsub (sk0, sk1, sk3, mpnw);
  mp::mpadd (*ar, *ai, sk0, mpnw);
  mp::mpsub (*br, *bi, sk1, mpnw);
  mp::mpmul (sk0, sk1, sk4, mpnw);
  mp::mpsub (sk4, sk3, sk1, mpnw);
  mp::mpmul (*br, *br, sk0, mpnw);
  mp::mpmul (*bi, *bi, sk3, mpnw);
  mp::mpadd (sk0, sk3, sk4, mpnw);
  mp::mpdiv (f, sk4, sk0, mpnw);
  mp::mpmul (sk2, sk0, *c.r, mpnw);
  mp::mpmul (sk1, sk0, *c.i, mpnw);
  
#if DEBUG == 2
  if (MPIDB >= 7) 
    debug("MPCDIV O ", c);
#endif
}

void mp_complex::mpceq (const mp_complex& a, mp_complex& b, int mpnw)
{
  int i;
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
#if DEBUG == 2
  if (MPIDB >= 10) cerr << "MPCEQ\n";
#endif
  
  int n1 = MIN (a.r->nw, mpnw);
  int n2 = MIN (a.i->nw, mpnw);
  b.r->sign = a.r->sign;
  b.r->nw = n1;
  b.r->exponent=a.r->exponent;
  b.i->sign=a.i->sign;
  b.i->nw=n2;
  b.i->exponent=a.i->exponent;
  
  for (i = 0; i<n1; i++)
    (*b.r)[i] = (*a.r)[i];
  for (i=0; i<n2; i++)
    (*b.i)[i] = (*a.i)[i];
}

void mp_complex::mpcmul(const mp_complex& a, const mp_complex& b, 
												mp_complex& c, int mpnw)
{ 
  mpsize mpnw2(mpnw+2);
  mp sk0(mpnw2), sk1(mpnw2), sk2(mpnw2), sk3(mpnw2);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(c);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 7) 
  {
    debug("MPCMUL I ", a);
    debug("MPCMUL I ", b);
  }
#endif
  
  mp::mpmul (*a.r, *b.r, sk0, mpnw);
  mp::mpmul (*a.i, *b.i, sk1, mpnw);
  mp::mpsub (sk0, sk1, *c.r, mpnw);
  mp::mpadd (sk0, sk1, sk2, mpnw);
  mp::mpadd (*a.r, *a.i, sk0, mpnw);
  mp::mpadd (*b.r, *b.i, sk1, mpnw);
  mp::mpmul (sk0, sk1, sk3, mpnw);
  mp::mpsub (sk3, sk2, *c.i, mpnw);
  
#if DEBUG == 2
  if (MPIDB >= 7) 
    debug("MPCMUL O ", c);
#endif
}

void mp_complex::mpcpwr (const mp_complex& a, int n, 
                         mp_complex& b, int mpnw)
{
  int j;
  const double cl2 = 1.4426950408889633;
  double t1;
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
	
#if DEBUG == 2
  if (MPIDB >= 7) 
    debug("MPCPWR I ", a);
#endif
  
  int na1 = MIN (a.r->nw, mpnw);
  int na2 = MIN (a.i->nw, mpnw);
  
  if (na1 == 0 && na2 == 0) 
  {
    if (n >= 0) 
    {
      zero(b);
#if DEBUG == 2
      if (MPIDB >= 7) 
        debug("MPCPWR O ", b);
#endif
      return;
    } 
    else 
    {
      if (MPKER[25] != 0) 
      {
        cerr << "*** MPCPWR: Argument is zero and N is negative or zero.\n";
        MPIER = 25;
        if (MPKER[MPIER] == 2)  mpabrt();
      }
      return;
    }
  }
  
  int nws = mpnw;
  mpnw = mpnw + 1;
  int nn = abs (n);
  
  if (nn == 0) 
  {
    b.r->nw=1; b.r->sign=1; b.r->exponent=0; (*b.r)[0]=1; (*b.r)[1]=0;
    mp::zero(*b.i);
#if DEBUG == 2
    if (MPIDB >= 7) 
      debug("MPCPWR O ", b);
#endif
    return;
  } 
  
  mp_complex *f = createComplex(6), *sk0 = createComplex(mpnw+3), 
    *sk1 = createComplex(mpnw+3), *sk2 = createComplex(mpnw+3);   
  mpceq (a, *sk0, mpnw); 
  f->r->nw=1; f->r->sign=1; f->r->exponent=0; (*f->r)[0]=1; (*f->r)[1]=0;
  bool skip = false;
  
  if (nn == 1) 
  {
    mpceq (*sk0, *sk2, mpnw); 
    skip = true;
  } 
  else if (nn == 2) 
  {
    mpcmul (*sk0, *sk0, *sk2, mpnw);
    skip = true;
  }
  
  //  Determine the least integer MN such that 2 ^ MN > NN.
  if (!skip)
  {
    int mn, kn, kk;
    t1 = nn;
    mn = int(cl2 * log (t1) + 1.0 + mprxx); // *cast*
    mpceq (*f, *sk2, mpnw);
    kn = nn;
    
    //  Compute B ^ N using the binary rule for exponentiation.
    
    for (j = 1; j<=mn; j++)
    {
      kk = kn / 2;
      if (kn != 2 * kk) 
      {
        mpcmul (*sk2, *sk0, *sk1, mpnw);
        mpceq (*sk1, *sk2, mpnw);
      }
      kn = kk;
      if (j < mn) 
      {
        mpcmul (*sk0, *sk0, *sk1, mpnw);
        mpceq (*sk1, *sk0, mpnw);
      }
    }
  }
  //  Compute reciprocal if N is negative.
  
  if (n < 0) 
  {
    mpceq (*f, *sk1, mpnw);
    mpcdiv (*sk1, *sk2, *sk0, mpnw);
    mpceq (*sk0, *sk2, mpnw);
  }
  mpceq (*sk2, b, mpnw);
  
  mpcroun (b, nws);
  
#if DEBUG == 2
  if (MPIDB >= 7) 
    debug("MPCPWR O ", b);
#endif
  delete sk0; delete sk1; delete sk2; delete f; 
}

void mp_complex::mpcsqt (const mp_complex& a, mp_complex& b, int mpnw)
{
  mpsize mpnw2(mpnw+2);
  mp sk0(mpnw2), sk1(mpnw2), sk2(mpnw2); 
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 6) 
    debug("MPCSQT I", a);
#endif
  
  if (a.r->nw == 0. && a.i->nw == 0.) 
  {
    zero(b);
#if DEBUG == 2
    if (MPIDB >= 6) 
      debug("MPCSQT O ", b);
#endif
    return;
  }
  
  mp::mpmul (*a.r, *a.r, sk0, mpnw);
  mp::mpmul (*a.i, *a.i, sk1, mpnw);
  mp::mpadd (sk0, sk1, sk2, mpnw);
  mp::mpsqrt (sk2, sk0, mpnw);
  mp::mpeq (*a.r, sk1, mpnw);
  sk1.sign = 1;
  mp::mpadd (sk0, sk1, sk2, mpnw);
  mp::mpmuld (sk2, mp_dpe(0.50), sk1, mpnw);
  mp::mpsqrt (sk1, sk0, mpnw);
  mp::mpmuld (sk0, mp_dpe(2.0), sk1, mpnw);
  if (a.r->nw >= 0.) 
  {
    mp::mpeq (sk0, *b.r, mpnw);
    mp::mpdiv (*a.i, sk1, *b.i, mpnw);
  } 
  else 
  {
    mp::mpdiv (*a.i, sk1, *b.r, mpnw);
    b.r->sign = 1;
    mp::mpeq (sk0, *b.i, mpnw);
    b.i->nw = a.i->nw;
  }
  
#if DEBUG == 2
  if (MPIDB >= 6) 
    debug("MPCSQT O ", b); 
#endif
}

void mp_complex::mpcsub (const mp_complex& a, const mp_complex& b, 
                         mp_complex& c, int mpnw)
{
  if (MPIER != 0)
  {
    if (MPIER == 99)  mpabrt();
    zero(c);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 9) cerr << "MPCSUB\n";
#endif
  
  mp::mpsub (*a.r, *b.r, *c.r, mpnw);
  mp::mpsub (*a.i, *b.i, *c.i, mpnw);
}


