#include "mpcomplex.h"

void mp_complex::mpcdvx (const mp_complex& a, const mp_complex& b, 
                         mp_complex& c, int mpnw)
{
  
  mpsize mpnw2(mpnw+2);
  mp f(mpsize(6)), sk0(mpnw2), sk1(mpnw2), sk2(mpnw2), sk3(mpnw2), sk4(mpnw2); 
  mp *ar=a.r, *ai=a.i, *br=b.r, *bi=b.i;
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(c);
    return;
  }
  
  
#if DEBUG == 2
  if (MPIDB >= 7) 
  {
    debug("MPCDVX I ", a);
    debug("MPCDVX I ", b);
  }
#endif
  
  
  if (b.r->nw  == 0. && b.i->nw == 0.) 
  {
    if (MPKER[18] != 0) 
    {
      cerr << "*** MPCDVX: Divisor is zero.\n";
      MPIER = 18;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  f.nw=1; f.exponent=0; f.sign=1; f[0]=1; f[1]=0;
  
  mp::mpmulx (*ar, *br, sk0, mpnw);
  mp::mpmulx (*ai, *bi, sk1, mpnw);
  mp::mpadd (sk0, sk1, sk2, mpnw);
  mp::mpsub (sk0, sk1, sk3, mpnw);
  mp::mpadd (*ar, *ai, sk0, mpnw);
  mp::mpsub (*br, *bi, sk1, mpnw);
  mp::mpmulx (sk0, sk1, sk4, mpnw);
  mp::mpsub (sk4, sk3, sk1, mpnw);
  mp::mpsqx (*br,  sk0, mpnw);
  mp::mpsqx (*bi, sk3, mpnw);
  mp::mpadd (sk0, sk3, sk4, mpnw);
  mp::mpdivx (f, sk4, sk0, mpnw);
  mp::mpmul (sk2, sk0, *c.r, mpnw);
  mp::mpmul (sk1, sk0, *c.i, mpnw);
  
#if DEBUG == 2
  if (MPIDB >= 7) 
    debug("MPCDVX O ", c);
#endif
  
}

void mp_complex::mpcmlx (const mp_complex& a, const mp_complex& b, 
                         mp_complex& c, int mpnw)
{
  mpsize mpnw2(mpnw+2);
  mp sk0(mpnw2), sk1(mpnw2), sk2(mpnw2), sk3(mpnw2);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(c);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 7) 
  {
    debug("MPCMLX I ", a);
    debug("MPCMLX I ", b);
  }
#endif
  
  mp::mpmulx (*a.r, *b.r, sk0, mpnw);
  mp::mpmulx (*a.i, *b.i, sk1, mpnw);
  mp::mpsub (sk0, sk1, *c.r, mpnw);
  mp::mpadd (sk0, sk1, sk2, mpnw);
  mp::mpadd (*a.r, *a.i, sk0, mpnw);
  mp::mpadd (*b.r, *b.i, sk1, mpnw);
  mp::mpmulx (sk0, sk1, sk3, mpnw);
  mp::mpsub (sk3, sk2, *c.i, mpnw);
  
#if DEBUG == 2
  if (MPIDB >= 7) 
    debug("MPCMLX O ", c);
#endif
  
}

void mp_complex::mpcpwx (const mp_complex& a, int n, 
                         mp_complex& b, int mpnw)
{
  
  int j;
  const double cl2 = 1.4426950408889633;
  double t1;
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 6) 
    debug("MPCPWX I ", a);
#endif
  
  int na1 = MIN (a.r->nw, mpnw);
  int na2 = MIN (a.i->nw, mpnw);
  int ncr = int(pow(2, MPMCR)); // *cast*
  
  //  Check if precision level of A is too low to justify advanced routine.
  if (na1 <= ncr && na2 <= ncr) 
  {
    mpcpwr (a, n, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 6) 
      debug("MPCPWX O ", b);
#endif
    return;
  }
  
  if (na1 == 0 && na2 == 0) 
  {
    if (n >= 0) 
    {
      zero(b);
#if DEBUG == 2
      if (MPIDB >= 6) 
        debug("MPCPWX O ", b);
#endif
      return;
    } 
    else 
    {
      if (MPKER[26] != 0) 
      {
        cerr << "*** MPCPWX: Argument is zero and N is negative or zero.\n";
        MPIER = 26;
        if (MPKER[MPIER] == 2)  mpabrt();
      }
      return;
    }
  }
  int nn = abs (n);
  
  if (nn == 0) 
  {
    b.r->nw=1; b.r->sign=1; b.r->exponent=0; (*b.r)[0]=1; (*b.r)[1]=0;
    mp::zero(*b.i);
#if DEBUG == 2
    if (MPIDB >= 6) 
      debug("MPCPWX O ", b);
#endif
    return;
  } 
  
  mp_complex *f = createComplex(6), *sk0 = createComplex(mpnw+3), 
    *sk1 = createComplex(mpnw+3), *sk2 = createComplex(mpnw+3);   
  mpceq (a, *sk0, mpnw); 
  f->r->nw=1; f->r->sign=1; f->r->exponent=0; (*f->r)[0]=1; (*f->r)[1]=0;
  bool skip = false;
  
  if (nn == 1) 
  {
    mpceq (*sk0, *sk2, mpnw); 
    skip = true;
  } 
  else if (nn == 2) 
  {
    mpcmul (*sk0, *sk0, *sk2, mpnw);
    skip = true;
  }
  
  //  Determine the least integer MN such that 2 ^ MN > NN.
  if (!skip)
  {
    int mn, kn, kk;
    t1 = nn;
    mn = int(cl2 * log (t1) + 1.0 + mprxx); // *cast*
    mpceq (*f, *sk2, mpnw);
    kn = nn;
    
    //  Compute B ^ N using the binary rule for exponentiation.
    for (j = 1; j<=mn; j++)
    {
      kk = kn / 2;
      if (kn != 2 * kk) 
      {
        mpcmlx (*sk2, *sk0, *sk1, mpnw);
        mpceq (*sk1, *sk2, mpnw);
      }
      kn = kk;
      if (j < mn) 
      {
        mpcmlx (*sk0, *sk0, *sk1, mpnw);
        mpceq (*sk1, *sk0, mpnw);
      }
    }
  }
  //  Compute reciprocal if N is negative.
  if (n < 0) 
  {
    mpceq (*f, *sk1, mpnw);
    mpcdvx (*sk1, *sk2, *sk0, mpnw);
    mpceq (*sk0, *sk2, mpnw);
  }
  mpceq (*sk2, b, mpnw);
  
#if DEBUG == 2
  if (MPIDB >= 6) 
    debug("MPCPWX O ", b);
#endif
  delete sk0; delete sk1; delete sk2; delete f; 
}

void mp_complex::mpcsqx (const mp_complex& a, mp_complex& b, int mpnw)
{
  mpsize mpnw2(mpnw+2);
  mp sk0(mpnw2), sk1(mpnw2), sk2(mpnw2); 
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 5) 
    debug("MPCSQX I", a);
#endif
  
  if (a.r->nw == 0. && a.i->nw == 0.) 
  {
    zero(b);
#if DEBUG == 2
    if (MPIDB >= 5) 
      debug("MPCSQX O ", b);
#endif
    return;
  }
  
  mp::mpsqx (*a.r, sk0, mpnw);
  mp::mpsqx (*a.i, sk1, mpnw);
  mp::mpadd (sk0, sk1, sk2, mpnw);
  mp::mpsqrx (sk2, sk0, mpnw);
  mp::mpeq (*a.r, sk1, mpnw);
  sk1.sign = 1;
  mp::mpadd (sk0, sk1, sk2, mpnw);
  mp::mpmuld (sk2, mp_dpe(0.50), sk1, mpnw);
  mp::mpsqrx (sk1, sk0, mpnw);
  mp::mpmuld (sk0, mp_dpe(2.0), sk1, mpnw);
  
  if (a.r->sign >= 0) 
  {
    mp::mpeq (sk0, *b.r, mpnw);
    mp::mpdivx (*a.i, sk1, *b.i, mpnw);
  } 
  else 
  {
    mp::mpdivx (*a.i, sk1, *b.r, mpnw);
    b.r->sign = 1;
    mp::mpeq (sk0, *b.i, mpnw);
    b.i->sign = a.i->sign;
  }
  
#if DEBUG == 2
  if (MPIDB >= 5) 
    debug("MPCSQX O ", b); 
#endif
}


