#ifndef MP_DPE
#define MP_DPE

#include "mpfun.h"

/**
 *  This class represents DPE numbers.
 *  @author Herman Harjono.
 *  @version Sep 21 1998.
 */
class mp_dpe: public mpfun
{
  friend ostream& operator<<(ostream& output, const mp_dpe& in)
    { output << in.a*pow(2,in.n); return output;}
	
 private:  
  
  /** 
   *  PT table. It's made static member of the class to preserve 
   *  thread-safety. Required in dpadd.
   *  @see #dpadd
   */
  static const double* pt;
  
  /**
   *  This initializes the pt table.
   */
  static double* init_pt();
  
  
  /**  
   *  Normalize dpe number c to a decent range if it's not.
   */
  static void normalizeDPE(mp_dpe& c);
  
 public:
  double a;
  int n;
  inline mp_dpe() {}
  inline mp_dpe(double A, int N=0) {a = A; n = N;}
  
  /** 
   *  Returns the value of a * 2^n. 
   */
  inline double value() const {return a * pow(2,n);}
  
  /** 
   *  This adds the dpe numbers a and b to yield 
   *  the sum (c,nc).
   */
  static void dpadd (const mp_dpe& a, const mp_dpe& b, mp_dpe& c);
  
  /** 
   * This converts the dpe number a to decimal form
   * i.e. b * 10^nb, where |b| is between 1 and 10.
   */
  static void dpdec(const mp_dpe& a, mp_dpe& b);
  
  /** 
   * This divides the dpe number a by b to yield the quotient
   * c.
   */
  static void dpdiv (const mp_dpe& a, const mp_dpe& b, mp_dpe& c);
  
  /**
   * This multiplies the dpe number a by b to yield the product
   * c.
   */
  static void dpmul(const mp_dpe& a, const mp_dpe& b, mp_dpe& c);
  
  /**
   *  This raises the dpe number a to the b power and places the
   *  result in c.
   */
  static void dppwr(const mp_dpe& a, const mp_dpe& b, mp_dpe& c);
  
  /**
   *  This computes the square root of the dpe number a and places the
   *  result in b.
   */
  static void dpsqrt(const mp_dpe& a, mp_dpe& b);
  
  /**
   *  This subtracts the dpe number b from a to yield the difference
   *  c.
   */
  static void dpsub(const mp_dpe& a, const mp_dpe& b, mp_dpe& c); 
  
};

#endif
