#ifndef MPFUN_H
#define MPFUN_H

#include "mpfun_Includes.h"

class mp_real;
/**  
* This class contains shared variables, macros, and functions that
* are used by other mp-classes that inherit it.
*
* @author Herman Harjono
* @version Sep 21, 1998.
*/ 
class mpfun
{
 public:

/**
  *  Precision level, in digits
  */
  static const int mpipl;
  
  /**
  *  Output precision, in digits.
  */
  static const int mpiou; 
  
  /**
  *  Log_10 of MP epsilon level
  */
  static const int mpiep; 
  
  /**
  *  Precision level, in words.
  */
  static const int mpwds; 

 /**
  * Current output precision (in digits).
  */
  static const int mpoud;
  
  static const int mp2; 

  static const int mp21;

  /**
  * Log 2.
  */
  static mp_real& mpl02; 
  
  /**
  * Log 10.
  */  
  static mp_real& mpl10;
  
  /**
  * Pi.
  */
  static mp_real& mppic;
  
  /**
  * Current MP epsilon value. 10^MPIEP.
  */
  static mp_real& mpeps;
  


protected:
	
/**  
 *  Depends on the system word size.
 * 
 *  On IEEE systems and most other 32 bit systems, set to 4096.D0
 *  On Cray system, set to 2048.D0.
 */
  static const double mpbbx;
  static const double mpbdx; 
  static const double mprdx; 
  static const double mpbx2; 
  static const double mprbx; 
  static const double mprx2; 
  static const double mprxx;
  
	
 /**  
  *  Depends on the system word size.
  * 
  *  On IEEE systems and most other 32 bit systems, set to 24.
  *  On Cray system, set to 22.
  */
  static const int mpnbt;
	
 /**  
  *  Depends on the system word size.
  *
  *  On IEEE systems and most other 32 bit systems, set to 32.
  *  On Cray system, set to 16.
  */
  static const int mpnpr;
	
	/**  
  *  Depends on the system word size.
  *
  *  On IEEE systems and most other 32 bit systems, set to 7.
  *  On Cray system, set to 8.
  */
  static const int mpmcrx;
	
	/**
  *  This is spacing parameters to avoid bank and cache conflicts in the FFT routines.
  *
  *  Value of 32 appears to work well on most systems.
  *  Set mpnrow = 64 on Crays.
  */
  static const int mpnrow;
	
	/**
  *  This is spacing parameters to avoid bank and cache conflicts in the FFT routines.
  *
  *  Value of 3 appears to work well on most systems. 
  */
  static const int mpnsp1; 
	
	/**
  *  This is spacing parameters to avoid bank and cache conflicts in the FFT routines.
  *
  *  Value of 17 appears to work well on most systems.
  */
  static const int mpnsp2;
	
	/**
  *  MPFUN debug level. Initial value: 0
  */
  static int MPIDB; 
	
	/**
  *  Number of words in debug output. Initial value: 22.
  */
  static int MPNDB;
	
	/**
  *  MPFUN error indicator. Initial value: 99.
  */
  static int MPIER; 
	
	/**
  *  Cross-over point for advanced routines. 
  *  Initial value: mpmcrx.
  *  @see #mpmcrx
  */
  static int MPMCR;
	
	/**
  *  MPFUN rounding option. Initial value: 1.
  */
  static int MPIRD;
	
	/**
  *  Array of error options. Initial value: 2.
  */
  static int MPKER[73+1];
  
	/** 
  *  Fortran's sign function.
  *  @return the absolute value of a with sign of b.
  */
  inline static double SIGN(double a, double b)
  { return (b>=0 ? fabs(a) : -fabs(a));}
	
	/**
  *  This routine terminates execution. 
  *
  *  Many users will want to replace the exit system call with a call to 
  *  a system routine that provides a traceback.
  *  Examples of code that produce traceback are included here (commented out)
  *  for some systems.
  */
  static void mpabrt();

  static DComplex *mpuu1;
  static DComplex *mpuu2;

 private:  
 
  /**
  *  This initializes the root of unity arrays mpuu1, which are required
  *  by the FFT routines that are called by MPMULX.
  *
  *  This routine must be called with N set to the largest precision level MPNW 
  *  that will be used in the subsequent application.  
  *  @see #mpuu2
  */
  static DComplex* init_mpuu1(int n);
  
  /**
  *  This initializes the root of unity arrays mpuu2, which are required
  *  by the FFT routines that are called by MPMULX. 
  *  
  *  This routine must be called with N set to the largest precision level MPNW 
  *  that will be use in the subsequent application.
  *  @see #mpuu1
  */
  static DComplex* init_mpuu2(int n);
  
	
public:
/**
*  Current precision level, in words.
*  Initial value: 16.
  */
  static const int MPNW; 
	
	
};

#endif



