#include "mp.h"



// mppic, mpl02, mpl10, mpeps are initialized in mp_real

mp::mp(const mp& in)
{
  maxnw = in.maxnw;
  if(maxnw>0)
  {
    mantissa = new float[maxnw];
    mpeq(in, *this, MPNW);
  }
  else 
  {
    exponent = in.exponent;
    sign = in.sign;
    nw = in.nw;
    mantissa = NULL;
  }
}



#if DEBUG == 2
void mp::debug(char* functionName, const mp& mpVar)
{
  int count; cout.precision(10);
  for(count=0;count< mpVar.nw;count++)
    cout << ' ' << count << "  " << functionName << "  " << mpVar[count] << '.' << endl;
  cout.precision(0);
}
#endif


mp::mp(const mpsize& maxNW)
{
  maxnw=maxNW.size;
  sign=1;
  nw=0;
  exponent=0;
  if(maxnw>=1)
  {
    mantissa = new float[maxnw];		
    mantissa[0]=0;
  }
  else
    mantissa = NULL;
}

mp::mp(const mpprecision& precision)
{
  maxnw = int((precision.size / 7.224719896) + 4);
  if(maxnw < 8) maxnw = 8; // since at least we need 6 (such as in mpdmc), therefore 
	// it's safe to allocate at least 8.
  mantissa = new float[maxnw];
  mantissa[0]=0;
  sign=1;
  nw=0;
  exponent=0;
}

mp::mp(int ia, const mpprecision& precision)
{
  maxnw = int((precision.size / 7.224719896) + 4);
  if(maxnw < 8) maxnw = 8;
  mantissa = new float[maxnw];
  mpdmc(mp_dpe(double(ia)), *this);
}

mp::mp(double da, const mpprecision& precision)
{
  maxnw = int((precision.size / 7.224719896) + 4);
  if(maxnw < 8) maxnw = 8; // since at least we need 6 (such as in mpdmc), therefore 
	// it's safe to allocate at least 8.
  mantissa = new float[maxnw];
  mpdmc(mp_dpe(da), *this);
}

mp::mp(char* ab, const mpprecision& precision)
{
  maxnw = int((precision.size / 7.224719896) + 4);
  if(maxnw < 8) maxnw = 8; // since at least we need 6 (such as in mpdmc), therefore 
	// it's safe to allocate at least 8.
  mantissa = new float[maxnw];
  mpdexc (ab, strlen(ab), *this, MIN(MPNW, maxnw-2));
}


int operator==(const mp& ja, const mp& jb)
{
  int ic;
  mp::mpcpr(ja, jb, ic, mpfun::MPNW);
  return ic == 0 ? 1 : 0;
}

int operator!=(const mp& ja, const mp& jb)
{
  int ic;  
  mp::mpcpr(ja, jb, ic, mpfun::MPNW);
  return ic != 0 ? 1 : 0;
}


int operator<=(const mp& ja, const mp& jb)
{
  int ic;
  mp::mpcpr(ja, jb, ic, mpfun::MPNW);
  return ic <= 0 ? 1 : 0;
}

int operator>=(const mp& ja, const mp& jb)
{
  int ic;
  mp::mpcpr(ja, jb, ic, mpfun::MPNW);
  return ic >= 0 ? 1 : 0;
}

int operator<(const mp& ja, const mp& jb)
{
  int ic;
  mp::mpcpr(ja, jb, ic, mpfun::MPNW);
  return ic < 0 ? 1 : 0;
}

int operator>(const mp& ja, const mp& jb)
{
  int ic;
  mp::mpcpr(ja, jb, ic, mpfun::MPNW);
  return ic > 0 ? 1 : 0;
}

double dble(const mp& ja) 
{
  return ja.toDPE().value();
}

int integer(const mp& ja)
{
  return int(ja.toDPE().value());
}

float real(const mp& ja)
{
  return float(ja.toDPE().value());
}


istream& operator>>(istream& s, mp& ja)
{
  int i = 0;
  char c = ' ', *az = new char[mp::mpipl + 100]; 
  while(s>>c)  
  {
    if (c == ',') break;
    if (c == '\n' || c == '\t' || c == ' ') continue;
    az[i++] = c;
  }
  if(i!=0)
    mp::mpdexc(az, i, ja, mp::MPNW);
  delete [] az;
  return s;
}

ostream& operator<<(ostream& s, const mp& ja)
{
  char *az = new char[mp::mpipl + 100];
  int nd;
  mp::mpouts(ja, mp::mpoud, az, nd, mp::MPNW);
  int i;
  for(i=0; i<nd; i++)
  {
    s << az[i];
    if((i+1)%78 ==0)
      s<< endl;
  }
	
  s << endl;
  s.flush();
  delete [] az;
  return s;
}




