#include "mp.h"

void mp::mpang (const mp& x, const mp& y, const mp& pi, mp& a, int mpnw)
{
  int k;
  const double cl2= 1.4426950408889633, cpi = 3.141592653589793;
  const int nit=3;
  mpsize mpnw3(mpnw+3);
  mp sk0(mpnw3), sk1(mpnw3),sk2(mpnw3), 
    sk3(mpnw3), sk4(mpnw3);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(a);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 5) 
  {
    mpdeb ("MPANG I ", x, mpnw);
    mpdeb ("MPANG I ", y, mpnw);
  }
#endif
  
  int ix = x.sign;
  int nx = MIN (x.nw, mpnw);
  int iy = y.sign;
  int ny = MIN (y.nw, mpnw);
  
  //  Check if both X and Y are zero.
  
  if (nx == 0 && ny == 0) 
  {
    if (MPKER[7] != 0) 
    {
      cerr << "*** MPANG: Both arguments are zero.\n";
      MPIER = 7;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  //  Check if Pi has been precomputed.
  mp_dpe t1;
  mpmdc (pi, t1);
  if (t1.n != 0 || fabs (t1.a - cpi) > mprx2) 
  {
    if (MPKER[8] != 0) 
    {
      cerr << "*** MPANG: PI must be precomputed.\n";
      MPIER = 8;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  //  Check if one of X or Y is zero.
  
  if (nx == 0) 
  {
    if (iy > 0) 
      mpmuld (pi, mp_dpe(0.5), a, mpnw); 
    else 
      mpmuld (pi, mp_dpe(-0.5), a, mpnw); 
    
#if DEBUG == 2
    if (MPIDB >= 5)  
      mpdeb ("MPANG O ", a, mpnw);
#endif
    return;
    
  } 
  else if (ny == 0) 
  {
    if (ix > 0) 
      zero(a);
    else 
      mpeq (pi, a, mpnw);
    
#if DEBUG == 2
    if (MPIDB >= 5)  
      mpdeb ("MPANG O ", a, mpnw);
#endif
    return;
  }
  
  int nws = mpnw;
  mpnw++;
  
  //  Determine the least integer MQ such that 2 ^ MQ >= MPNW.
  
  t1.a = nws;
  int mq = int(cl2 * log (t1.a) + 1.0 - mprxx); // *cast*
  
  //  Normalize x and y so that x^2 + y^2 = 1.
  
  mpmul (x, x, sk0, mpnw);
  mpmul (y, y, sk1, mpnw);
  mpadd (sk0, sk1, sk2, mpnw);
  mpsqrt (sk2, sk3, mpnw);
  mpdiv (x, sk3, sk1, mpnw);
  mpdiv (y, sk3, sk2, mpnw);
  
  //  Compute initial approximation of the angle.
  
  mpmdc (sk1, t1);
  mp_dpe t2;
  mpmdc (sk2, t2);
  t1.n = MAX (t1.n, -66);
  t2.n = MAX (t2.n, -66);
  t1.a = t1.a * pow(2.0, t1.n);
  t2.a = t2.a * pow(2.0, t2.n);
  mp_dpe t3;
  t3.a = atan2 (t2.a, t1.a);
  mpdmc (t3, a);
  
  //  The smaller of x or y will be used from now on to measure convergence.
  //  This selects the Newton iteration (of the two listed above) that has the
  //  largest denominator.
  
  int kk;
  if (fabs (t1.a) <= fabs (t2.a)) 
  {
    kk = 1;
    mpeq (sk1, sk0, mpnw);
  } 
  else {
    kk = 2;
    mpeq (sk2, sk0, mpnw);
  }
  
  mpnw = 3;
  int iq = 0;
  
  //  Perform the Newton-Raphson iteration described above with a dynamically
  //  changing precision level MPNW (one greater than powers of two).
  
  for (k = 2;  k<=mq; k++)
  {
    mpnw = MIN (2 * mpnw - 2, nws) + 1;
    bool cont = true;
    while(cont)
    {
      mpcssn (a, pi, sk1, sk2, mpnw);
      if (kk == 1)
      {
        mpsub (sk0, sk1, sk3, mpnw);
        mpdiv (sk3, sk2, sk4, mpnw);
        mpsub (a, sk4, sk1, mpnw);
      } 
      else 
      {
        mpsub (sk0, sk2, sk3, mpnw);
        mpdiv (sk3, sk1, sk4, mpnw);
        mpadd (a, sk4, sk1, mpnw);
      }
      mpeq (sk1, a, mpnw);
      if (k == mq - nit && iq == 0)
        iq = 1;
      else
        cont = false;
    }
  }
  
  mproun (a, nws);
  
#if DEBUG == 2
	if (MPIDB >= 5)  
		mpdeb ("MPANG O ", a, nws);
#endif
	
}

void mp::mpcssh (const mp& a, const mp& al2, mp& x, mp& y, int mpnw)
{
  mpsize mpnw3(mpnw+3);
  mp f(mpsize(6)),sk0(mpnw3), sk1(mpnw3), 
    sk2(mpnw3), sk3(mpnw3); 
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(x); zero(y);
    return;
  }
#if DEBUG == 2
  if (MPIDB >= 5)  mpdeb ("MPCSSH I ", a, mpnw);
#endif
  int nws = mpnw;
  mpnw = mpnw + 1;
  f.sign=1; f.nw=1; f.exponent=0;
  f[0]=1; f[1]=0;
  
  mpexp (a, al2, sk0, mpnw);
  mpdiv (f, sk0, sk1, mpnw);
  mpadd (sk0, sk1, sk2, mpnw);
  mpmuld (sk2, mp_dpe(0.5), sk3, mpnw);
  mpeq (sk3, x, mpnw);
  mpsub (sk0, sk1, sk2, mpnw);
  mpmuld (sk2, mp_dpe(0.5), sk3, mpnw);
  mpeq (sk3, y, mpnw);  
  
  
  mproun (x, nws);
  mproun (y, nws);
  
#if DEBUG == 2
  if (MPIDB >= 5) 
  {
    mpdeb ("MPCSSH O", x, nws);
    mpdeb ("MPCSSH O", y, nws);
  }
#endif
}

void mp::mpcssn (const mp& a, const mp& pi, mp& x, mp& y, int mpnw)
{
  const double cpi = 3.141592653589793;
  double t2;
  mpsize mpnw3(mpnw+3);
  mp f(mpsize(6)), sk0(mpnw3), sk1(mpnw3), 
    sk2(mpnw3), sk3(mpnw3), sk4(mpnw3),
    sk5(mpnw3), sk6(mpnw3);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    if (MPIER == 99)  mpabrt();
    zero(x); zero(y);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 6)  mpdeb ("MPCSSN I ", a, mpnw);
#endif
	
  int na = MIN (a.nw, mpnw);
  int l1; 
  if (na == 0) 
  {
    x.sign=1; x.nw=1; x.exponent=0; x[0]=1;
    zero(y);
    l1 = 0;
#if DEBUG == 2
    if (MPIDB >= 6) 
    {
      cerr << "Iteration count: " << l1 << endl;
      mpdeb ("MPCSSN O ", x, mpnw);
      mpdeb ("MPCSSN O ", y, mpnw);
    }
#endif
    return;   
  }
  
  //  Check if Pi has been precomputed.
  mp_dpe t1;
  mpmdc (pi, t1);
  if (t1.n != 0 || fabs (t1.a - cpi) > mprx2) 
  {
    if (MPKER[28] != 0) 
    {
      cerr << "*** MPCSSN: PI must be precomputed.\n";
      MPIER = 28;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  int nws = mpnw;
  mpnw = mpnw + 1;
  f.nw=1; f.sign=1; f.exponent=0;
  f[0]=1; f[1]=0;
  
  //  Reduce to between - Pi and Pi.
  
  mpmuld (pi, mp_dpe(2.0), sk0, mpnw);
  mpdiv (a, sk0, sk1, mpnw);
  mpnint (sk1, sk2, mpnw);
  mpsub (sk1, sk2, sk3, mpnw);
  
  //  Determine nearest multiple of Pi / 2, and within a quadrant, the nearest
  //  multiple of Pi / 16.  Through most of the rest of this subroutine, KA and
  //  KB are the integers a and b of the algorithm above.
  
  mpmdc (sk3, t1);
  int ka, kb;
  if (t1.n >= - mpnbt) 
  {
    t1.a *= pow(2.0, t1.n);
    t2 = 4.0 * t1.a;  
		
    ka = int(NINT (t2)); // *cast* 
    kb = int(NINT (8.0 * (t2 - ka))); // *cast*
  } 
  else 
  {
    ka = 0;
    kb = 0;
  }
  
  t1.a = (8 * ka + kb) / 32.0;
  t1.n = 0;
  mpdmc (t1, sk1);
  mpsub (sk3, sk1, sk2, mpnw);
  mpmul (sk0, sk2, sk1, mpnw);
  
  //  Compute cosine and sine of the reduced argument s using Taylor's series.
  
  if (sk1.nw == 0) 
  {
    zero(sk0);
    l1 = 0;
  }
  else
  {
    mpeq (sk1, sk0, mpnw);
    mpmul (sk0, sk0, sk2, mpnw);
    l1 = 0;
    
    // 100
    do 
    {
      l1 = l1 + 1;
      if (l1 == 10000) 
      {
        if (MPKER[29] != 0) 
        {
          cerr << "*** MPCSSN: Iteration limit exceeded.\n";
          MPIER = 29;
          if (MPKER[MPIER] == 2)  mpabrt();
          return;
        }
      }
      
      t2 = - (2.0 * l1) * (2.0 * l1 + 1.0);
      mpmul (sk2, sk1, sk3, mpnw);
      mpdivd (sk3, mp_dpe(t2), sk1, mpnw);
      mpadd (sk1, sk0, sk3, mpnw);
      mpeq (sk3, sk0, mpnw);
    }
    //  Check for convergence of the series.
    while (sk1.nw != 0 && sk1.exponent >= sk0.exponent - mpnw); 
  }
  //  Compute Cos (s) = Sqrt [1 - Sin^2 (s)].
  
  mpeq (sk0, sk1, mpnw);
  mpmul (sk0, sk0, sk2, mpnw);
  mpsub (f, sk2, sk3, mpnw);
  mpsqrt (sk3, sk0, mpnw);
  
  //  Compute cosine and sine of b * Pi / 16.
  int kc;
  kc = abs (kb);
  f[0] = 2.;
  if (kc == 0) 
  {
    sk2.sign=1; sk2.nw=1; sk2.exponent=0;
    sk2[0]=1;
    zero(sk3);
  } 
  else 
  {
    switch (kc)
    {
    case 1: 
      mpsqrt (f, sk4, mpnw);
      mpadd (f, sk4, sk5, mpnw);
      mpsqrt (sk5, sk4, mpnw);
      break;
    case 2: 
      mpsqrt (f, sk4, mpnw);
      break;
    case 3: 
      mpsqrt (f, sk4, mpnw);
      mpsub (f, sk4, sk5, mpnw);
      mpsqrt (sk5, sk4, mpnw);  
      break;
    case 4: 
      zero(sk4);
      break;
    }
    mpadd (f, sk4, sk5, mpnw);
    mpsqrt (sk5, sk3, mpnw);
    mpmuld (sk3, mp_dpe(0.5), sk2, mpnw);
    mpsub (f, sk4, sk5, mpnw);
    mpsqrt (sk5, sk4, mpnw);
    mpmuld (sk4, mp_dpe(0.5), sk3, mpnw);
  }
  if (kb < 0) sk3.sign = - sk3.sign;
  
  //  Apply the trigonometric summation identities to compute cosine and sine of
  //  s + b * Pi / 16.
  
  mpmul (sk0, sk2, sk4, mpnw);
  mpmul (sk1, sk3, sk5, mpnw);
  mpsub (sk4, sk5, sk6, mpnw);
  mpmul (sk1, sk2, sk4, mpnw);
  mpmul (sk0, sk3, sk5, mpnw);
  mpadd (sk4, sk5, sk1, mpnw);
  mpeq (sk6, sk0, mpnw);
  
  //  This code in effect applies the trigonometric summation identities for
  //  (s + b * Pi / 16) + a * Pi / 2.
  
  switch(ka)
  {
  case 0: 
    mpeq (sk0, x, mpnw);
    mpeq (sk1, y, mpnw);
    break;
  case 1: 
    mpeq (sk1, x, mpnw);
    x.sign = - x.sign;
    mpeq (sk0, y, mpnw);
    break;
  case -1: 
    mpeq (sk1, x, mpnw);
    mpeq (sk0, y, mpnw);
    y.sign = - y.sign;
    break; 
  case 2: case -2: 
    mpeq (sk0, x, mpnw);
    x.sign = - x.sign;
    mpeq (sk1, y, mpnw);
    y.sign = - y.sign;
    break;
  }
  
  mproun (x, nws);
  mproun (y, nws);
  
#if DEBUG == 2
  if (MPIDB >= 6) 
  {
    cerr << "Iteration count: " << l1 << endl;
    mpdeb ("MPCSSN O ", x, nws);
    mpdeb ("MPCSSN O ", y, nws);
  }
#endif
}

void mp::mpexp (const mp& a, const mp& al2, mp& b, int mpnw)
{
  int i, l1;
  const int nq = 8;
  const double alt = 0.693147180559945309;
  mpsize mpnw3(mpnw+3);
  mp f(mpsize(6)), sk0(mpnw3), sk1(mpnw3), 
    sk2(mpnw3), sk3(mpnw3);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
#if DEBUG == 2
  if (MPIDB >= 7)  mpdeb ("MPEXP I ", a, mpnw);
#endif
  
  mp_dpe t1;
  mpmdc (a, t1);
  t1.a = t1.value();
	
  //  Unless the argument is near Log (2), Log(2) must be precomputed.  This
  //  exception is necessary because MPLOG calls MPEXP to initialize Log (2).
  
  mp_dpe t2;
  if (fabs (t1.a - alt) > mprdx) 
  {
    mpmdc (al2, t2);
    if (t2.n != - mpnbt || fabs (t2.a * pow(0.50, mpnbt) - alt) > mprx2) 
    {
      if (MPKER[34] != 0) 
      {
        cerr << "*** MPEXP: LOG (2) must be precomputed.\n";
        MPIER = 34;
        if (MPKER[MPIER] == 2)  mpabrt();
      }
      return;
    }
  }
  
  //  Check for overflows and underflows.
  
  if (t1.a >= 1e9) 
  {
    if (t1.a > 0.0) 
    {
      if (MPKER[35] != 0) 
      {
        cerr.width(12);
        cerr.precision(6);
        cerr << "*** MPEXP: Argument is too large "
          << t1.a << " x 10 ^";
        cerr.width(0);
        cerr.precision(0);
        cerr << t1.n;
        MPIER = 35;
        if (MPKER[MPIER] == 2)  mpabrt();
      }
      return;
    } 
    else 
    {
      zero(b);
      l1 = 0;
#if DEBUG == 2
      if (MPIDB >= 7) 
      {
        cerr << "Iteration count: " << l1 << endl;
        mpdeb ("MPEXP O ", b, mpnw);
      }
#endif
      return;
    }
  }
  
  int nws = mpnw;
  mpnw++;
  f.sign=1; f.nw=1; f.exponent=0;
  f[0]=1; f[1]=0;
  
  //  Compute the reduced argument A' = A - Log(2) * Nint [A / Log(2)].  Save
  //  NZ = Nint [A / Log(2)] for correcting the exponent of the final result.
  int nz;
  if (fabs (t1.a - alt) > mprdx) 
  {
    mpdiv (a, al2, sk0, mpnw);
    mpnint (sk0, sk1, mpnw);
    mpmdc (sk1, t1);
    nz = int(t1.value() + SIGN (mprxx, t1.a)); // *cast*
    mpmul (al2, sk1, sk2, mpnw);
    mpsub (a, sk2, sk0, mpnw);
  } 
  else 
  {
    mpeq (a, sk0, mpnw);
    nz = 0;
  }
  
  double tl = sk0.exponent - mpnw;
  
  //  Check if the reduced argument is zero.
  
  bool skip = false;
  if (sk0.nw == 0) 
  {
    sk0.nw=1; sk0.sign=1;
    sk0.exponent=0;
    l1 = 0;
    skip = true;
  }
  
  if (!skip)
  {
    //  Divide the reduced argument by 2 ^ NQ.  
    mpdivd (sk0, mp_dpe(1.0, nq), sk1, mpnw);
    
    //  Compute Exp using the usual Taylor series.    
    mpeq (f, sk2, mpnw);
    mpeq (f, sk3, mpnw);
    l1 = 0;
    t2.n = 0; 
    do 
    {
      l1 = l1 + 1;
      if (l1 == 10000) 
      {
        if (MPKER[36] != 0) 
        {
          cerr << "*** MPEXP: Iteration limit exceeded.\n";
          MPIER = 36;
          if (MPKER[MPIER] == 2)  mpabrt();
          return;
        }
      }
      
      t2.a = l1;
      mpmul (sk2, sk1, sk0, mpnw);
      mpdivd (sk0, t2, sk2, mpnw);
      mpadd (sk3, sk2, sk0, mpnw);
      mpeq (sk0, sk3, mpnw);
    }
    //  Check for convergence of the series.
    while (sk2.nw != 0. && sk2.exponent >= tl); 
    
    //  Raise to the (2 ^ NQ)-th power.   
    for (i = 1; i<=nq; i++)
    {
      mpmul (sk0, sk0, sk1, mpnw);
      mpeq (sk1, sk0, mpnw);
    }   
  }
  //  Multiply by 2 ^ NZ.
  mpmuld (sk0, mp_dpe(1.0, nz), sk1, mpnw);
  mpeq (sk1, b, mpnw);  
  mproun (b, nws);
  
#if DEBUG == 2
  if (MPIDB >= 7) 
  {
    cerr << "Iteration count: " << l1 << endl;
    mpdeb ("MPEXP O ", b, nws);
  }
#endif  
	
}

void mp::mplog (const mp& a, const mp& al2, mp& b, int mpnw)
{
  int k;
  const double alt = 0.693147180559945309, 
    cl2 = 1.4426950408889633;
  const int nit = 3;
  mpsize mpnw3(mpnw+3);
  mp sk0(mpnw3), sk1(mpnw3), sk2(mpnw3);
  
  if (MPIER != 0) 
	{
    if (MPIER == 99)  mpabrt();
    b.nw = 0; b.exponent=0; b.sign=1;
    return;
	}
  
#if DEBUG == 2
  if (MPIDB >= 6)  mpdeb ("MPLOG I ", a, mpnw);
#endif
  
  int na = MIN (a.nw, mpnw);
  
  if (a.sign < 0 || na == 0) 
	{
    if (MPKER[50] != 0) 
    {
      cerr << "*** MPLOG: Argument is less than or equal to zero.\n";
      MPIER = 50;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
	}
  
  //  Unless the input is close to 2, Log (2) must have been precomputed.
  mp_dpe t1, t2;
  mpmdc (a, t1);
  if (fabs (t1.a - 2.0) > 1e-3 || t1.n != 0) 
  {
    mpmdc (al2, t2);
    if (t2.n != - mpnbt || fabs (t2.a * pow(0.50, mpnbt) - alt) > mprx2) 
    {
      if (MPKER[51] != 0) 
      {
        cerr << "*** MPLOG: LOG (2) must be precomputed.\n";
        MPIER = 51;
        if (MPKER[MPIER] == 2)  mpabrt();
      }
      return;
    }
  }
  
  //  Check if input is exactly one.
  if (a.nw == 1 && a.sign == 1 && a.exponent == 0 && a[0] == 1.) 
	{
    b.nw = 0; b.exponent=0; b.sign=1;
#if DEBUG == 2
    if (MPIDB >= 6)  mpdeb ("MPLOG O ", b, mpnw);
#endif    
    return;
	}
  int nws = mpnw;
  
  //  Determine the least integer MQ such that 2 ^ MQ >= MPNW.
  t2.a = nws;
  int mq = int(cl2 * log (t2.a) + 1.0 - mprxx); // *cast*
  
  
  //  Compute initial approximation of Log (A).
  t1.a = log (t1.a) + t1.n * alt;
  t1.n = 0;
  mpdmc (t1, b);
  mpnw = 3;
  int iq = 0;
  
  //  Perform the Newton-Raphson iteration described above with a dynamically
  //  changing precision level MPNW (one greater than powers of two).
  
  for(k = 2; k<=mq; k++)
  {
    mpnw = MIN (2 * mpnw - 2, nws) + 1 ;
    bool cont = true;
    while(cont)
    { 
      mpexp (b, al2, sk0, mpnw);
      mpsub (a, sk0, sk1, mpnw);
      mpdiv (sk1, sk0, sk2, mpnw);
			
      mpadd (b, sk2, sk1, mpnw);
			
      mpeq (sk1, b, mpnw);
      if (k == mq - nit && iq == 0) 
        iq = 1;
      else
        cont = false;
    }
  }
  
  mproun (b, nws);
  
#if DEBUG == 2
  if (MPIDB >= 6)  mpdeb ("MPLOG O ", b, nws);
#endif
	
}

void mp::mppi (mp& pi, int mpnw)
{
  int k;
  double t1;
  const double cl2 = 1.4426950408889633;
  mpsize mpnw3(mpnw+3);
  mp f(mpsize(6)), sk0(mpnw3), sk1(mpnw3), 
    sk2(mpnw3), sk3(mpnw3), sk4(mpnw3);
  
  if (MPIER != 0) 
	{
    if (MPIER == 99)  mpabrt();
    zero(pi);
    return;
  }
  
  //  Perform calculations to one extra word accuracy.
  int nws = mpnw;
  mpnw = mpnw + 1;
  
  //  Determine the number of iterations required for the given precision level.
  //  This formula is good only for this Pi algorithm.
  
  t1 = nws * log10 (mpbdx);
  int mq = int(cl2 * (log (t1) - 1.0) + 1.0); // *cast*
  
  //  Initialize as above.
  
  sk0.nw=1; sk0.sign=1; sk0.exponent=0; sk0[0]=1;
  f.nw=1; f.sign=1; f.exponent=0; f[0]=2; f[1]=0;
  
  mpsqrt (f, sk2, mpnw);
  mpmuld (sk2, mp_dpe(0.50), sk1, mpnw);
  f.exponent = -1;
  f[0] = float(0.50 * mpbdx); // *cast*
  mpsub (sk2, f, sk4, mpnw);
  
  //  Perform iterations as described above.
  
  for (k = 1; k<= mq; k++)
  {
    mpadd (sk0, sk1, sk2, mpnw);
    mpmul (sk0, sk1, sk3, mpnw);
    mpsqrt (sk3, sk1, mpnw);
    mpmuld (sk2, mp_dpe(0.50), sk0, mpnw);
    mpsub (sk0, sk1, sk2, mpnw);
    mpmul (sk2, sk2, sk3, mpnw);
    t1 = pow(2.0, k);
    mpmuld (sk3, mp_dpe(t1), sk2, mpnw);
    mpsub (sk4, sk2, sk3, mpnw);
    mpeq (sk3, sk4, mpnw);
  }
  
  //  Complete computation.
  mpadd (sk0, sk1, sk2, mpnw);
  mpmul (sk2, sk2, sk3, mpnw);
  mpdiv (sk3, sk4, sk2, mpnw);
  mpeq (sk2, pi, mpnw);
  
  mproun (pi, nws);
  
#if DEBUG == 2
  if (MPIDB >= 7)  mpdeb ("MPPI O ", pi, nws);
#endif
}


