#include "mp.h"

void mp::mpadd (const mp& a, const mp& b, 
                mp& c, int mpnw)
{
  int i;
  
  if (MPIER != 0) 
  {
    zero(c);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 9) 
  {
    debug("mpadd ", a);
    debug("mpadd ", b);
  }
#endif
  
  int na = MIN (a.nw, mpnw);
  int nb = MIN (b.nw, mpnw);
  
  //  checks for zero inputs.
  if (na == 0) 
  {
    
    //  a is zero -- the result is b.
    c.sign=b.sign;
    c.nw=nb;
    c.exponent=b.exponent;
    
    for(i=0;i<nb;i++)
       c[i] = b[i];
    
#if DEBUG == 2
    if (MPIDB >= 9) 
      debug("mpadd ", c);
#endif
    
    return;
  } 
  else if (nb == 0) 
  {
    
    //  b is zero -- the result is a.
    
    c.sign=a.sign;
    c.nw=na;
    c.exponent=a.exponent;
    
    for (i=0;i<na;i++)
      c[i]=a[i];
    
#if DEBUG == 2
    if (MPIDB >= 9) 
      debug("mpadd ", c);
#endif
    
    return;
  }
  
  double* d = new double[mpnw+5];  
  double db;
  if (a.sign == b.sign) 
    db = 1.0;  
  else
    db = -1.0;
  
  int ixa = a.exponent;  
  int ixb = b.exponent; 
  int ish = ixa - ixb;
  
#if DEBUG == 2  
  if (MPIDB >= 9) 
    cerr <<  "ish = " << ish << endl;
#endif
  int nd,ixd;
  
  if (ish >= 0)
  {
    //  A has greater exponent than B, so B must be shifted to the right.
    int m1 = MIN (na, ish);
    int m2 = MIN (na, nb + ish);
    int m3 = na;
    int m4 = MIN (MAX (na, ish), mpnw + 1); 
    int m5 = MIN (MAX (na, nb + ish), mpnw + 1); 
    d[0]=0;
    d[1]=0;
    
    for (i = 0; i<m1; i++)
      d[i+2] = a[i];
    
    for (i = m1; i<m2; i++)
      d[i+2] = a[i] + db * b[i-ish];
    
    for (i = m2; i<m3; i++)
      d[i+2] = a[i];
    
    for (i = m3; i<m4; i++)
      d[i+2] = 0.0;
    
    for (i = m4; i <m5; i++)
      d[i+2] = db * b[i-ish];
    
    nd = m5;
    ixd = ixa;
    d[nd+2] = 0.0;
    d[nd+3] = 0.0;
  } 
  else
  {
    //  B has greater exponent than A, so A must be shifted to the right.
    int nsh = - ish;
    int m1 = MIN (nb, nsh);
    int m2 = MIN (nb, na + nsh);
    int m3 = nb;
    int m4 = MIN (MAX (nb, nsh), mpnw+1); 
    int m5 = MIN (MAX (nb, na + nsh), mpnw+1); 
    d[0]=0;
    d[1]=0;
    
    for (i = 0; i<m1; i++)
      d[i+2] = db * b[i]; 
    
    for (i = m1; i<m2; i++)
      d[i+2] = a[i-nsh] + db * b[i]; 
    
    for (i = m2; i<m3; i++)
      d[i+2] = db * b[i]; 
    
    for (i = m3; i<m4; i++)
      d[i+2] = 0.0;
    
    for (i = m4; i< m5; i++)
      d[i+2] = a[i-nsh];      
    
    nd = m5;
    ixd = ixb;
    d[nd+2] = 0.0;
    d[nd+3] = 0.0;
  }
  
  //  Call mpnorm to fix up result and store in c.
  d[0] = SIGN(nd, a.sign);
  d[1]=ixd;
  //mpnorm(d,c,mpnw, false); // !!!!!!!!!!!!
    mpnorm(d,c,mpnw);
  delete [] d;
  
#if DEBUG == 2
  if (MPIDB >= 9) 
    debug("mpadd ", c);
#endif
}

void mp::mpcbrt (const mp& a, mp& b, int mpnw)
{
  const double cl2 = 1.4426950408889633;
  const int nit = 3;
  mpsize mpnw3(mpnw+3);
  mp f(mpsize(6)), sk0(mpnw3), sk1(mpnw3), 
    sk2(mpnw3);
  
  if (MPIER != 0)
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 7) 
    debug("mpcbrt ", a);
#endif
  
  int na = MIN (a.nw, mpnw); 
  
  if (na == 0)
  {
    zero(b);
#if DEBUG == 2
    debug("mpcbrt ", b);
#endif
    return;
  }
  
  if (a.sign < 0.0) 
  {
    if (MPKER[13] != 0) 
    {
      cerr << "*** MPCBRT: Argument is negative." << endl;
      MPIER = 13;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  // saving initial mpnw;
  int nws=mpnw;
  
  //  Determine the least integer MQ such that 2 ^ MQ >= MPNW.
  mp_dpe t1;
  t1.a = mpnw; 
  int mq = int(cl2 * log (t1.a) + 1.0 - mprxx); // *cast*
  
  //  Compute A^2 outside of the iteration loop.
  
  mpnw++;
  mpmul (a, a, sk0, mpnw);
  
  //  Compute the initial approximation of A ^ (-2/3).
  mpmdc (a, t1);
  mp_dpe t2;
  t2.n = - 2 * t1.n / 3;
  t2.a = pow((t1.a * pow(2.0,(t1.n + 3.0 * t2.n / 2.0))), (-2.0 / 3.0));
  mpdmc (t2, b);
  f.sign=1;
  f.nw=1;
  f.exponent=0;
  f[0]=1;
  f[1]=0;
  
  mpnw=3; 
  int iq = 0;
  
  //  Perform the Newton-Raphson iteration described above with a dynamically
  //  changing precision level MPNW (one greater than powers of two).
  
  for (int k = 2; k<=mq - 1;k++)
  {
    int nw1 = mpnw; 
    mpnw=MIN (2 * mpnw - 2, nws) + 1;
    int nw2 = mpnw; 
    
    bool cont=true;
    while (cont)
    {
      mpmul (b, b, sk1, mpnw);
      mpmul (b, sk1, sk2, mpnw);
      mpmul (sk0, sk2, sk1, mpnw);
      mpsub (f, sk1, sk2, mpnw);
      
      mpnw=nw1;
      
      mpmul (b, sk2, sk1, mpnw);
      mpdivd (sk1, mp_dpe(3.0), sk2, mpnw);
      
      mpnw=nw2;
      
      mpadd (b, sk2, sk1, mpnw);
      mpeq (sk1, b, mpnw);
      
      if (k == mq - nit && iq == 0) 
      {
        iq = 1;
      }
      else
        cont=false;
    }
  }
  
  //  Perform last iteration using Karp's trick.
  
  mpmul (a, b, sk0, mpnw);
  int nw1 = mpnw; 
  mpnw = MIN (2 * mpnw - 2, nws) + 1;
  
  int nw2 = mpnw; 
  
  mpmul (sk0, sk0, sk1, mpnw);
  mpmul (sk0, sk1, sk2, mpnw);
  mpsub (a, sk2, sk1, mpnw);
  
  mpnw=nw1;
  
  mpmul (sk1, b, sk2, mpnw);
  mpdivd (sk2, mp_dpe(3.0), sk1, mpnw);
  
  mpnw=nw2;
  
  mpadd (sk0, sk1, sk2, mpnw);
  mpeq (sk2, b, mpnw);
  
  mproun (b, nws);
  
#if DEBUG == 2
  if (MPIDB >= 7) 
    debug("mpcbrt ", b);
#endif
}

void mp::mpcpr (const mp& a, const mp& b, int& ic, int mpnw)
{
  
  int i;
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    ic = 0;
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 9) 
  {
    debug("MPCPR I ", a);
    debug("MPCPR I ", b);
  }
#endif
  
  int ia = a.sign;
  if (a.nw == 0.) ia = 0;
  int ib = b.sign;
  if (b.nw == 0.) ib = 0;
  
  int na = MIN (a.nw, mpnw);
  int nb = MIN (b.nw, mpnw);
  
  //  Compare signs.
  if (ia != ib) 
    ic = int(SIGN (1, ia - ib)); // *cast*
  
  //  The signs are the same.  Compare exponents.
  else if (a.exponent != b.exponent) 
    ic = int(ia * SIGN (1, a.exponent - b.exponent)); // *cast*
  
  //  The signs are the exponents are the same. 
  //  Compare the actual (common) mantissas.
  else  
  { 
    bool sameMantissas=true;
    for (i = 0; i<MIN (na, nb);i++)
    {
      if (a[i] != b[i]) 
      {
        ic = int(ia * SIGN (1., a[i] - b[i])); // *cast*
        sameMantissas=false;
        break;
      }
    }
    
    if(sameMantissas) 
      if (na != nb) // compares the length of mantissa array
        ic = int(ia * SIGN (1, na - nb));// *cast*
      else
        //  The signs, exponents, mantissas and lengths are the same.  Thus A = B.
        ic = 0;
  }
  
  // The signs and exponents are the same. Compare lengths of mantissas.
  
#if DEBUG == 2
  if (MPIDB >= 9) 
    cerr << "MPCPR O "  << ic << endl;
#endif
}

void mp::mpdiv (const mp& a, const mp& b, mp& c, int mpnw)
{
  
  int i,j;
  double rb, ss, t0, t1, t2;
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(c);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 8) 
  {
    debug("MPDIV I ", a);
    debug("MPDIV I ", b);
  }
#endif
  
  int na = MIN (a.nw, mpnw);
  int nb = MIN (b.nw, mpnw);
  
  //  Check if dividend is zero.
  if (na == 0) 
  {
    zero(c);
#if DEBUG == 2
    if (MPIDB >= 8)
      debug("MPDIV O ", c);
    
    return;
#endif
  }
  
  if (nb == 1 && b[0] == 1.) 
  {
    
    //  Divisor is 1 or -1 -- result is A or -A.
    c.nw=na;
    c.sign=a.sign*b.sign;
    c.exponent=a.exponent-b.exponent;
    
    for (i = 0; i<na;i++)
      c[i] = a[i];
    
#if DEBUG == 2
    if (MPIDB >= 8)
      debug("MPDIV O ", c);
#endif
    
    return;
  }
  
  //  Check if divisor is zero.
  if (nb == 0) 
  {
    if (MPKER[31] != 0) 
    {
      cerr << "*** MPDIV: Divisor is zero.\n";
      MPIER = 31;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  //  Initialize trial divisor and trial dividend.
  double* d = new double[mpnw+4];  
  t0 = mpbdx * b[0];
  if (nb >= 2) t0 = t0 + b[1];
  if (nb >= 3) t0 = t0 + mprdx * b[2];
  if (nb >= 4) t0 = t0 + mprx2 * b[3];
  rb = 1.0 / t0;
  int md = MIN (na + nb, mpnw);
  d[0]=0.0;
  
  for (i=1; i<=na; i++)
    d[i] = a[i-1];
  
  for (i = na + 1; i<md + 4; i++)
    d[i] = 0.0;
  
  
  //  Perform ordinary long division algorithm.  First compute only the first
  //  NA words of the quotient.
  
  for(j = 2; j<=na + 1; j++)
  {
    t1 = mpbx2 * d[j-2] + mpbdx * d[j-1] + d[j] + mprdx * d[j+1];
    t0 = int (rb * t1);
    int j3 = j - 3;
    int i2 = MIN (nb, mpnw + 2 - j3) + 2;
    int ij = i2 + j3;
    
    int i3;
    for (i = 2; i<i2; i++)
    {
      i3 = i + j3;
      d[i3] -= t0 * b[i-2]; // *cast*
    }
    
    //  Release carries periodically to avoid overflowing the exact integer
    //  capacity of double floating point words in D.
    if (((j - 1)%mpnpr) == 0) 
    {
      //dir$ ivdep
      for (i = j; i<ij;i++)
      {
        t1 = d[i];
        t2 = int (mprdx * t1);
        d[i] = t1 - mpbdx * t2; // *cast*
        d[i-1] += t2; // *cast*
      }
    }
    d[j-1] += mpbdx * d[j-2]; // *cast*
    d[j-2] = t0; // *cast* 
  }
  
  //  Compute additional words of the quotient, as long as the remainder
  //  is nonzero.
  
  
  bool stopped=false;
  for (j = na + 2; j<=mpnw + 3; j++)
  {
    t1 = mpbx2 * d[j-2] + mpbdx * d[j-1] + d[j];
    if (j <= mpnw + 2) t1 = t1 + mprdx * d[j+1];
    t0 = int (rb * t1);
    int j3 = j - 3;
    int i2 = MIN (nb, mpnw + 2 - j3) + 2;
    int ij = i2 + j3;
    ss = 0.0;
    
    int i3;
    for (i = 2; i<i2;i++)
    {
      i3 = i + j3;
      d[i3] -= t0 * b[i-2]; // *cast*
      ss = ss + fabs (d[i3]);
    }
    
    
    
    if (((j - 1)%mpnpr) == 0) 
    {
      //dir$ ivdep
      for (i = j; i<ij; i++)
      {
        t1 = d[i];
        t2 = int (mprdx * t1);
        d[i] = t1 - mpbdx * t2; // *cast*
        d[i-1] += t2; // *cast*
      }
    }
    d[j-1] += mpbdx * d[j-2]; // *cast*
    d[j-2] = t0; // *cast*
    if (ss == 0.0)
    {
      stopped=true;
      break;
    }
    if (ij <= mpnw + 1) d[ij+2] = 0.0;
  }
  
  //  Set SIGN and exponent, and fix up result.
  
  if(!stopped)
    j = mpnw + 3;
  
  d[j-1] = 0.0;
  int is;
  if (d[0] == 0.0) 
    is=1;
  else 
    is=2;
  
  int nc = MIN (j - 1, mpnw);
  d[nc+2] = 0.0;
  d[nc+3] = 0.0;
  
  for(i=j; i>=2; i--)
    d[i] =  d[i-is];
  
  d[0] = SIGN(nc, a.sign*b.sign);
  d[1] = a.exponent - b.exponent + is - 2;  
  
  mpnorm (d, c, mpnw);
  delete [] d;
  
#if DEBUG == 2
  if(MPIDB >= 8)
    debug("MPDIV O ", c);
#endif
}

void mp::mpdivd(const mp& a, const mp_dpe& b, 
                mp& c, int mpnw)
{
  int j,k;
  double bb, br, dd, t1;
  mp f(mpsize(6));
  
  // has nothing to  do anything with the subscript
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(c);
    return;
  }
#if DEBUG == 2
  if (MPIDB >= 9) 
  {
    debug("MPDIVD I ", a);
    cerr << "MIDIVD I " <<  b.a << endl;
    cerr.width(25);
    cerr.precision(15);
    cerr << b.n << endl;
    cerr.width(0);
    cerr.precision(0);
  }
#endif
  
  int na = MIN (a.nw, mpnw);
  int ib = int(SIGN (1.0, b.a)); // *cast*
  
  //  Check if dividend is zero.
  
  if (na == 0) 
  {
    zero(c);
#if DEBUG == 2
    if (MPIDB >= 9) 
      debug("MPDIVD O ", c);
#endif
    return;
  }
  
  //  Check if divisor is zero.
  
  if (b.a == 0.0) 
  {
    if (MPKER[32] != 0) 
    {
      cerr << "*** MPDIVD: Divisor is zero.\b.n";     
      MPIER = 32;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  int n1 = b.n / mpnbt;
  int n2 = b.n - mpnbt * n1;
  bb = fabs (b.a) * pow(2.0,n2);
  
  //  Reduce BB to within 1 and MPBDX.
  if (bb >= mpbdx) 
  {
    
    for (k = 1; k<=100;k++)
    {
      bb = mprdx * bb;
      if (bb < mpbdx) 
      {
        n1 += k;
        break;
      }
    }
  }
  else if (bb < 1.0) 
  {
    for (k = 1; k<=100; k++)
    {
      bb = mpbdx * bb;
      if (bb >= 1.0)
      {
        n1 -= k;
        break;
      }
    }
  }
  
  //  If B cannot be represented exactly in a single mantissa word, use MPDIV.
  
  if (bb != int (bb)) 
  {
    mp_dpe dpb(SIGN(bb, b.a), n1 * mpnbt);
    mpdmc (dpb , f);
    mpdiv (a, f, c, mpnw);
#if DEBUG == 2
    if (MPIDB >= 9) 
      debug("MPDIVD O ", c);
#endif
    return;
  }
  
  double* d = new double[mpnw+4];  
  
  br = 1.0 / bb;
  dd = a[0];
  
  //  Perform short division (not vectorizable at present).  Continue as long as
  //  the remainder remains nonzero.
  
  bool skipJ=false;
  
  for(j = 2; j<= mpnw + 3;j++)
  {
    t1 = int (br * dd);
    d[j] = t1; 
    dd = mpbdx * (dd - t1 * bb);
    if (j <= na) 
      dd = dd + a[j-1];
    
    else 
      if (dd == 0.0)
      {
        skipJ=true;
        break;
      }
  }
  //  Set SIGN and exponent of result.
  
  if(!skipJ)
    j = mpnw + 3;
  
  int nc = MIN (j - 1, mpnw);
  d[0] = SIGN(nc, a.sign*ib);
  d[1] = a.exponent - n1;
  
  if (j <= mpnw + 2) d[j+1] = 0.0;
  if (j <= mpnw + 1) d[j+2] = 0.0;
  
  mpnorm (d, c, mpnw);
  
  delete [] d;
#if DEBUG == 2
  if (MPIDB >= 9) 
    debug("MPDIVD O ", c);
#endif
}

void mp::mpdmc(const mp_dpe& a, mp& b)
{
  int i,k;
  double aa;
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 9) 
  {
    cerr.width(25);
    cerr.precision(15);
    cerr << "MPDMC I " << a.a << '\t';
    cerr.width(0);
    cerr.precision(0);
    cerr << a.n << endl;
  }
#endif
  
  //  Check for zero.
  if (a.a == 0.0) 
  {
    zero(b);
    
#if DEBUG == 2
    if (MPIDB >= 9) 
      debug("MDMC O ", b);
#endif
    return;
  }
  
  int n1 = a.n / mpnbt;
  int n2 = a.n - mpnbt * n1;
  aa = fabs (a.a) * pow(2.0, n2);
  
  //  Reduce AA to within 1 and MPBDX.
  if (aa >= mpbdx) 
  {
    for(k = 1; k<=100; k++)
    {
      aa = mprdx * aa;
      if (aa < mpbdx) 
      {
        n1 = n1 + k;
        break;
      }
    }
  }
  else if (aa < 1.0) 
  {
    for (k = 1; k<=100; k++)
    {
      aa = mpbdx * aa;
      if (aa >= 1.0) 
      {
        n1 = n1 - k;
        break;
      }
    } 
  }
  
  //  Store successive sections of AA into B.
  b.exponent=n1;
  b[0] = float(int (aa));  // *cast* float
  aa = mpbdx * (aa - b[0]);
  b[1] = float(int (aa)); // *cast* float
  aa = mpbdx * (aa - b[1]);
  b[2] = float(int (aa)); // *cast* float
  aa = mpbdx * (aa - b[2]);
  b[3] = float(int (aa)); // *cast* float
  b[4] = 0.;
  b[5] = 0.;
  
  for (i = 3; i>=0; i--)
    if (b[i] != 0.) break;
    
    aa = i + 1;
    b.sign=int(SIGN(1, a.a)); // *cast*
    b.nw=int(aa); // *cast*
    
#if DEBUG == 2
    if (MPIDB >= 9) 
      debug("MDMC O ", b);  
#endif
}

void mp::mpeq(const mp& a, mp& b, int mpnw)
{
  int i;
  if (MPIER != 0)
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 10) 
    cerr << "MPEQ\n";
#endif 
  
  int na = MIN (a.nw, mpnw);
  if (na == 0) 
  {
    zero(b);
    return;
  }
  
  b.nw=na;
  b.sign=a.sign;
  b.exponent=a.exponent;
  
  for (i = 0; i<=na;i++)
    b[i] = a[i];
}

void mp::mpinfr (const mp& a, mp& b, mp& c, int mpnw)
{
  int i;
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    zero(c);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 9) 
    debug("MPINFR I ", a);
#endif
  
  //  Check if  A  is zero.
  int na = MIN (a.nw, mpnw);
  int ma = a.exponent;
  if (na == 0)  
  {
    zero(b); zero(c);
#if DEBUG == 2
    if (MPIDB >= 9)  
    {
      debug("MPINFR O ", b);
      debug("MPINFR O ", c);  
    }
#endif
  }
  
  if (ma >= mpnw - 1) 
  {
    if (MPKER[40] != 0) 
    {
      cerr << "*** MPINFR: Argument is too large.\n";
      MPIER = 40;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  //  Place integer part in  B.
  int nb = MIN (MAX (ma + 1, 0), na);
  if (nb == 0) 
    zero(b); 
  else 
  {
    b.nw=nb; 
    b.sign=a.sign; 
    b.exponent=ma;
    b[nb] = 0.;
    b[nb+1] = 0.;
    
    for(i = 0; i<nb; i++)
      b[i] = a[i];
  }
  
  //  Place fractional part in c.
  int nc = na - nb;
  if (nc <= 0) 
    zero(c);
  else 
  {
    c.nw=nc;
    c.sign=a.sign;
    c.exponent=ma-nb;
    c[nc] = 0.;
    c[nc+1] = 0.;
    
    for(i = 0; i<nc; i++)
      c[i] = a[i+nb];   
  }
  
  //  Fix up results.  B may have trailing zeros and c may have leading zeros.
  mproun (b, mpnw);
  mproun (c, mpnw);
  
#if DEBUG == 2
  if (MPIDB >= 9)  
  {
    debug("MPINFR O ", b);
    debug("MPINFR O ", c);  
  }
#endif
}

void mp::mpmdc(const mp& a, mp_dpe& b)
{
  double aa;
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    b.a = 0.0;
    b.n = 0;
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 9) 
    debug("MPMDC I ", a);
#endif
  
  bool isAZero = false;
  if (a.nw==0)  
  {
    b.a = 0.0;
    b.n = 0;
    isAZero=true;
  }
  
  if(!isAZero)
  {
    int na = a.nw;
    aa = a[0];
    if (na >= 2) aa += mprdx * a[1];
    if (na >= 3) aa += mprx2 * a[2];
    if (na >= 4) aa += mprdx * mprx2 * a[3];
    
    b.n = mpnbt * a.exponent;
    b.a = SIGN (aa, double (a.sign));
  }
  
#if DEBUG == 2
  if (MPIDB >= 9) 
  {
    cerr.width(10);
    cerr.precision(1);
    cerr << "MPMDC O "  << b.a;
    cerr.width(0);
    cerr.precision(0);
    cerr << '\t' << b.n << endl;
  }
#endif
}

void mp::mpmul(const mp& a, const mp&b, mp& c, int mpnw)
{
  int i, j;
  double t1, t2;
  
  if (MPIER != 0)
  {
    if (MPIER == 99)  mpabrt();
    zero(c);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 8) 
  {
    debug("MPMUL I ", a);
    debug("MPMUL I ", b);
  }
#endif
  
  int na = MIN (a.nw, mpnw);
  int nb = MIN (b.nw, mpnw);
  
  if (na == 0 || nb == 0) 
  {
    //  One of the inputs is zero -- result is zero.
    zero(c);
#if DEBUG == 2
    if (MPIDB >= 8) 
      debug("MPMUL O ", c);
#endif
    return;
  }
  if (na == 1 && a[0] == 1.)
  {
    //  A is 1 or -1 -- result is B or -B.
    c.sign=a.sign*b.sign;
    c.nw=nb;
    c.exponent=a.exponent+b.exponent;
    
    for(i = 0; i<nb;i++)
      c[i] = b[i];
    
#if DEBUG == 2
    if (MPIDB >= 8) 
      debug("MPMUL O ", c);
#endif
    
    return;
  }
  else if (nb == 1 && b[0] == 1.) 
  {
    
    //  B is 1 or -1 -- result is A or -A.
    c.sign = a.sign*b.sign;
    c.nw = na;
    c.exponent = a.exponent + b.exponent;
    
    for (i = 0; i<na; i++)
      c[i] = a[i];
    
#if DEBUG == 2
    if (MPIDB >= 8) 
      debug("MPMUL O ", c);
#endif
    return;
  }
  
  double* d = new double[mpnw+4];
  
  int nc = MIN (na + nb, mpnw);
  double d2 = a.exponent + b.exponent; // *cast*
  
  for(i = 0; i<nc + 4; i++)
    d[i] = 0.0;
  
  //  Perform ordinary long multiplication algorithm.  Accumulate at most MPNW+4
  //  mantissa words of the product.
  
  for (j = 3; j<= na + 2; j++)
  {
    t1 = a[j-3];
    int j3 = j - 3;
    int n2 = MIN (nb + 2, mpnw + 4 - j3);
    
    for(i = 2; i<n2; i++) 
      d[i+j3] += t1 * b[i-2];
    
    //  Release carries periodically to avoid overflowing the exact integer
    //  capacity of double floating point words in D.
    
    if (((j - 2)% mpnpr) == 0) 
    {
      int i1 = MAX (3, j - mpnpr);
      int i2 = n2 + j3;
      
      //dir$ ivdep
      for(i = i1-1; i< i2; i++)
      {
        t1 = d[i];
        t2 = int (mprdx * t1);
        d[i] = t1 - mpbdx * t2;
        d[i-1] +=  t2;
      }
    }
  }
  
  //  If D(1) is nonzero, shift the result one cell right.
  
  if (d[1] != 0.0) 
  {
    d2 += 1.0;
    
    //dir$ ivdep
    for (i = nc + 3; i>=2; i--)
      d[i] = d[i-1];
    
  }
  d[0] = SIGN (nc, a.sign*b.sign);
  d[1] = d2;
  
  //  Fix up result, since some words may be negative or exceed MPBDX.
  
  mpnorm (d, c, mpnw);
  delete [] d;
  
#if DEBUG == 2
  if (MPIDB >= 8) 
    debug("MPMUL O ", c);
#endif
}

void mp::mpmuld (const mp& a, const mp_dpe& b, mp& c, int mpnw)
{
  int i,k;
  double bb;
  mp f(mpsize(6));
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(c);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 9) 
  {
    debug("MPMULD I ", a);
    cerr.width(25);
    cerr.precision(15);
    cerr << "MPMULD I " << b.a;
    cerr.width(0);
    cerr.precision(0);
    cerr << '\t' << b.n << endl; 
  }
#endif
  
  //  Check for zero inputs.
  
  int na = MIN (a.nw, mpnw);
  int ib = int(SIGN(1.0, b.a)); // *cast*
  if (na == 0 || b.a == 0.0) 
  {
    zero(c);
#if DEBUG == 2
    if (MPIDB >= 9) 
      debug("MPMULD O ", c);
#endif
    return;
  }
  int n1 = b.n / mpnbt;
  int n2 = b.n - mpnbt * n1;
  bb = fabs (b.a) * pow(2.0, n2);
  
  //  Reduce BB to within 1 and MPBDX.
  
  if (bb >= mpbdx) 
  {
    
    for (k = 1; k<=100; k++)
    {
      bb = mprdx * bb;
      if (bb < mpbdx) 
      {
        n1 = n1 + k;
        break;
      }
    }
  }
  else if (bb < 1.0) 
  {
    for (k = 1; k<= 100; k++)
    {
      bb = mpbdx * bb;
      if (bb >= 1.0) 
      {
        n1 = n1 - k;
        break;
      }
    }
  }
  
  //  If B cannot be represented exactly in a single mantissa word, use MPMUL.
  
  if (bb != int (bb)) 
  {
    mp_dpe dpb(SIGN (bb, b.a), n1 * mpnbt); 
    mpdmc (dpb, f);
    mpmul (f, a, c, mpnw);
#if DEBUG == 2
    if (MPIDB >= 9) 
      debug("MPMULD O ", c);
#endif
    return; 
  }
  
  double* d = new double[mpnw+4];
  
  //  Perform short multiply operation.
  
  //dir$ ivdep
  for(i = 2; i<na + 2; i++)
    d[i] = bb * a[i-2];
  
  //  Set the exponent and fix up the result.
  
  d[0] = SIGN (na, a.sign * ib);
  d[1] = a.exponent + n1;
  d[na+2] = 0.0;
  d[na+3] = 0.0;
  
  mpnorm (d, c, mpnw);
  
  delete [] d;
#if DEBUG == 2
  if (MPIDB >= 9) 
    debug("MPMULD O ", c);
#endif  
}

void mp::mpnint (const mp& a, mp& b, int mpnw)
{
  int i;
  mp f(mpsize(6)), s(mpsize(mpnw+2));
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 8) 
    debug("MPNINT I ", a);
#endif
  
  int na = MIN (a.nw, mpnw);
  if (na == 0)  
  {
    //  A is zero -- result is zero.
    zero(b);
#if DEBUG == 2
    if (MPIDB >= 8) 
      debug("MPINT O ", b);
#endif
    return;
  }
  if (a.exponent >= mpnw) 
  {
    //  A cannot be represented exactly as an integer.
    if (MPKER[56] != 0) 
    {
      cerr << "*** MPNINT: Argument is too large.\t" << a.exponent << '\t' << mpnw << endl;
      MPIER = 56;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  f.nw=1;
  f.sign=1;
  f.exponent=-1;
  f[0]=float(0.5 * mpbdx); // *cast*
  f[1]=0.;
  
  //  Add or subtract 1/2 from the input, depending on its SIGN.
  if (a.sign == 1) 
    mpadd (a, f, s, mpnw);
  else 
    mpsub (a, f, s, mpnw);
  
  int ic = s.sign; 
  int nc = s.nw;
  int mc = s.exponent;
  
  //  Place integer part of S in B.
  
  int nb = MIN (MAX (mc + 1, 0), nc);
  if (nb == 0) 
    zero(b);
  else 
  {
    b.nw=nb;
    b.sign=ic;
    b.exponent=mc;
    b[nb] = 0.;
    b[nb+1] = 0.;
    
    for(i = 0; i<nb; i++)
      b[i] = s[i];
  }
  
#if DEBUG == 2
  if (MPIDB >= 8) 
    debug("MPINT O ", b);
#endif
}

void mp::mpnorm(double d[],  mp& a, int mpnw)
{
  double t1, t2, t3;
  int i;
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(a);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 9) 
  {
    int no = MIN(int(fabs(d[0])), MPNDB) + 4; // *cast*
    cerr.width(18);
    cerr.precision(1);
    
    cerr << "MPNORM I ";
    for(i=0; i<no;i++)
    {
      cerr << d[i] << '\t';
      if (i%4 == 0)
        cerr << endl;
    }     
    // reset to default;
    cerr.width(0);
    cerr.precision(0);      
  }
#endif
  
  int ia = int(SIGN (1.0, d[0])); // sign
  int na = MIN (int (fabs (d[0])), mpnw);
  if (na == 0)  
  {
    zero(a);
#if DEBUG == 2
    if (MPIDB >= 9) 
      debug("MPNORM O ", a);
#endif
    return;
  }
  int n4 = na + 4;
  double a2 = d[1]; 
  d[1] = 0.0;
  
  bool needToNormalize=true;
  while(needToNormalize)
  {
    /****  110 ****/
    // !>
    //  Try a vectorized fixup loop three times, unless A is very short.  This
    //  should handle 99% of the inputs.  On RISC computers, it is more
    //  efficient to completely bypass this loop, by defining RISC = 1 
    
    bool breakLoop = false;
#if RISC == 0
    double s1;
    int k;
    if (na > 8) 
    { 
      for(k = 1; k<=3; k++)
      {
        s1 = 0.0;
        
        //dir$ ivdep
        for (i = 2; i<n4; i++)
        {
          t2 = mprdx * d[i];
          t1 = int (t2);
          if (t2 < 0.0 && t1 != t2) t1--;
          d[i] -= t1 * mpbdx;
          d[i-1] += t1;
          s1 += fabs (t1);
        }
        
        if (s1 == 0.0) 
        {
          breakLoop=true;
          break;
        }
      }
    }
#endif
    
    
    //  Still not fixed - use recursive loop.  This loop is not vectorizable,
    //  but it is guaranteed to complete the job in one pass.
    if(!breakLoop) //here that changes
    { 
      t1=0;
      for(i = n4-1; i>=2; i--)
      {
        t3 = t1 + d[i];
        t2 = mprdx * (t3);
        t1 = int (t2);
        if (t2 < 0.0 && t1 != t2) t1--;
        d[i] = t3 - t1 * mpbdx;
      } 
      d[1] += t1;
    }
    
    if (d[1] < 0.0) 
    { 
      //  D(1) is negative -- negate all words and re-normalize.
      ia = - ia;
      d[2] += mpbdx * d[1];
      d[1] = 0.0;
      
      for(i = 1; i< n4; i++)
        d[i] = - d[i];
    }
    else if (d[1] > 0.0) 
    {
      
      //  The fixup loops above "spilled" a nonzero number into D(1).  Shift the
      //  entire number right one cell.  The exponent and length of the result
      //  are increased by one.
      
      for (i = n4-2; i>=1; i--)
        a[i-1] = float(d[i]); // *cast*
      
      na = MIN (na + 1, mpnw);
      
      a2++;
      needToNormalize=false;
    } 
    else  //here the last one
    {
      for (i = 2; i< n4; i++)
        a[i-2] = float(d[i]); // *cast*
      needToNormalize=false;
    }
  }
  
  //  Perform rounding and truncation.
  a.nw=na;
  a.sign=ia;
  a.exponent = int(a2);
  
  mproun (a, mpnw);
 
  
#if DEBUG == 2
  if (MPIDB >= 9) 
    debug("MPNORM O ", a);
#endif
}

void mp::mpnpwr (const mp& a, int n, mp& b, int mpnw)
{
  int j;
  const double cl2 = 1.4426950408889633; 
  double t1;
  mpsize mpnw3(mpnw+3);
  mp f1(mpsize(6)), sk0(mpnw3), sk1(mpnw3);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 7) 
    debug("MPNPWR I ", a);
#endif
  
  int na = MIN (a.nw, mpnw);
  if (na == 0) 
  {
    if (n >= 0) 
    {
      zero(b);
#if DEBUG == 2
      if (MPIDB >= 7) 
        debug("MPNPWR O ", b);
#endif
      return;
    } 
    else 
    {
      if (MPKER[57] != 0) 
      {
        cerr << "*** MPNPWR: Argument is zero and N is negative or zero.\n";
        MPIER = 57;
        if (MPKER[MPIER] == 2)  mpabrt();
      }
      return;
    }
  }
  
  int nws = mpnw;
  mpnw = mpnw + 1;
  int nn = abs (n);
  f1.sign=1;
  f1.nw=1;
  f1.exponent=0;
  f1[0]=1.;
  f1[1]=0.;
  
  bool skip=false;
  
  switch(nn)
  {
  case 0:
    mpeq (f1, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 7) 
      debug("MPNPWR O ", b);
#endif
    return;
    
  case 1:
    mpeq (a, b, mpnw);
    skip=true;
    break;
    
  case 2:
    mpmul (a, a, sk0, mpnw);
    mpeq (sk0, b, mpnw);
    skip=true;
    break;
  }
  
  if(!skip)
  {
    //  Determine the least integer MN such that 2 ^ MN > NN.
    t1 = nn;
    int mn = int(cl2 * log (t1) + 1.0 + mprxx); // cast
    mpeq (f1, b, mpnw);
    mpeq (a, sk0, mpnw);
    int kn = nn;
    
    //  Compute B ^ N using the binary rule for exponentiation.
    
    for (j = 1; j<= mn; j++)
    {
      int kk = kn / 2;
      if (kn != 2 * kk) 
      {
        mpmul (b, sk0, sk1, mpnw);
        mpeq (sk1, b, mpnw);
      }
      kn = kk;
      if (j < mn) 
      {
        mpmul (sk0, sk0, sk1, mpnw);
        mpeq (sk1, sk0, mpnw);
      }
    }
  }
  
  //  Compute reciprocal if N is negative.
  if (n < 0) 
  {
    mpdiv (f1, b, sk0, mpnw);
    mpeq (sk0, b, mpnw);
  }
  
  mproun (b, nws);
  
#if DEBUG == 2
  if (MPIDB >= 7) 
    debug("MPNPWR O ", b);
#endif
}

void mp::mpnrt (const mp& a, int n, mp& b, int mpnw)
{
  int k;
  double t2, tn;
  const double alt = 0.693147180559945309, cl2 = 1.4426950408889633;
  const int nit = 3, n30 = int(pow(2,30)); // *cast*
  mpsize mpnw3(mpnw+3);
  mp f1(mpsize(6)), f2(mpsize(6)), 
    sk0(mpnw3), sk1(mpnw3), sk2(mpnw3), sk3(mpnw3);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 7) 
    debug("MPNRT I ", a);
#endif  
  
  int na = MIN (a.nw, mpnw);
  
  if (na == 0) 
  {
    zero(b);
#if DEBUG == 2
    if (MPIDB >= 7) 
      debug("MPNRT O ", b);
#endif
    return;
  }
  
  if (a.sign < 0) 
  {
    if (MPKER[59] != 0) 
    {
      cerr << "*** MPNRT: Argument is negative.\n";
      MPIER = 59;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  if (n <= 0 || n > n30) 
  {
    if (MPKER[60] != 0) 
    {
      cerr << "*** MPNRT: Improper value of N: " << n << endl;
      MPIER = 60;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  //  If N = 1, 2 or 3,  MPEQ, MPSQRT or MPCBRT.  These are faster.
  switch (n)
  {
  case 1:
    mpeq (a, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 7) 
      debug("MPNRT O ", b);
#endif
    return;
  case 2:
    mpsqrt (a, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 7) 
      debug("MPNRT O ", b);
#endif
    return;
  case 3:
    mpcbrt (a, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 7) 
      debug("MPNRT O ", b);
#endif
    return;
  }
  
  int nws = mpnw;
  f1.nw=1;
  f1.sign=1;
  f1.exponent=0;
  f1[0]=1;
  f1[1]=0;
  
  //  Determine the least integer MQ such that 2 ^ MQ >= MPNW.
  mp_dpe t1;
  t1.a = mpnw;
  int mq = int(cl2 * log (t1.a) + 1.0 - mprxx); // *cast*
  
  //  Check how close A is to 1.
  mpsub (a, f1, sk0, mpnw);
  if (sk0.nw == 0) 
  {
    mpeq (f1, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 7) 
      debug("MPNRT O ", b);
#endif
    return;
  }
  
  mpmdc (sk0, t1);
  int n2 = int(cl2 * log (fabs (t1.a))); // *cast*
  t1.a = t1.a * pow(0.5, n2);
  t1.n = t1.n + n2;
  if (t1.n <= -30) 
  {
    t2 = n;
    n2 = int(cl2 * log (t2) + 1.0 + mprxx); // *cast*
    int n3 = - mpnbt * mpnw / t1.n;
    if (n3 < 1.25 * n2) 
    {
      //  A is so close to 1 that it is cheaper to use the binomial series.
      mpnw = mpnw + 1;
      mpdivd (sk0, mp_dpe(t2), sk1, mpnw);
      mpadd (f1, sk1, sk2, mpnw);
      k = 0;
      
      int temp = t1.n;
      t1.n=0;
      do 
      {
        k = k + 1;
        t1.a = 1 - k * n;
        t2 = (k + 1) * n;
        mpmuld (sk1, t1, sk3, mpnw);
        mpdivd (sk3, mp_dpe(t2), sk1, mpnw);
        mpmul (sk0, sk1, sk3, mpnw);
        mpeq (sk3, sk1, mpnw);
        mpadd (sk1, sk2, sk3, mpnw);
        mpeq (sk3, sk2, mpnw);
      }
      while (sk1.nw != 0 && sk1.exponent >= - mpnw); 
      
      t1.n = temp;
      mpeq (sk2, b, mpnw);
      mpdiv (f1, sk2, sk0, mpnw);
      mproun (b, nws);
      
#if DEBUG == 2
      if (MPIDB >= 7) 
        debug("MPNRT O ", b);
#endif
      return;
    }
  }
  
  //  Compute the initial approximation of A ^ (-1/N).
  tn = n;
  mp_dpe dpn(n, 0);
  mpmdc (a, t1);
  mp_dpe dp1;
  dp1.n = int(- t1.n / tn); // *cast*
  dp1.a = exp (-1.0 / tn * (log (t1.a) + (t1.n + tn * dp1.n) * alt));
  mpdmc (dp1, b);
  mpdmc (dpn, f2);
  mpnw = 3;
  int iq = 0;
  
  //  Perform the Newton-Raphson iteration described above with a dynamically
  //  changing precision level MPNW (one greater than powers of two).
  for (k = 2; k<= mq;k++)
  {
    mpnw = MIN (2 * mpnw - 2, nws) + 1;
    bool loop=true;
    while (loop)
    {
      mpnpwr (b, n, sk0, mpnw);
      mpmul (a, sk0, sk1, mpnw);
      mpsub (f1, sk1, sk0, mpnw);
      mpmul (b, sk0, sk1, mpnw);
      mpdivd (sk1, mp_dpe(tn), sk0, mpnw);
      mpadd (b, sk0, sk1, mpnw);
      mpeq (sk1, b, mpnw);
      if (k == mq - nit && iq == 0) 
        iq = 1;
      else loop=false;
    }
  }
  //  Take the reciprocal to give final result.
  mpdiv (f1, b, sk1, mpnw);
  mpeq (sk1, b, mpnw);
  
  mproun (b, nws);
  
#if DEBUG == 2
  if (MPIDB >= 7) 
    debug("MPNRT O ", b);
#endif
}

void mp::mprand (mp& a, int mpnw)
{
  int i;
  const double f7 = 78125, s0 = 314159265;
  double t1, t2;
  static double t30, r30=0, sd = s0;
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(a);
    return;
  }
  MUTEX_LOCK();
  if(r30 == 0) 
  {
    r30 = 1.0;
    t30 = 1.0;
    
    for(i = 1; i<=30; i++)
    {
      r30 = 0.50 * r30;
      t30 = 2.0 * t30;
    }
  }
  
  a.nw=mpnw;
  a.sign=1;
  a.exponent=-1;
  
  for(i = 0; i<=mpnw + 1; i++)
  {
    t1 = f7 * sd;
    t2 = int (r30 * t1);
    sd = t1 - t30 * t2;
    a[i] = float(int (mpbdx * r30 * sd)); // *cast*
  }
  MUTEX_UNLOCK();
  mproun (a, mpnw);
  
#if DEBUG == 2
  if (MPIDB >= 9) 
    debug("MPRAND O ", a);
#endif
}

void mp::mproun(mp& a, int mpnw)
{
  const double amx = 2.e6;
  int i;
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(a);
    return;
  }
  
  //  Check for initial zeroes.
  
  int a2 = a.exponent;
  a.exponent = 0;
  int na = MIN (a.nw, mpnw);
  int n1 = na + 1;
  if (a[0] == 0) 
  {
    
    //  Find the first nonzero word and shift the entire number left.  The length
    //  of the result is reduced by the length of the shift.
    
    bool allZero=true;
    for (i = 1; i<=n1; i++)
    {
      if (a[i] != 0) 
      {
        allZero=false;
        break;          // bug fixed here
      }
    }
    
    if(allZero)
    {
      zero(a);
      return;
    }
    
    int k = i; // bug fixed here
    
    //dir$ ivdep
    for (i = 0; i<= n1 - k; i++)
      a[i] = a[i+k];
    
    
    a2 -= k;
    na -= MAX (k - 2, 0);
  }
  
  //  Perform rounding depending on MPIRD.
  
  if (na == mpnw && MPIRD >= 1) 
  {
    if ((MPIRD == 1) && (a[na] >= 0.5 * mpbdx) || 
      (MPIRD == 2) && (a[na] >= 1))
      a[na-1]++;
    
    
    //  Release carries as far as necessary due to rounding.
    bool loopBreak = false;
    for(i = na - 1; i>=0; i--) // bug fixed here
    {
      if (a[i] < mpbdx) 
      {
        loopBreak=true;
        break;
      }
      a[i] -= float(mpbdx);  // *cast*
      if(i != 0)
        a[i-1]++;
      else
     	a.exponent++;
    }
    
    //  Release of carries due to rounding continued all the way to the start --
    //  i.e. number was entirely 9's.
    if(!loopBreak)
    {
      a[0] = float(a.exponent);//a[0] = float(a.exponent + 1); // *cast*
      na = 1;
      a2++;
    }
  }
  
  if (a[na-1] == 0) 
  {
    
    //  At least the last mantissa word is zero.  Find the last nonzero word
    //  and adjust the length of the result accordingly.
    
    bool allZero=true;
    for (i = na - 1; i>=0; i--) // bug fixed here
    {
      if (a[i] != 0)
      {
        allZero=false;
        break;
      }
    } 
    if(allZero)
    {
      zero(a);
      return;
    }
    na = i + 1; // bug fixed here
  }
  
  
  //  Check for overflow and underflow.
  
  if (a2 < - amx) 
  {
    if (MPKER[68] != 0) 
    {
      cerr << "*** MPROUN: Exponent underflow.\n";
      MPIER = 68;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
  } 
  else if (a2 > amx) 
  {
    if (MPKER[69] != 0) 
    {
      cerr << "*** MPROUN: Exponent overflow.\n";
      MPIER = 69;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
  }
  
  //  Check for zero.
  
  if (a[0] == 0.) 
    zero(a);
  else 
  {
    a.nw=na;
    a.exponent=a2;
    a[na] = 0.;
    a[na+1] = 0.;
  }
}

void mp::mpsqrt(const mp& a, mp& b, int mpnw)
{
  int k;
  const double cl2 = 1.4426950408889633;
  const int nit = 3;
  
  double t2;
  mpsize mpnw3(mpnw+3);
  mp f(mpsize(6)), sk0(mpnw3), sk1(mpnw3), sk2(mpnw3);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2 
  if (MPIDB >= 7) 
    debug("MPSQRT I ", a);
#endif
  
  int na = MIN (a.nw, mpnw);
  
  if (na == 0) 
  {
    zero(b);
#if DEBUG == 2
    if (MPIDB >= 7) 
      debug("MPSQRT O ", b);
#endif
    return;
  }
  
  if (a.sign < 0.0) 
  {
    if (MPKER[70] != 0) 
    {
      cerr << "*** MPSQRT: Argument is negative.\n";
      MPIER = 70;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  int nws = mpnw;
  
  //  Determine the least integer MQ such that 2 ^ MQ >= MPNW.
  mp_dpe t1(mpnw,0);
  int mq = int(cl2 * log (t1.a) + 1.0 - mprxx); // *cast* 
  
  //  Compute the initial approximation of 1 / Sqrt(A).
  int iq;
  
  mpmdc (a, t1);
  mp_dpe dp1;
  
  dp1.n = - t1.n / 2;
  t2 = sqrt (t1.a * pow(2.0, (t1.n + 2 * dp1.n)));
  dp1.a = 1.0/t2; 
  mpdmc (dp1, b);
  
  f.nw=1;
  f.sign=1;
  f.exponent=0;
  f[0]=1.;
  f[1]=0.;
  mpnw = 3;
  iq = 0;
  
  //  Perform the Newton-Raphson iteration described above with a dynamically
  //  changing precision level MPNW (one greater than powers of two). 
  int nw1, nw2;
  for(k = 2; k<= mq - 1; k++)
  {
    nw1 = mpnw;
    mpnw = MIN (2 * mpnw - 2, nws) + 1;
    nw2 = mpnw;
    bool stop=false;
    while(!stop)
    {
      mpmul (b, b, sk0, mpnw);
      mpmul (a, sk0, sk1, mpnw);
      mpsub (f, sk1, sk0, mpnw);
      mpnw = nw1;
      mpmul (b, sk0, sk1, mpnw);
      mpmuld (sk1, mp_dpe(0.50), sk0, mpnw);
      mpnw = nw2;
      mpadd (b, sk0, sk1, mpnw);
      mpeq (sk1, b, mpnw);
      if (k == mq - nit && iq == 0) 
        iq = 1;
      else
        stop=true;
    }
  }
  
  //  Perform last iteration using Karp's trick.  
  mpmul (a, b, sk0, mpnw);
  nw1 = mpnw;
  mpnw = MIN (2 * mpnw - 2, nws) + 1;
  nw2 = mpnw;
  mpmul (sk0, sk0, sk1, mpnw);
  mpsub (a, sk1, sk2, mpnw);
  mpnw = nw1;
  mpmul (sk2, b, sk1, mpnw);
  mpmuld (sk1, mp_dpe(0.50), sk2, mpnw);
  mpnw = nw2;
  mpadd (sk0, sk2, sk1, mpnw);
  mpeq (sk1, b, mpnw);
  
  //  Restore original precision level.
  mproun (b, nws);
  
#if DEBUG == 2
  if (MPIDB >= 7) 
    debug("MPSQRT O ", b);
#endif
  
}

void mp::mpsub (const mp& a, const mp& b, mp& c, int mpnw)
{
  
  int i;
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(c);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 9) 
    cerr << "MPSUB\n";
#endif
  
  //  Check if A = B.  This is necessary because A and B might be same array,
  //  in which case negating B below won't work.
  
  // check if A == B points to the same object 
  if(&a == &b) 
  {
    zero(c);
#if DEBUG == 2
    if(MPIDB >= 9)
      debug("MPSUB O ", c);
#endif
    return;
  }
  
  // check if their mantissas are the same
  if(a.sign == b.sign && a.nw == b.nw && a.exponent == b.exponent)
  {
    for(i=0; i<a.nw;i++)
    {
      if(a[i] != b[i])break;
    }
    if (i==a.nw)
    {
      zero(c);
#if DEBUG == 2
      if(MPIDB >= 9)
        debug("MPSUB O ", c);
#endif
      return;
    }
  }
  
  //  Make bb shares b's mantissa and negate bb's sign.
  mp bb(mpsize(0));
  bb.sign=-b.sign;
  bb.nw=b.nw;
  bb.exponent=b.exponent;
  bb.mantissa=b.mantissa;
  
  mpadd (a, bb, c, mpnw);
  bb.mantissa = NULL;
}


