#include "mp.h"
#include <string.h>
#include <ctype.h>

#if DEBUG == 2
void mp::mpdeb (char* cs, const mp& a, int mpnw)
{
  char b[160];
	
  if (MPIER != 0) 
    return;
  
  int n;
  int ids = MPIDB;
  MPIDB = 0;
  mpnw = MIN (mpnw, 10);
  mpoutc (a, b, n, mpnw);
  n = MIN (n, 70);
  cerr << cs << ' ';
  for(int i=0; i<n; i++) 
    cerr << b[i];
  
  cerr << endl;
  
  MPIDB = ids;
}
#endif

void mp::mpinpc (const char a[], int n, mp& b, int mpnw)
{ 
  int i, j, is, id; 
  double bi;
  char ai;
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2
	if (MPIDB >= 7) 
	{
		int no = MIN (n, int (7.225 * MPNDB) + 20);
		cerr << "MPINPC I ";
		for(i=0; i<no;i++)
		{
			cerr <<  a[i];
			if (i%78 == 0)
				cerr << endl;
		} 
	}
#endif
  
  char *ca = new char[81];
  
  mpsize mpnw3(mpnw+3);
  mp f(mpsize(6)), sk0(mpnw3), sk1(mpnw3), sk2(mpnw3);
  int nws = mpnw++;
  int i1 = 0;
  int nn = 0;
  
  //  Find the carat, period, plus or minus sign, whichever comes first.
  
  bool caretFound=false;
  
  for (i = 0; i<n; i++)
  {
    ai = a[i];
    if (ai == '^')
    {
      caretFound=true;
      break;
    }
    if (ai == '.' || ai == '+' || ai == '-') 
      break;
  }
  
  for(j=0;j<81;j++)
		ca[j]='\0';
	
	
  if (caretFound)
  {
    //  Make sure number preceding the caret is 10.
    int i2 = i-1;
    if (i2 > 79)
    {
      delete [] ca;
      mpinpcExit();
      return;
    }
		
    j=0;
    for(i = 0; i<=i2; i++)
    {
      ai = a[i];
      if (ai == ' ')
        continue;
      else if (!isdigit(ai))
      {
        delete [] ca;
        mpinpcExit();
        return;
      }
      ca[j++] = ai;
    }
		
    if (ca[0]!='1' || ca[1]!='0')
    {
      delete [] ca;
      mpinpcExit();
      return;
    }
    i1 = i2 + 2;
    
		
    bool exit=true;
    //  Find the x or *.
    for(i = i1; i< n; i++)
    {
      ai = a[i];
      if (ai == 'x' || ai == '*')
      {
        exit=false;
        break;
      }
    }
    if(exit)
    {
      delete [] ca;
      mpinpcExit();
      return;
    }
		
		
    //  Convert the exponent.
    
    i2 = i - 1;
    int l1 = i2 - i1;
    if (l1 > 79) 
    {
      delete [] ca;
      mpinpcExit();
      return; 
    }
		
		
    id = 0;
    is = 1;
    
    j=0;
    for (i = 0; i<=l1; i++)
    {
      ai = a[i+i1];
      if (ai == ' ' || ai == '+') 
        continue;
      else if (ai == '-' && id == 0) 
      {
        id = 1;
        is = -1;
      } 
      else 
      {
        if (!isdigit(ai))
        {
          delete [] ca;
          mpinpcExit();
          return;
        }
        id = 1; 
        ca[j++] = ai;
      }
    }
		
    ca[j]='\0';
    nn=atoi(ca);
    
    nn = is * nn;
    i1 = i2 + 2;
  }
  //  Find the next nonblank character.
  bool exit=true;
  
  for(i = i1; i< n; i++)
  {
    if (a[i] != ' ')
    {
      exit=false;
      break;
    }
  }
  
  if(exit)
  {
    delete [] ca;
    mpinpcExit();
    return;
  }
  
  //  Check if the nonblank character is a plus or minus SIGN.
  
  i1 = i;
  if (a[i1] == '+') 
  {
    i1 = i1 + 1;
    is = 1;
  } 
  else if (a[i1] == '-') 
  {
    i1 = i1 + 1;
    is = -1;
  } 
  else 
    is = 1;
  
  int nb = 0;
  int ib = 0;
  id = 0;
  int ip = 0;
  zero(sk2);
  f.nw=1;
  f.sign=1;
  f.exponent=0;
  int it = 0;
  
  int mm;
  bool cont=true;
  while(cont)
  {
    ip = 0;
    
    for(mm=0; mm<6; mm++)
      ca[mm]='0';
		
    //  Scan for digits, looking for the period also.  On the first pass we just
    //  count, so that on the second pass it will come out right.
    
    for(i = i1; i< n; i++)
    {
      ai = a[i];
      if (ai == ' ') 
      {
      } 
      else if (ai == '.') 
      {
        if (ip != 0)
        {
          delete [] ca;
          mpinpcExit();
          return;
        }
        ip = id;
      } 
      else if (!isdigit(ai))
      {
        delete [] ca;
        mpinpcExit();
        return;
      } 
      else 
      {
        id++;
        ca[ib++] = ai;
      }
      if (ib == 6 || i == (n-1) && ib != 0) 
      {
        if (it != 0) 
        {
          nb++;
          ca[ib]='\0';
          
          bi=atoi(ca);
					
          mpmuld (sk2, mp_dpe(1e6), sk0, mpnw);
          
          if (bi != 0) 
          {
            f.nw=1; f.sign=1;
            f[0]=float(bi); // *cast*
          } 
          else 
          {
            f.nw=0; f.sign=1;
          }
          mpadd (sk0, f, sk2, mpnw);
          for(mm=0; mm<6; mm++)
            ca[mm]='0';
        }
        if ((i+1) != n) ib = 0;
      }
    }
    
    if (it == 0) 
    {
      ib = 6 - ib;
      if (ib == 6) 
        ib = 0;
      it = 1;
    }
    else
      cont=false;
  }
  
	
  if (is == -1) 
    sk2.sign = - sk2.sign;
  
  if (ip == 0) 
    ip = id;
  
  nn = nn + ip - id;
  f.nw=1; f.sign=1;
  f[0]=10;
  mpnpwr (f, nn, sk0, mpnw);
  mpmul (sk2, sk0, sk1, mpnw);
  mpeq (sk1, b, mpnw);
  mproun (b, nws);
  
  delete [] ca;
#if DEBUG == 2
	if (MPIDB >= 7) 
		debug("MPINPC O ", b);
#endif
}

void mp::mpinpcExit()
{
  if (MPKER[41] != 0) 
  {
    cerr << "*** MPINPC: Syntax error in literal string.\n";
    MPIER = 41;
    if (MPKER[MPIER] == 2)  mpabrt();
  }
}


void mp::mpoutc(const mp& a, char* b, int& n, int mpnw)
{
  int i, j, nn;
  const double al2 = 0.301029995663981195;
  double aa, t1;
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    b[0] = '\0';
    n = 0;
    return;
  }
	
#if DEBUG == 2
  int no;
  if (MPIDB >= 7) 
    debug("MPOUTC I ", a);
#endif
	
  mpsize mpnw3(mpnw+3);
  mp f(6), sk0(mpnw3), sk1(mpnw3);
  char* ca = new char[16+1];
  
  int na = MIN (a.nw, mpnw);
  mpnw = mpnw + 1;
  f.sign=1; f.nw=1; f.exponent=0;
  f[0]=10.;
  
  //  Determine exact power of ten for exponent.
  int nx;
  if (na != 0) 
  {
    aa = a[0];
    if (na >= 2) aa = aa + mprdx * a[1];
    if (na >= 3) aa = aa + mprx2 * a[2];
    if (na >= 4) aa = aa + mprdx * mprx2 * a[3];
    t1 = al2 * mpnbt * a.exponent + log10 (aa);
    if (t1 >= 0.0) 
      nx = int(t1); // *cast*
    else 
      nx = int(t1 - 1.0); // *cast*
    
    mpnpwr (f, nx, sk0, mpnw);
    mpdiv (a, sk0, sk1, mpnw);
		
    //  If we didn't quite get it exactly right, multiply or divide by 10 to fix.
    bool cont = true;
    while (cont) 
    {
      if (sk1.exponent < 0)
      {
        nx = nx - 1;
        mpmuld (sk1, mp_dpe(10.0), sk0, mpnw);
        mpeq (sk0, sk1, mpnw);
      }
      else if(sk1[0] >= 10.) 
      {
        nx = nx + 1;
        mpdivd (sk1, mp_dpe(10.0), sk0, mpnw);
        mpeq (sk0, sk1, mpnw);
      }
      else
        cont = false;
    }
    
    sk1.sign = 1; 
  }
  else 
    nx = 0;
  
  
  //  Place exponent first instead of at the very end as in Fortran.
  
  b[0] = '1';
  b[1] = '0';
  b[2] = ' ';
  b[3] = '^';
  
  sprintf(ca, "%1d", nx);
  int len = strlen(ca);
  int blank = 14-len;
  for(i=4;i<blank;i++)
    b[i]=' ';
	
  for(i = 0;i< len; i++)
    b[blank+i] = ca[i]; 
  
  b[14] = ' ';
  b[15] = 'x';
  b[16] = ' ';
  
	
  //  Insert sign and first digit.
	
  if (a.sign == -1) 
    b[17] = '-';
  else  
    b[17] = ' ';
	
  if (na != 0) 
    nn = int(sk1[0]); // *cast*
  else 
    nn = 0;
	
  sprintf(ca, "%1d", nn);   
	
  b[18] = ca[0];
  b[19] = '.';
	
  int ix = 20;
  if (na == 0) 
  {     
    n = ix;
#if DEBUG == 2
    if (MPIDB >= 7)
    {
      no = MIN (n, 6 * MPNDB + 20);
      cerr << "MPOUTC O "; 
      for(i=0; i<no; i++)
        cerr << b[i];
    }
#endif
    delete [] ca;
    b[n]='\0';
    return;
  }
	
  f[0] = float(nn); // *cast*
  mpsub (sk1, f, sk0, mpnw);
  
  if (sk0.nw == 0) 
  {     
    n = ix;
#if DEBUG == 2
    if (MPIDB >= 7)
    {
      no = MIN (n, 6 * MPNDB + 20);
      cerr << "MPOUTC O "; 
      for(i=0; i<no; i++)
        cerr << b[i];
    }
#endif
    delete [] ca;
    b[n]='\0';
    return;
  }
  
  mpmuld (sk0, mp_dpe(1e6), sk1, mpnw);
	
  int nl = int(MAX (mpnw * log10 (mpbdx) / 6.0 - 1.0, 1.0)); // *cast*
  
  //  Insert the digits of the remaining words.
  bool skip = false;
  for (j = 1; j<=nl; j++)
  {
    if (sk1.exponent == 0.) 
    {
      nn = int(sk1[0]); // *cast*
      f.nw=1; f.sign=1; 
      f[0] = float(nn); // *cast*
    } 
    else 
    {
      f.nw=0; f.sign=1;
      nn = 0;
    }
    
    sprintf(ca, "%06d", nn);
    for (i = 0; i<6; i++)
      b[i+ix] = ca[i];
    
    ix  += 6;
    mpsub (sk1, f, sk0, mpnw);
    mpmuld (sk0, mp_dpe(1e6), sk1, mpnw);
    if (sk1.nw == 0) 
    {
      skip = true;
      break;
    }
  }
  //  Check if trailing zeroes should be trimmed.
	
  if(!skip)
    j = nl + 1;
  
  int l = --ix;
  if (b[l] == '0' || (j > nl && b[l-1] == '0' && 
    b[l-2] == '0' && b[l-3] == '0')) 
  {
    b[l] = '\0';
    bool loopbreak = false;
    for (i = l - 1; i>=20; i--)
    {
      if (b[i] != '0') 
      {
        ix = i;
        loopbreak = true;
        break;
      }
      b[i] = '\0';
    }
    if(!loopbreak)
      ix = 20;
    
    //  Check if trailing nines should be rounded up.
    
  } 
  else if (j > nl && b[l-1] == '9' && b[l-2] == '9' 
    && b[l-3] == '9') 
  {
    b[l] = '\0';
    
    skip = false;
    for (i = l - 1; i>=20; i--)
    {
      if (b[i] != '9') 
      {
        skip = true;
        break;
      }
      b[i] = '\0';
    }
    
    //  We have rounded away all digits to the right of the decimal point, and the
    //  digit to the left of the digit is a 9.  Set the digit to 1 and increase
    //  the exponent by one.
    
    if(!skip)
    {
      ix = 20;
      if (b[18] == '9') 
      {
        b[18] = '1';
        sprintf(ca, "%10d", nx+1);
        
        for (i = 0; i<10; i++)
          b[i+4] = ca[i];   
      } 
      else 
      {
        ca[0] = b[18]; ca[1]='\0';
        nn = atoi(ca);
        
        sprintf(ca, "%1d", nn+1);
        
        b[18] = ca[0];
      }
    }
    
    else
    {
      ca[0] = b[i]; ca[1]='\0';
      nn = atoi(ca);
      sprintf(ca, "%1d", nn+1);
      
      b[i] = ca[0];
      ix = i;
    }
  }
  
  n = ix;
#if DEBUG == 2
  if (MPIDB >= 7)
  {
    no = MIN (n, 6 * MPNDB + 20);
    cerr << "MPOUTC O "; 
    for(i=0; i<no; i++)
      cerr << b[i];
  }
#endif
  b[++n]='\0';
  delete [] ca;
}


void mp::mpdexc (const char a[], int l, mp& b, int mpnw)
{
  int i;
  bool foundExponent = false;
  for( i = 0; i< l; i++)
  {
    if (a[i] == 'D' || a[i] == 'E' || a[i] == 'd' ||
      a[i] == 'e') 
    {
      foundExponent = true;
      break;
    }
  }
  if(!foundExponent)
  {
    mpinpc (a, l, b, mpnw);
    return;
  }
	
  char *c = new char[mpipl+101];
  int i1 = i + 1; 
  int l1 = i; 
  int l2 = l - i1; 
  c[0] = '1';
  c[1] = '0';
  c[2] = '^';
  
  for(i = 0; i<l2; i++)
    c[i+3] = a[i+i1];
  
  c[l2+3] = 'x';
  
  for(i = 0; i<l1; i++)
    c[i+l2+4] = a[i];
  c[i+l2+4]='\0';
	
  mpinpc (c, l1 + l2 + 4, b, mpnw);
  delete [] c;
}

void mp::mpouts (const mp& a, int la, char* cs, int& l, int mpnw)
{
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    return;
  }
  int ll = int(la / log10 (mpbdx) + 2.0);
  mpnw = MIN (mpnw, ll);
	if(mpipl<10000)
		mpoutc (a, cs, l, mpnw);
	else
		mpoutx(a,cs,l,mpnw);
  l = MIN (l, la + 20) + 1;
  cs[l-1] = ',';
  cs[l]='\0';
  return;
}

