#include "mp.h"

void mp::mpagmx (mp& a, mp& b, int mpnw)
{
  mpsize mpnw2(mpnw+2);
  mp sk0(mpnw2), sk1(mpnw2);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(a); zero(b);
    return;
  }
  
  // sk0 is initialized to zero by constructor
  int l1 = 0;
  int s1;
  // 100 
  mp_dpe dpe1(0.50, 0);
  do
  { 
    l1++;
    if (l1 == 50) 
    {
      if (MPKER[5] != 0)
      {
        cerr << "*** MPAGMX: Iteration limit exceeded.\n";
        MPIER = 5;
        if (MPKER[MPIER] == 2)  mpabrt();
      }
    }
    s1 = sk0.exponent;
    mpadd (a, b, sk0, mpnw);
    mpmuld (sk0, dpe1, sk1, mpnw);
    mpmulx (a, b, sk0, mpnw);
    mpsqrx (sk0, b, mpnw);
    mpeq (sk1, a, mpnw);
    mpsub (a, b, sk0, mpnw);
  }
  //  Check for convergence.
  while (sk0.nw != 0. && (sk0.exponent < s1 || sk0.exponent >= -2));
  
#if DEBUG == 2
  if (MPIDB >= 6) 
  {
    cerr << "MPAGMX: Iter. " << l1 
      << " Tol. achieved ";
    cerr.width(9); cerr.precision(1);
    cerr << sk0.exponent << endl;
    cerr.width(0); cerr.precision(0);
  }
#endif
}

void mp::mpcshx (const mp& a, const mp& pi, const mp& al2, 
                 mp& x, mp& y, int mpnw)
{
  mpsize mpnw2(mpnw+2);
  mp f(mpsize(6)), sk0(mpnw2), sk1(mpnw2), sk2(mpnw2);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(x); zero(y);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 5)  mpdeb ("MPCSHX I ", a, mpnw);
#endif
  
  f.sign=1; f.nw=1; f.exponent=0;
  f[0]=1; f[1]=0;
  
  mp_dpe dpe1(0.5);
  mpexpx (a, pi, al2, sk0, mpnw);
  mpdivx (f, sk0, sk1, mpnw);
  mpadd (sk0, sk1, sk2, mpnw);
  mpmuld (sk2, dpe1, x, mpnw);
  mpsub (sk0, sk1, sk2, mpnw);
  mpmuld (sk2, dpe1, y, mpnw);
  
#if DEBUG == 2
  if (MPIDB >= 5) 
  {
    mpdeb ("MPCSHX O ", x, mpnw);
    mpdeb ("MPCSHX O ", y, mpnw);
  }
#endif
}

void mp::mpexpx (const mp& a, const mp& pi, const mp& al2, mp& b, int mpnw)
{
  int k;
  const double alt = 0.693147180559945309, cl2 = 1.4426950408889633, 
    cpi = 3.141592653589793238;
  const int nit = 1;
  mpsize mpnw2(mpnw+2);
  mp f1(mpsize(6)), sk0(mpnw2), sk1(mpnw2), sk2(mpnw2);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
#if DEBUG == 2
  if (MPIDB >= 5)  mpdeb ("MPEXPX I ", a, mpnw);
#endif
  
  int ncr = int(pow(2, MPMCR));
  
  mp_dpe t1;
  mpmdc (a, t1);
  t1.a *=  pow(2.0, t1.n);
  
  //  Check if precision level is too low to justify the advanced routine.
  
  if (mpnw <= ncr) 
  {
    mpexp (a, al2, b, mpnw);
    
#if DEBUG == 2
    if (MPIDB >= 6)  
      mpdeb ("MPEXPX O ", b, mpnw);
#endif
    return;
  }
  
  //  Check if Log(2) has been precomputed.
  
  mp_dpe t2;
  mpmdc (al2, t2);
  if (t2.n != - mpnbt || fabs (t2.a * pow(0.50, mpnbt) - alt) > mprx2)
  {
    if (MPKER[37] != 0) 
    {
      cerr << "*** MPEXPX: LOG (2) must be precomputed.\n";
      MPIER = 37;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  
  //    Check if Pi has been precomputed.
  
  mpmdc (pi, t2);
  if (t2.n != 0 || fabs (t2.a - cpi) > mprx2) 
  {
    if (MPKER[38] != 0) 
    {
      cerr << "*** MPEXPX: PI must be precomputed.\n";
      MPIER = 38;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  //  Check for overflows and underflows.
  
  if (t1.a >= 1e9) 
  {
    if (t1.a > 0.0) 
    {
      if (MPKER[39] != 0) 
      {
        cerr.width(12);
        cerr.precision(6);
        cerr << "*** MPEXPX: Argument is too large "
          << t1.a << " x 10 ^";
        cerr.width(0);
        cerr.precision(0);
        cerr << t1.n;
        MPIER = 39;
        if (MPKER[MPIER] == 2)  mpabrt();
      }
      return;
    } 
    else 
    {
      zero(b);
#if DEBUG == 2
      if (MPIDB >= 6)  
        mpdeb ("MPEXPX O ", b, mpnw);
#endif
      return;
    }
  }
  
  int nws = mpnw;
  f1.sign=1; f1.nw=1; f1.exponent=0;
  f1[0]=1; f1[1]=0;
  
  //  Determine the least integer MQ such that 2 ^ MQ >= MPNW.
  t2.a = nws;
  int mq = int(cl2 * log (t2.a) + 1.0 - mprxx);
  
  mpadd (a, f1, sk0, mpnw);
  
  //  Compute initial approximation to Exp (A).
  
  mpnw = ncr;
  mpexp (a, al2, b, mpnw);
  int iq = 0;
  
  //  Perform the Newton-Raphson iteration described above with a dynamically
  //  changing precision level MPNW.
  for(k = MPMCR + 1; k<=mq; k++)
  {
    mpnw = MIN (2 * mpnw, nws);
    bool cont = true;
    while(cont)
    {
      mplogx (b, pi, al2, sk1, mpnw);
      mpsub (sk0, sk1, sk2, mpnw);
      mpmulx (b, sk2, sk1, mpnw);
      mpeq (sk1, b, mpnw);
      if (k == mq - nit && iq == 0) 
        iq = 1;
      else
        cont=false;
    }
  }
  
#if DEBUG == 2
  if (MPIDB >= 6)  
    mpdeb ("MPEXPX O ", b, mpnw);
#endif
}

void mp::mplogx (const mp& a, const mp& pi, const mp& al2, mp& b, int mpnw)
{
  const double alt = 0.693147180559945309, cpi = 3.141592653589793;
  const int mzl = -5;
  
  double st, tn;
  mpsize mpnw2(mpnw+2);
  mp f1(mpsize(6)), f4(mpsize(6)), 
    sk0(mpnw2), sk1(mpnw2), sk2(mpnw2), sk3(mpnw2);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(b);
    return;
  }
  
#if DEBUG == 2
  if (MPIDB >= 6)  mpdeb ("MPLOGX I ", a, mpnw);
#endif
  
  int na = MIN (a.nw, mpnw);
  int ncr = int(pow(2, MPMCR)); // *cast*
  
  //  Check if precision level is too low to justify the advanced routine.
  
  if (mpnw <= ncr) 
  {
    mplog (a, al2, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 6)  mpdeb ("MPLOG O ", b, mpnw);
#endif
    return;    
  }
  
  if (a.sign < 0 || na == 0) 
  {
    if (MPKER[52] != 0) 
    {
      cerr << "*** MPLOGX: Argument is less than or equal to zero.\n";
      MPIER = 52;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  //  Check if Pi has been precomputed.
  mp_dpe t1;
  mpmdc (pi, t1);
  if (t1.n != 0 || fabs (t1.a - cpi) > mprx2) 
  {
    if (MPKER[53] != 0) 
    {
      cerr << "*** MPLOGX: PI must be precomputed.\n";
      MPIER = 53;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  //  Unless the input is 2, Log (2) must have been precomputed.
  mp_dpe t2;
  int it2;
  if (a.nw != 1 || a.exponent != 0 || a[0] != 2 || a.sign != 1) 
  {
    it2 = 0;
    mpmdc (al2, t2);
    if (t2.n != - mpnbt || fabs (t2.a * pow(0.50, mpnbt) - alt) > mprx2) 
    {
      if (MPKER[54] != 0) 
      {
        cerr << "*** MPLOGX: LOG (2) must be precomputed.\n";
        MPIER = 54;
        if (MPKER[MPIER] == 2)  mpabrt();
      }
      return;
    }
  }
  else  
    it2 = 1;
  
  
  f1.nw=1; f1.sign=1; f1.exponent=0; f1[0]=1; f1[1]=0;
  f4.nw=1; f4.sign=1; f4.exponent=0; f4[0]=4; f4[1]=0;
  
  
  //  If argument is 1, the result is zero.  If the argument is extremely close
  //  to 1.  If so, employ a Taylor's series instead.
  
  mpsub (a, f1, sk0, mpnw);
  
  if (sk0.nw == 0) 
  {
    zero(b);
#if DEBUG == 2
    if (MPIDB >= 6)  mpdeb ("MPLOG O ", b, mpnw);
#endif
    return;
  } 
  else if (sk0.exponent <= mzl) 
  {
    mpeq (sk0, sk1, mpnw);
    mpeq (sk1, sk2, mpnw);
    int i1 = 1;
    int is = 1;
    int tl = sk0.exponent - mpnw - 1;
    
    do
    {
      i1++;
      is = - is;
      st = is * i1;
      mpmulx (sk1, sk2, sk3, mpnw);
      mpdivd (sk3, mp_dpe(st), sk2, mpnw);
      mpadd (sk0, sk2, sk3, mpnw);
      mpeq (sk3, sk0, mpnw);
    }
    while (sk2.exponent >= tl); 
    
    mpeq (sk0, b, mpnw);
#if DEBUG == 2
    if (MPIDB >= 6)  mpdeb ("MPLOG O ", b, mpnw);
#endif
    return;
  }
  
  //  If input is exactly 2, set the exponent to a large value.  Otherwise
  //  multiply the input by a large power of two.
  
  mpmdc (a, t1);
  t2.n = mpnbt * (mpnw / 2 + 2) - t1.n;
  tn = t2.n;
  t2.a = 1.0;
  if (it2 == 1) 
    mpdmc (t2, sk0);
  else 
    mpmuld (a, t2, sk0, mpnw);
  
  
  //  Perform AGM iterations. 
  mpeq (f1, sk1, mpnw);
  mpdivx (f4, sk0, sk2, mpnw);
  mpagmx (sk1, sk2, mpnw);
  
  //  Compute B = Pi / (2 * A), where A is the limit of the AGM iterations.
  
  mpmuld (sk1, mp_dpe(2.0), sk0, mpnw);
  mpdivx (pi, sk0, sk1, mpnw);
  
  //  If the input was exactly 2, divide by TN.  Otherwise subtract TN * Log(2).
  
  if (it2 == 1) 
    mpdivd (sk1, mp_dpe(tn), sk0, mpnw);
  else 
  {
    mpmuld (al2, mp_dpe(tn), sk2, mpnw);
    mpsub (sk1, sk2, sk0, mpnw);
  }
  mpeq (sk0, b, mpnw);
  
#if DEBUG == 2
  if (MPIDB >= 6)  mpdeb ("MPLOG O ", b, mpnw);
#endif
  
}

void mp::mppix (mp& pi, int mpnw)
{
  int k;
  double t1;
  const double cl2 = 1.4426950408889633;
  mpsize mpnw2(mpnw+2);
  mp f(mpsize(6)), sk0(mpnw2), sk1(mpnw2), sk2(mpnw2), sk3(mpnw2), sk4(mpnw2);
  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    zero(pi);
    return;
  }
  
  int ncr = int(pow(2, MPMCR)); // *cast*
  
  //  Check if precision level is too low to justify the advanced routine.
  
  if (mpnw <= ncr) 
  {
    mppi (pi, mpnw);
#if DEBUG == 2
    if (MPIDB >= 7)  mpdeb ("MPPI O ", pi, mpnw);
#endif
    return;
  }
  
  //  Determine the number of iterations required for the given precision level.
  //  This formula is good only for this Pi algorithm.
  
  t1 = mpnw * log10 (mpbdx);
  int mq = int(cl2 * (log (t1) - 1.0) + 1.0); // *cast*
  
  //  Initialize as above.
  sk0.nw=1; sk0.sign=1; sk0.exponent=0; sk0[0]=1;
  f.nw=1; f.sign=1; f.exponent=0; f[0]=2; f[1]=0;
  
  mpsqrx (f, sk2, mpnw);
  mpmuld (sk2, mp_dpe(0.50), sk1, mpnw);
  f.exponent = -1;
  f[0] = float(0.50 * mpbdx); // *cast*
  mpsub (sk2, f, sk4, mpnw);
  
  //  Perform iterations as described above.
  
  mp_dpe dpe1(0.50);
  for (k = 1; k<= mq; k++)
  {
    mpadd (sk0, sk1, sk2, mpnw);
    mpmulx (sk0, sk1, sk3, mpnw);
    mpsqrx (sk3, sk1, mpnw);
    mpmuld (sk2, dpe1, sk0, mpnw);
    mpsub (sk0, sk1, sk2, mpnw);
    
    mpsqx (sk2, sk3, mpnw);
    t1 = pow(2.0, k);
    
    mpmuld (sk3, mp_dpe(t1), sk2, mpnw);
    
    mpsub (sk4, sk2, sk3, mpnw);
    mpeq (sk3, sk4, mpnw);
  }
  
  //  Complete computation.
  mpadd (sk0, sk1, sk2, mpnw);
  mpsqx (sk2, sk3, mpnw);
  mpdivx (sk3, sk4, sk2, mpnw);
  mpeq (sk2, pi, mpnw);
  
#if DEBUG == 2
  if (MPIDB >= 7)  mpdeb ("MPPIX O ", pi, mpnw);
#endif    
}
