#include "mp.h"

void mp::mpfftcr (int is, int m, int n, DComplex x[], double y[])
{
  int k;
  
  int mx = int(real(mpuu1[0])); // *cast*
  
  //  Check if input parameters are invalid.
  
  if ((is != 1 && is != -1) || m < 3 || m > mx) 
  {
    if (MPKER[27] != 0) 
    {
      cerr << "*** MPFFTCR: Either the UU arrays have not been initialized "
        << "or one of the input parameters is invalid: " 
        << is << '\t' << m << '\t' << mx << endl;
      MPIER = 27;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
	
  DComplex *dc1 = new DComplex [n/2], a1, a2, x1, x2;
	
  int n1 = int(pow(2, (m / 2))); // *cast*
  int n2 = n / 2;
  int n4 = n / 4;
  
  //  Construct the input to MPFFT1.
  dc1[0] = 0.50 * DComplex (real(x[0] + x[n2]),  real(x[0] - x[n2]));
  
  if (is == 1) 
    dc1[n4] = conjg(x[n4]);
  else 
    dc1[n4] = x[n4];
  
  int ku = n2;
  
  if (is == 1) 
  {
    //dir$ ivdep
    for (k = 1; k< n4; k++)
    {
      x1 = x[k];
      x2 = conjg (x[n2-k]);
      a1 = x1 + x2;
      a2 = DComplex(0.0, 1.0) * mpuu1[k+ku] * (x1 - x2);
      dc1[k] = 0.50 * (a1 + a2);
      dc1[n2-k] = 0.50 * conjg (a1 - a2);
    }
  } 
  else 
  {
    //dir$ ivdep
    for (k = 1; k< n4; k++)
    {
      x1 = x[k];
      x2 = conjg (x[n2-k]);
      a1 = x1 + x2;
      a2 = DComplex(0.0, 1.0) * conjg (mpuu1[k+ku]) * (x1 - x2);
      dc1[k] = 0.50 * (a1 + a2);
      dc1[n2-k] = 0.50 * conjg (a1 - a2);
    }
  }
  
  //  Perform a normal N/2-point FFT on DC1.
  
  mpfft1 (is, m - 1, n1, n2 / n1, dc1, x);
  
  //  Copy DC1 to Y such that DC1(k) = Y(2k-1) + i Y(2k).
  
  for(k = 0; k< n / 2; k++)
  {
    y[2*k] = real(dc1[k]);
    y[2*k+1] = aimag (dc1[k]);
  }
  delete [] dc1;
}

void mp::mpfftrc (int is, int m, int n, const double x[], DComplex y[])
{
  int k;
  int mx = int(real(mpuu1[0])); // *cast*
  
  //  Check if input parameters are invalid.
  
  if ((is != 1 && is != -1) || m < 3 || m > mx) 
  {
    if (MPKER[67] != 0) 
    {
      cerr << "*** MPFFTRC: Either the UU arrays have not been initialized "
        << "or one of the input parameters is invalid: " 
        << is << '\t' << m << '\t' << mx << endl;
      MPIER = 67;
      if (MPKER[MPIER] == 2)  mpabrt();
    }
    return;
  }
  
  DComplex *dc1 = new DComplex[n/2], a1, a2, z1, z2;
  
  int n1 = int (pow(2,(m / 2))); // *cast*
  int n2 = n / 2;
  int n4 = n / 4;
  
  //  Copy X to DC1 such that DC1(k) = X(2k-1) + i X(2k).
  
  for(k = 0; k< n2; k++)
    dc1[k] = DComplex (x[2*k], x[2*k+1]);
  
  //  Perform a normal N/2-point FFT on DC1.
  
  mpfft1 (is, m - 1, n1, n2 / n1, dc1, y);
  
  //  Reconstruct the FFT of X.
  
  y[0] = DComplex (2.0 * (real (dc1[0]) + aimag (dc1[0])), 0.0);
  
  if (is == 1) 
    y[n4] = 2.0 * dc1[n4];
  else 
    y[n4] = 2.0 * conjg (dc1[n4]);
  
  y[n2] = DComplex (2.0 * (real (dc1[0]) - aimag (dc1[0])), 0.0);
  int ku = n2;
  
  if (is == 1) 
  {
    //!dir$ ivdep
    for(k = 1; k< n4; k++)
    {
      z1 = dc1[k];
      z2 = conjg (dc1[n2-k]);
      a1 = z1 + z2;
      a2 = DComplex(0.0, -1.0) * mpuu1[k+ku] * (z1 - z2);
      y[k] = a1 + a2;
      y[n2-k] = conjg (a1 - a2);
    }
  } 
  else 
  {
    //!dir$ ivdep
    for (k = 1; k< n4; k++)
    {
      z1 = dc1[k];
      z2 = conjg (dc1[n2-k]);
      a1 = z1 + z2;
      a2 = DComplex(0.0, -1.0) * conjg (mpuu1[k+ku]) * (z1 - z2);
      y[k] = a1 + a2;
      y[n2-k] = conjg (a1 - a2);
    }
  }
  delete [] dc1;
}

void mp::mpfft1 (int is, int m, int n1, int n2, DComplex x[], DComplex y[])
{
  int i, j, k;
  DComplex *q1 = new DComplex[mpnsp2], *q2 = new DComplex[mpnsp2], 
    **z1 = new DComplex*[mpnrow+mpnsp1], **z2 = new DComplex*[mpnrow+mpnsp1];
  
  int yrow = n2 + mpnsp1;
  for(i = 0; i<mpnrow+mpnsp1; i++)
  {
    z1[i] = new DComplex[n1];
    z2[i] = new DComplex[n1];
  }
	
  int m1 = (m + 1) / 2;
  int m2 = m - m1;
  int nr1 = MIN (n1, mpnrow);
  int nr2 = MIN (n2, mpnrow);
  int ku = int(real(mpuu2[m-1])); // *cast*
  
  for (i = 0; i<n1; i += nr1)
  {
    //  Copy NR1 rows of X (treated as a N1 x N2 complex array) into Z1.
    for(k = 0; k<nr1; k++)
    {
      for(j = 0; j<n2; j++)
        z1[k][j] = x[j*n1 + i+k];
    }
    
    //  Perform NR1 FFTs, each of length N2.
    
    mpfft2 (is, nr1, m2, n2, z1, z2);
    
    //  Multiply the resulting NR1 x N2 complex block by roots of unity and
    //  store transposed into the appropriate section of Y.
    
    int iu = i + ku - n1 - 1;

    if (is == 1) 
    {
      for(k = 0; k<nr1; k++)
      {
        for (j = 0; j< n2; j++)
          y[(i+k)*yrow + j] = mpuu2[iu+k+ (j+1)*n1] * z1[k][j];
      } 
    } 
    else 
    {
      for(k = 0; k<nr1; k++)
      {
        for (j = 0; j<n2; j++)
          y[(i+k)*yrow + j] = conjg (mpuu2[iu+k+ (j+1)*n1]) * z1[k][j];
      }
    }
  }
  
  for (i = 0; i< n2; i += nr2)
  {
    
    //  Copy NR2 rows of the Y array into Z2.
    
    for(k = 0; k< nr2; k++)
    {
      for(j = 0; j< n1; j++)
        z2[k][j] = y[j*yrow + i+k];
    }
    
    //  Perform NR2 FFTs, each of length N1.
    
    mpfft2 (is, nr2, m1, n1, z2, z1);
    
    //  Copy NR2 x N1 complex block back into X array.  It's a little more
    //  complicated if M is odd.
    
    if ((m%2) == 0) 
    {
      for(k = 0; k< nr2; k++)
      {
        for (j = 0; j< n1; j++)
          x[i+k + j*n1] = z2[k][j];
      }
    } 
    else 
    {
      int j2;
      for (j = 0; j<n1 / 2; j++)
      {
        j2 = 2 * j;
        //dir$ ivdep
        for (k = 0; k< nr2; k++)
        {
          x[i+k + j*n1] = z2[k][j2]; 
          x[i+k+n2 + n1*j] = z2[k][j2+1]; 
        }   
      }
    }
  }
  
  delete [] q1; delete[] q2;
  for(i = 0; i<mpnrow+mpnsp1; i++)
  {delete [] z1[i]; delete[] z2[i];}
  delete [] z1; delete [] z2;
}

void mp::mpfft2 (int is, int ns, int m, int n, DComplex **x, DComplex **y)
{
  int l, j, i;
  
  //  Perform the second variant of the Stockham FFT.

  for(l = 1; l<=m; l+=2)
  {
    mpfft3 (is, l, ns, m, n, x, y);
    if (l == m)
    {
      for(i = 0; i<ns; i++) 
      {
	for (j = 0; j<n; j++)
	  x[i][j] = y[i][j];
      }
      return;
    }
    mpfft3 (is, l + 1, ns, m, n, y, x);
  }  
}

void mp::mpfft3 (int is, int l, int ns, int m, int n, DComplex **x, DComplex **y)
{
  
  DComplex u1, x1, x2;
  int i,j, k;
  
  //  Set initial parameters.
  int n1 = n / 2;
  int lk = int(pow(2,(l - 1))); // *cast*
  int li = int(pow(2, (m - l))); // *cast*
  int lj = 2 * lk;
  int ku = li;
  
  int i11, i12, i21, i22;
  for(i = 0; i<= li - 1; i++)
  {
    i11 = i * lk + 1;
    i12 = i11 + n1;
    i21 = i * lj + 1;
    i22 = i21 + lk;
    if (is == 1) 
      u1 = mpuu1[i+ku];
    else 
      u1 = conjg (mpuu1[i+ku]);
    
    for (k = -1; k< lk - 1; k++)
    {
      //!dir$ ivdep
      for (j = 0; j< ns; j++)
      {
        x1 = x[j][i11+k];
        x2 = x[j][i12+k];
        y[j][i21+k] = x1 + x2;
        y[j][i22+k] = u1 * (x1 - x2);
      }
    }
  }
}


void mp::mplconv (int iq, int n, int nsq, double a[], double b[], double c[])
{
  int i,j,k;
  const double cl2 = 1.4426950408889633, erm = 0.438;
#if CRAY == 1
  const int mbt = 48;
#else
  const int mbt = 53;
#endif
  double an, t1, t2;
  
  
  //  Handle the case where N is less than NCR1 = 2 ** (MPMCR-1).  If IQ < 0, 
  //  only the second half of the result vector is returned, since the first 
  //  half won't be used.
  
  int ncr1 = int(pow(2, (MPMCR - 1))); // *cast*
  int n1, n2;
  if (n < ncr1) 
  {
    switch(iq)
    {
    case 1:
      for(k = 0; k<2 * n; k++)
      {
        t1 = 0.0;
        n1 = MAX (k - n + 2, 1);
        n2 = MIN (k+1, n);
        
        for(j = n1-1; j<n2; j++)
          t1 += a[j] * a[k-j];
        
        c[k] = t1;
      }
      break;
      
    case 2: 
      for(k = 0; k< 2 * n; k++)
      {
        t1 = 0.0;
        n1 = MAX (k - n + 2, 1);
        n2 = MIN (k+1, n);
        
        for(j = n1-1; j< n2; j++)
          t1 += a[j] * b[k-j];
        
        c[k] = t1;
      }
      break;
      
    case -1:
      for (k = 0; k< n - 1; k++)
        c[k] = 0.0;
      
      for(k = n-1; k< 2 * n; k++)
      {
        t1 = 0.0;
        n1 = k - n + 2;
        n2 = n;
        
        for(j = n1-1; j< n2; j++)
          t1 += a[j] * a[k-j];
        
        c[k] = t1;
      }
      break;
      
    case -2:
      for (k = 0; k< n - 1; k++)
        c[k] = 0.0;
      
      for(k = n-1; k<2 * n; k++)
      {
        t1 = 0.0;
        n1 = k - n + 2;
        n2 = n;
        
        for (j = n1 - 1; j< n2; j++)
          t1 += a[j] * b[k-j];
        
        c[k] = t1;
      }
      break;
    }
    return;
  }
  
  double *d1 = new double[3*n+2], *d2 = new double[3*n+2], 
    *d3 = new double[3*n+2];
  DComplex *dc1 = new DComplex[3*n/2+nsq*mpnsp1+3], 
    *dc2 = new DComplex[3*n/2+nsq*mpnsp1+3];
  
  //  Determine M1 and N1.  Note that by this reckoning, N1 <= 1.5 N.  This is
  //  the reason for the 3*n/2 dimensions above.
  
  t1 = 0.750 * n;
  int m1 = int(cl2 * log (t1) + 1.0 - mprxx); // *cast*
  n1 = int (pow(2, m1)); // *cast*
  int m2 = m1 + 1;
  n2 = 2 * n1;
  int n4 = 2 * n2;
  int nm = MIN (2 * n, n2);
  
  if (abs (iq) == 1) 
  {
    for(i = 0; i<n; i++)
      d1[i] = a[i];
    
    for(i = n ; i< n2; i++)
      d1[i] = 0.0;
    
    //  Perform a forward real-to-complex FFT on the vector in a.
    
    mpfftrc (1, m2, n2, d1, dc1);
    
    //  Square the resulting complex vector.
    
    for(i = 0; i< n1 + 1; i++)
      dc1[i] *= dc1[i];
  } 
  else 
  {
    for(i = 0; i<n; i++)
    {
      d1[i] = a[i];
      d2[i] = b[i];
    }
    
    for(i = n; i< n2; i++)
    {
      d1[i] = 0.0;
      d2[i] = 0.0;
    }
    
    //  Perform forward real-to-complex FFTs on the vectors in a and b.
    
    mpfftrc (1, m2, n2, d1, dc1);
    mpfftrc (1, m2, n2, d2, dc2);
    
    //  Multiply the resulting complex vectors.
    
    for(i = 0; i<= n1; i++)
      dc1[i] *= dc2[i];
  }
  
  //  Perform an inverse complex-to-real FFT on the resulting data.
  
  mpfftcr (-1, m2, n2, dc1, d3);
  
  //  Divide by N4.
  // here!!!!!!!!!
  an = 1.0 / n4;
  
  for(i = 0; i< nm; i++)
  {
    t1 = an * d3[i];
    t2 = NINT (t1);
    // D1(I) = ABS (T2 - T1)
    c[i] = t2;
  }
  
  //  Find the largest FFT roundoff error.  Roundoff error is minimal unless
  //  exceedingly high precision (i.e. over one million digits) is used.  Thus
  //  this test may be disabled in normal use.  To disable this test, set the
  //  skip flag to false and comment out the line of the previous loop
  //  that begins D1(I) =.  To enable, for  the reverse.
  
  //  This code can be used as a rigorous system integrity test.  First set
  //  MBT according to the system being used, and { set ERM to be fairly
  //  small, say 0.001 or whatever is somewhat larger than the largest FFT
  //  roundoff error typically encountered for a given precision level on the
  //  computer being used.  Enable this test as explained in the previous
  //  paragraph.  Then if an anomalously large roundoff error is detected, a
  //  hardware or compiler error has likely occurred.
  
  bool skip = true;
  if(!skip)
  {
    int i1;
    t1 = 0.0;
    
    for(i = 0; i< nm; i++)
    {
      if (d1[i] > t1) 
      {
        i1 = i;
        t1 = d1[i];
      }
    }
    //  Check if maximum roundoff error exceeds the limit ERM, which is set above.
    //  Also determine the number of fractional bits and how large the error is in
    //  terms of units in the last place (ulp).
    
    if (t1 > erm)  
    {
      if (MPKER[55] != 0) 
      {
        int i2, i3, i4, i5;
        t2 = an * d1[i1];
        i2 = int(cl2 * log (t1) + 1.0 + mprxx); // *cast*
        i3 = int(cl2 * log (t2) + 1.0 + mprxx); // *cast*
        i4 = mbt + i2 - i3;
        i5 = int(t1 * pow(2, i4) + mprxx); // *cast
        cerr << "*** MPLCONV: Excessive FFT roundoff error\t"
          << i1 << '\t';
        cerr.width(10); cerr.precision(6);
        cerr << t1; cerr.width(0); cerr.precision(0);
        cerr << i4 << '\t' << i5 << endl;
        MPIER = 55;
        if (MPKER[MPIER] == 2)  mpabrt();
      }
    }
    
  }
  //  Handle case where n > n1.
  int m, m21, ms;
  if (n > n1) 
  {
    m = n - n1;
    m2 = 2 * m;
    m21 = 2 * m - 1;
    ms = int(sqrt (3.0 * m21) + mprxx); // *cast*
    k = n1 - m + 1;
    
    if (abs (iq) == 1) 
    {
      for(i = 0; i< m21; i++)
        d1[i] = a[k+i];
      
      mplconv (-1, m21, ms, d1, d2, d3);
    } 
    else 
    {
      for (i = 0; i<m21; i++)
      {
        d1[i] = a[k+i];
        d2[i] = b[k+i];
      }
      
      mplconv (-2, m21, ms, d1, d2, d3);
    }
    
    int ii;
    for(i = 0; i< m2; i++)
    {
      ii = i + m2 - 2;
      c[i] -= d3[ii];
      c[i+n2] = d3[ii];
    }
  }
  
  delete [] d1; delete [] d2; delete[] d3;
  delete [] dc1; delete [] dc2;
}

