#include "mp.h"
#include <string.h>

void mp::mpoutx (const mp& a, char* b, int& n, int mpnw)
{
  int i;
  const double al2 = 0.301029995663981195;  
  if (MPIER != 0) 
  {
    if (MPIER == 99)  mpabrt();
    b[0] = '\0';
    n = 0;
    return;
  }
  
#if DEBUG == 2
  int no;
  if (MPIDB >= 7) 
    debug("MPOUTX I ", a);
#endif
  
  int na = MIN (a.nw, mpnw);
  
  int ncr = int(pow(2, MPMCR));
  
  //  Check if actual precision level of argument is too low to justify the
  //  advanced routine.
  
  if (na <= ncr) 
  {
    mpoutc (a, b, n, mpnw);
#if DEBUG == 2
    if (MPIDB >= 7)
    {
      no = MIN (n, 6 * MPNDB + 20);
      cerr << "MPOUTCX O "; 
      for(i=0; i<no; i++)
        cerr << b[i];
    }
#endif
    return;
  }
  double t1, t2;
  char dig[] = "0123456789";
  char c1[17], c2[17];
  char *b1 = new char[8*mpnw+31], *b2 = new char[8*mpnw+31];
	
  mpsize mpnw2(mpnw+2);
  mp sk0(mpnw2), sk1(mpnw2), sk2(mpnw2), sk3(mpnw2), sk4(mpnw2);
  
  //  Normalize input to an integer by multiplying by a suitable power of 10.
  
  t1 = a[0] + mprdx * a[1] + mprx2 * a[2];
  t2 = log10 (t1);
  int m1 = int (MAX (al2 * mpnbt * (a.nw - a.exponent) - t2, 0.0)); // *cast*
  mp_dpe dpe1(10,0); 
  mpdmc (dpe1, sk0);
  mpnpwx (sk0, m1, sk2, mpnw);
  mpmulx (a, sk2, sk1, mpnw);
  sk1.sign = 1;
  
  //  Split large integer into two approximately equal decimal sections.
  mp_dpe dpe2, dpe3;
  mpmdc(sk1, dpe2);
  
  mp_dpe::dpdec (dpe2, dpe3);
  
  int m2 = dpe3.n / 2;
  mpnpwx (sk0, m2, sk3, mpnw);
  mpdivx (sk1, sk3, sk0, mpnw);
  mpinfr (sk0, sk2, sk4, mpnw);
  mpmulx (sk2, sk3, sk0, mpnw);
  mpsub (sk1, sk0, sk3, mpnw);
  
  //  Recursively convert each section.
  
  int mpnws = mpnw;
  mpnw = sk2.nw + 1;
  int nb1, nb2;
  mpoutx (sk2, b1, nb1, mpnw);
  mpnw = sk3.nw + 1;
  mpoutx (sk3, b2, nb2, mpnw);
  mpnw = mpnws;
  
  //  Obtain decimal exponents from each section.
  for(i = 0; i<17; i++)
  { 
    c1[i] = '\0';
    c2[i] = '\0';
  }
  
  for(i = 0; i<10; i++)
  {
    c1[i] = b1[i+4];
    c2[i] = b2[i+4];
  }
  
  int ie1 = atoi(c1);
  int ie2 = atoi(c2);
  
  //  Set exponent of result.
  
  int ie = ie1 + m2 - m1;
  
  sprintf(c1, "%14d", ie);
  
  for (i = 0; i< 4; i++)
    b[i] = b1[i];
  
  for(i = 4; i< 14; i++)
    b[i] = c1[i];
  
  //  Copy mantissa of first section.
  
  for(i = 14; i<nb1; i++)
    b[i] = b1[i];
  
  int i2, i1 = ie1 + m2 - ie2 + 19;
  
  //  If first section is too long, then round trailing digits (probably 9s).
  char* location;
  if (nb1 > i1) 
  {
    location = strchr (dig, b[i1]);
		
    i2 = int(location - dig);
    if (i2 >= 5) 
    {
      bool skip = false;
      for(i = i1-1; i>=20; i--)
      {
        if (b[i] != '9')
        {
          skip = true;
          break;
        }
        b[i] = '0';
      }
      
      if(!skip)
      {
        cerr << "*** MPOUTX: Exceptional case -- contact author\n";
        exit(-1);
      }
      
      location = strchr (dig, b[i]);
      i2 = int(location - dig);
      sprintf(c1, "%1d", (i2+1));
      b[i] = c1[0];
    }
  } 
  else if (nb1 < i1) 
  {
    
    //  If first section is too short, then insert zeroes in gap.
    
    for(i = nb1; i< i1; i++)
      b[i] = '0';
  }
  
  //  Copy mantissa of second section.
  
  b[i1] = b2[18];
  n = MIN (i1 + nb2 - 19, int (7.225 * mpnw + 30));
  
  for(i = i1 + 1; i< n; i++)
    b[i] = b2[i-i1+19];
  
  //  Fix SIGN.
  
  if (a.sign == -1) b[17] = '-';
  
  // Put null terminating character;
  b[n] = '\0';
  
#if DEBUG == 2
  if (MPIDB >= 7)
  {
    no = MIN (n, 6 * MPNDB + 20);
    cerr << "MPOUTCX O "; 
    for(i=0; i<no; i++)
      cerr << b[i];
  }
#endif
  delete [] b1; delete [] b2;
}
