/** 
 * This file is created to avoid initialization problems with 
 * static members. By putting all the initialization in a file,
 * initialization order is enforced. 
 */

#include "mpreal.h"
#include "mpfun_Mutexes.h"

const double *mp_dpe::pt=init_pt();

// initializing static members
const double 
#if CRAY == 0 
mpfun::mpbbx=4096.0;
#else
mpfun::mpbbx=2048.0;
#endif 
const double mpfun::mpbdx=mpfun::mpbbx*mpfun::mpbbx; 
const double mpfun::mpbx2=mpfun::mpbdx*mpfun::mpbdx; 
const double mpfun::mprbx=1.0/mpfun::mpbbx;
const double mpfun::mprdx=mpfun::mprbx*mpfun::mprbx; 
const double mpfun::mprx2=mpfun::mprdx*mpfun::mprdx; 
const double mpfun::mprxx=16.0*mpfun::mprx2;

const int
#if CRAY == 0
mpfun::mpnbt=24, 
mpfun::mpnpr=32, 
mpfun::mpmcrx=7, 
mpfun::mpnrow=16,
#else
mpfun::mpnbt=22, 
mpfun::mpnpr=16, 
mpfun::mpmcrx=8, 
mpfun::mpnrow=64,
#endif 
mpfun::mpnsp1=2, 
mpfun::mpnsp2=9;


int 
mpfun::MPIDB=0,
mpfun::MPNDB=22,
mpfun::MPMCR=mpfun::mpmcrx,
mpfun::MPIRD=1,
mpfun::MPIER=99,
mpfun::MPKER[73+1]={-1,2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,
2,2,2,2,2,2,2,2,2,2};

//initializing static variables
const int mpfun::mpipl = 1000, 
mpfun::mpiou = 56, /*56*/ 
mpfun::mpiep = 10 - mpfun::mpipl, 
mpfun::mpwds = int((mpfun::mpipl /  7.224719896) + 1),
mpfun::mp2 = mpfun::mpwds + 2,
mpfun::mp21 = mpfun::mp2 + 1,
mpfun::mpoud = mpfun::mpiou, 
mpfun::MPNW = mpfun::mpwds;

DComplex *mpfun::mpuu1 = mpfun::init_mpuu1(MPNW+1), 
*mpfun::mpuu2 = mpfun::init_mpuu2(MPNW+1);

mp_real& mpfun::mppic=mp::init_mppic();
mp_real& mpfun::mpl02=mp::init_mpl02(); 
mp_real& mpfun::mpl10=mp::init_mpl10();
mp_real& mpfun::mpeps=mp::init_mpeps();



/********** Init functions **********/


void mpfun::mpabrt()
{
  if (MPIER == 99) 
    cerr << "*** The MPFUN library has not been initialized. \n";
  else
    cerr << "*** mpabrt: execution terminated, error code =" << MPIER << endl;
  
  //  use this line on cray systems.
  
  // abort
  
  //  use this line plus the c routine tracbk (available from author) on
  //  silicon graphics iris systems.
  
  // tracbk
  
  //  on other systems, merely terminate execution.
  
  exit(-1);
}

DComplex* mpfun::init_mpuu1(int n)
{
  DComplex *mpuu;
  const double cl2 = 1.4426950408889633, pi = 3.141592653589793238;
  double t1, ti;
  int j, i;
  
  t1 = 0.75 * n;
  int m = int(cl2 * log (t1) + 1.0 - mprxx); // *cast*
  int mq = m + 2;
  int nq = int(pow(2, mq)); // *cast*
  mpuu = new DComplex[nq];
  mpuu[0] = mq;
  
  int ku = 1;
  int ln = 1;
  
  for(j = 1; j<= mq; j++)
  {
    t1 = pi / ln;
    
    for (i = 0; i<= ln - 1; i++)
    {
      ti = i * t1;
      mpuu[i+ku] = DComplex (cos (ti), sin (ti));
    }
    
    ku +=  ln;
    ln *= 2;
  }
  return mpuu;
}

DComplex* mpfun::init_mpuu2(int n)
{
  DComplex* mpuu;
  const double cl2 = 1.4426950408889633, pi = 3.141592653589793238;
  double t1, tpn;
  int k, j, i;
  
  t1 = 0.75 * n;
  int m = int(cl2 * log (t1) + 1.0 - mprxx); // *cast*
  int mq = m + 2;
  int nq = int(pow(2, mq)); // *cast*
  mpuu = new DComplex [mq+nq];
  
  int ku = mq + 1;
  mpuu[0] = mq;
  
  int mm, nn, mm1, mm2, nn1, nn2, iu;
  for (k = 2; k<= mq - 1; k++)
  {
    mpuu[k-1] = ku;
    mm = k;
    nn = int(pow(2, mm)); // *cast*
    mm1 = (mm + 1) / 2;
    mm2 = mm - mm1;
    nn1 = int(pow(2, mm1)); // *cast*
    nn2 = int(pow(2, mm2)); // *cast*
    tpn = 2.0 * pi / nn;
    
    for(j = 0; j<nn2; j++)
    {
      for(i = 0; i<nn1; i++)
      {
        iu = ku + i + j * nn1;
        t1 = tpn * i * j;
        mpuu[iu-1] = DComplex (cos (t1), sin (t1));
      }
    }
    ku += nn;
  }
  return mpuu;
}

mp_real& mp::init_mppic()
{
  // other stuff
  MPIER = 0;
  mpsize a(mp21);
  mp_real* temp = new mp_real(a);
  mp::mppix(*temp, MPNW+1);
  return *temp;
}

mp_real& mp::init_mpl02()
{
  mpsize a(mp21);
  // mpl02
  mp_real *t1 = new mp_real(a), t2(mpsize(6));
  mp::mpdmc(mp_dpe(2.0), t2);
  mp::mplogx(t2, mp::mppic, *t1, *t1, MPNW+1);
  return *t1;
}

mp_real& mp::init_mpl10()
{
  mpsize a(mp21);
  mp_real *t1 = new mp_real(a), t2(mpsize(6));
  mp::mpdmc(mp_dpe(10.0, 0), t2);
  mp::mplogx(t2, mp::mppic, mp::mpl02, *t1, MPNW+1);
  return *t1;
	
}

mp_real& mp::init_mpeps()
{
  mpsize a(mp21);
  mp_real *t1 = new mp_real(a), t2(mpsize(6));
  mp::mpdmc(mp_dpe(10), t2);
  mp::mpnpwx(t2, mpiep, *t1, MPNW+1);
  
  // last words is not reliable
  mp::mppic.nw--; mp::mpl02.nw--;mp::mpl10.nw--;t1->nw--;
  return *t1;
}








