#include "mpint.h"

mp_int::mp_int(double da, const mpprecision& precision)
:mp(precision)
{
  mp mpt1(mpsize(6)), mpt2(mpsize(8));
  mpdmc(mp_dpe(da), mpt1);
  mpinfr(mpt1, *this, mpt2, MIN(maxnw - 2, MPNW));
}

mp_int::mp_int(char* ab, const mpprecision& precision)
:mp(precision)
{
  int mpnw = MIN(MPNW, maxnw-2);
  mpsize mpnw2(mpnw+2);
  mp mpt1(mpnw2), mpt2(mpnw2);
  mpdexc (ab, strlen(ab), mpt1, mpnw);
  mpinfr (mpt1, *this, mpt2, mpnw);
}

mp_int& mp_int::operator=(const int& ib)
{
  mpdmc(mp_dpe(double(ib)), *this);
  return *this;
}

mp_int& mp_int::operator=(const double& db)
{
  mp mpt1(mpsize(6)), mpt2(mpsize(8));
  mpdmc(mp_dpe(db), mpt1);
  mpinfr(mpt1, *this, mpt2, MPNW);
  return *this;
}

mp_int& mp_int::operator=(const char* ab)
{
  int mpnw = MIN(MPNW, maxnw-2);
  mpsize mpnw2(mpnw+2);
  mp mpt1(mpnw2), mpt2(mpnw2);
  mpdexc (ab, strlen(ab), mpt1, mpnw);
  mpinfr (mpt1, *this, mpt2, mpnw);
  return *this;
}

mp_int& mp_int::operator=(const mp_int& jb)
{
  int mpnw = MIN(MPNW, maxnw - 1);
  if(&jb != this)
    mpeq(jb, *this, mpnw);
	
  return *this;
}

mp_int& mp_int::operator=(const mp_real& qb)
{
  int mpnw = MIN(MPNW, maxnw - 2);
  mp mpt1(mpsize(mpnw+1)), mpt2(mpsize(mpnw+2));
  mpeq(qb, mpt1, mpnw);
  mpinfr(mpt1, *this, mpt2, mpnw);
	
  return *this;
}

mp_int& mp_int::operator=(const mp_complex& zb)
{
  if(zb.getReal() != this)
  {
    int mpnw = MIN(MPNW, maxnw - 2);
    mp mpt1(mpsize(mpnw+1)), mpt2(mpsize(mpnw+2));
		
    mpeq(*zb.getReal(), mpt1, mpnw);
    mpinfr(mpt1, *this, mpt2, mpnw);
  }
  return *this;
}

mp_int operator+(const mp_int& ja, const mp_int& jb)
{
  mp_int res, mpt1, mpt2;
  
  mp::mpadd(ja, jb, mpt1, mpfun::MPNW);
  mp::mpinfr(mpt1, res, mpt2, mpfun::MPNW);
  return res;
}

mp_int operator-(const mp_int& ja, const mp_int& jb)
{
  mp_int res, mpt1, mpt2;
  
  mp::mpsub(ja, jb, mpt1, mpfun::MPNW);
  mp::mpinfr(mpt1, res, mpt2, mpfun::MPNW);
  return res;
}

mp_int operator-(const mp_int& ja)
{
  mp_int res;
	
  mp::mpeq(ja, res, mpfun::MPNW);
  res.sign = - ja.sign;
  return res;
}

mp_int operator*(const mp_int& ja, const mp_int& jb)
{
  mp_int res, mpt1, mpt2;
  mp::mpmulx(ja, jb, mpt1, mpfun::MPNW);
  mp::mpinfr(mpt1, res, mpt2, mpfun::MPNW);
  return res;
}

mp_int operator/(const mp_int& ja, const mp_int& jb)
{
  mp_int res, mpt1, mpt2;
  mp::mpdivx(ja, jb, mpt1, mpfun::MPNW);
  mp::mpinfr(mpt1, res, mpt2, mpfun::MPNW);
  return res;
}

mp_int operator%(const mp_int& ja, const mp_int& jb)
{
  mp_int res, mpt1, mpt2, mpt3;
  mp::mpdivx(ja, jb, mpt1, mpfun::MPNW);
  mp::mpinfr(mpt1, mpt2, mpt3, mpfun::MPNW);
  mp::mpmulx(jb, mpt2, mpt1, mpfun::MPNW);
  mp::mpsub(ja, mpt1, res, mpfun::MPNW);
  return res;
}

mp_int mp_int::operator++()
{
  mp_int one("2");
  one.nw = 1; one.exponent = 0;
  one[0]=1; one[1] = 0;
  *this = *this + one;
  return *this;
}

mp_int mp_int::operator++(int)
{
  mp_int one("2"), temp;
  one.nw = 1; one.exponent = 0;
  one[0]=1; one[1] = 0;
  mpeq(*this, temp, MPNW);
  *this = *this + one;
  return temp;
}

mp_int mp_int::operator--()
{
  mp_int one("2");
  one.nw = 1; one.exponent = 0;
  one[0]=1; one[1] = 0;
  *this = *this - one;
  return *this;
}

mp_int mp_int::operator--(int)
{
  mp_int one("2"), temp;
  one.nw = 1; one.exponent = 0;
  one[0]=1; one[1] = 0;
  mpeq(*this, temp, MPNW);
  *this = *this - one;
  return temp;
}

mp_int abs(const mp_int& ja)
{
  mp_int res;  
  mp::mpeq(ja, res, mpfun::MPNW);
  res.sign=1;
  return res;
}

mp_int maximum(const mp_int& ja, const mp_int& jb)
{
  int ic;
  mp_int res;
  
  mp::mpcpr(ja, jb, ic, mpfun::MPNW);
  if (ic >= 0)
		mp::mpeq(ja, res, mpfun::MPNW);
  else
		mp::mpeq(jb, res, mpfun::MPNW);
  return res;
}

mp_int minimum(const mp_int& ja, const mp_int& jb)
{
  int ic;
  mp_int res;
  
  mp::mpcpr(ja, jb, ic, mpfun::MPNW);
  if (ic < 0)
		mp::mpeq(ja, res, mpfun::MPNW);
  else
		mp::mpeq(jb, res, mpfun::MPNW);
  return res;
}

mp_int sign(const mp_int& ja, const mp_int& jb)
{
  mp_int res;
  mp::mpeq(ja, res, mpfun::MPNW);
  res.sign=jb.sign;
  return res;
}

mp_int power(const mp_int& ja, const mp_int& jb)
{
  mp_int res, mpt1, mpt2;
  mp::mplogx(ja, mp::mppic, mp::mpl02, mpt1, mpfun::MPNW);
  mp::mpmulx(mpt1, jb, mpt2, mpfun::MPNW);
  mp::mpexpx(mpt2, mp::mppic, mp::mpl02, mpt1, mpfun::MPNW);
  mp::mpnint(mpt1, res, mpfun::MPNW);
  return res;
}


mp_int power(const int ia, const mp_int& jb)
{
  mp_int res, mpt1, mpt2, mpt3;
  mp::mpdmc(mp_dpe(double(ia)), mpt1);
  mp::mplogx(mpt1, mp::mppic, mp::mpl02, mpt2, mpfun::MPNW);
  mp::mpmulx(mpt2, jb, mpt3, mpfun::MPNW);
  mp::mpexpx(mpt3, mp::mppic, mp::mpl02, mpt1, mpfun::MPNW);
  mp::mpnint(mpt1, res, mpfun::MPNW);
  return res;
}

mp_int power(const mp_int& ja, const int ib)
{
  mp_int res, mpt1;  
  mp::mpnpwx(ja, ib, mpt1, mpfun::MPNW);
  mp::mpnint(mpt1, res, mpfun::MPNW);
  return res;
}




