#include "mpcomplex.h"


mp_real& mp_real::operator=(const int& ib)
{
  mpdmc(mp_dpe(double(ib)), *this);
  return *this;
}

mp_real& mp_real::operator=(const double& db)
{
  mpdmc(mp_dpe(double(db)), *this);
  return *this;
}

mp_real& mp_real::operator=(const mp_real& jb)
{
  if(&jb != this)
  {
    int mpnw = MIN(MPNW, maxnw - 1);
    mpeq(jb, *this, mpnw);
  }
  return *this;
}

mp_real& mp_real::operator=(const mp_int& jb)
{
  int mpnw = MIN(MPNW, maxnw - 1);
  mpeq(jb, *this, mpnw);
  return *this;
}

mp_real& mp_real::operator=(const mp_complex& zb)
{
  if(zb.getReal() != this)
  {
    int mpnw = MIN(MPNW, maxnw - 1);
    mpeq(*zb.getReal(), *this, mpnw);
  }
  return *this;
}

mp_real& mp_real::operator=(const char * ab)
{
  int mpnw = MIN(MPNW, maxnw-2);
  mpdexc(ab, strlen(ab), *this, mpnw);
  return *this;
}

mp_real operator+(const mp_real& qa, const mp_real& qb)
{
  mp_real res;  
  mp::mpadd(qa, qb, res, mpfun::MPNW);
  return res;
}

mp_real operator-(const mp_real& qa, const mp_real& qb)
{
  mp_real res;  
  mp::mpsub(qa, qb, res, mpfun::MPNW);
  return res;
}

mp_real operator-(const mp_real& qa)
{
  mp_real res;
	
  mp::mpeq(qa, res, mpfun::MPNW);
  res.sign = -qa.sign;
  return res;
}

mp_real operator*(const mp_real& qa, const mp_real& qb)
{
  mp_real res;  
  mp::mpmulx(qa, qb, res, mpfun::MPNW);
  return res;
}

mp_real operator/(const mp_real& qa, const mp_real& qb)
{
  mp_real res;  
  mp::mpdivx(qa, qb, res, mpfun::MPNW);
  return res;
}

mp_real mp_real::operator++()
{
  mp_real one("2");
  one.nw = 1; one.exponent = 0;
  one[0]=1; one[1] = 0;
  *this = *this + one;
  return *this;
}

mp_real mp_real::operator++(int)
{
  mp_real one("2"), temp;
  one.nw = 1; one.exponent = 0;
  one[0]=1; one[1] = 0;
  mpeq(*this, temp, MPNW);
  *this = *this + one;
  return temp;
}

mp_real mp_real::operator--()
{
  mp_real one("2");
  one.nw = 1; one.exponent = 0;
  one[0]=1; one[1] = 0;
  *this = *this - one;
  return *this;
}

mp_real mp_real::operator--(int)
{
  mp_real one("2"), temp;
  one.nw = 1; one.exponent = 0;
  one[0]=1; one[1] = 0;
  mpeq(*this, temp, MPNW);
  *this = *this - one;
  return temp;
}

mp_real abs(const mp_real& qa)
{
  mp_real res;
	
  mp::mpeq(qa, res, mpfun::MPNW);
  res.sign = 1;
  return res;
}

mp_real maximum(const mp_real& ja, const mp_real& jb)
{
  int ic;
  mp_real res;
  
  mp::mpcpr(ja, jb, ic, mpfun::MPNW);
  if (ic >= 0)
		mp::mpeq(ja, res, mpfun::MPNW);
  else
		mp::mpeq(jb, res, mpfun::MPNW);
  return res;
}

mp_real minimum(const mp_real& ja, const mp_real& jb)
{
  int ic;
  mp_real res;
  
  mp::mpcpr(ja, jb, ic, mpfun::MPNW);
  if (ic < 0)
    mp::mpeq(ja, res, mpfun::MPNW);
  else
    mp::mpeq(jb, res, mpfun::MPNW);
  return res;
}

mp_real sign(const mp_real& ja, const mp_real& jb)
{
  mp_real res;
  mp::mpeq(ja, res, mpfun::MPNW);
  res.sign=jb.sign;
  return res;
}

mp_real power(const mp& ja, const mp& jb)
{
  mp_real res, mpt1, mpt2;
  mp::mplogx(ja, mp::mppic, mp::mpl02, mpt1, mpfun::MPNW);
  mp::mpmulx(mpt1, jb, mpt2, mpfun::MPNW);
  mp::mpexpx(mpt2, mp::mppic, mp::mpl02, mpt1, mpfun::MPNW);
  return res;
}

mp_real power(const int ia, const mp_real& qb)
{
  mp_real res, mpt2, mpt3;
  mp mpt1(mpsize(6));
  
  mp::mpdmc(mp_dpe(double(ia)), mpt1);
  mp::mplogx(mpt1, mp::mppic, mp::mpl02, mpt2, mpfun::MPNW);
  mp::mpmulx(mpt2, qb, mpt3, mpfun::MPNW);
  mp::mpexpx(mpt3, mp::mppic, mp::mpl02, res, mpfun::MPNW);
  return res;
}

mp_real power(const mp_real& qa, const int ib)
{
  mp_real res;
  mp::mpnpwx(qa, ib, res, mpfun::MPNW);
  return res;
}

mp_real power(const double da, const mp& qb)
{
  mp_real res, mpt2, mpt3;
  mp mpt1(mpsize(6));
  
  mp::mpdmc(mp_dpe(da), mpt1);
  mp::mplogx(mpt1, mp::mppic, mp::mpl02, mpt2, mpfun::MPNW);
  mp::mpmulx(mpt2, qb, mpt3, mpfun::MPNW);
  mp::mpexpx(mpt3, mp::mppic, mp::mpl02, res, mpfun::MPNW);
  return res;
}

mp_real power(const mp& qa, const double db)
{
  mp_real res, mpt1, mpt2;
	
  mp::mplogx(qa, mp::mppic, mp::mpl02, mpt1, mpfun::MPNW);
  mp::mpmuld(mpt1, mp_dpe(db), mpt2, mpfun::MPNW);
  mp::mpexpx(mpt2, mp::mppic, mp::mpl02, res, mpfun::MPNW);
  return res;
}

mp_real acos(const mp_real& qa)
{
  mp_real res, mpt1, mpt2, mpt3;
	
  mp::mpdmc(mp_dpe(1), mpt1);
  mp::mpmulx(qa, qa, mpt2, mpfun::MPNW);
  mp::mpsub(mpt1, mpt2, mpt3, mpfun::MPNW);
  mp::mpsqrx(mpt3, mpt1, mpfun::MPNW);
  mp_complex::mpangx(qa, mpt1, mp::mppic, res, mpfun::MPNW);
  return res;
}

mp_real aint(const mp_real& qa)
{
  mp_real res, mpt1;
  mp::mpinfr(qa, res, mpt1, mpfun::MPNW);
  return res;
}

mp_real anint(const mp_real& qa)
{
  mp_real res;
  mp::mpnint(qa, res, mpfun::MPNW);
  return res;
}

mp_real asin(const mp_real& qa)
{
  mp_real res, mpt1, mpt2, mpt3;
  
  mp::mpdmc(mp_dpe(1), mpt1);
  mp::mpmulx(qa, qa, mpt2, mpfun::MPNW);
  mp::mpsub(mpt1, mpt2, mpt3, mpfun::MPNW);
  mp::mpsqrx(mpt3, mpt1, mpfun::MPNW);
  mp_complex::mpangx(mpt1, qa, mp::mppic, res, mpfun::MPNW);
  return res;
}

mp_real atan(const mp_real& qa)
{
  mp_real res;
  mp mpt1(mpsize(6));
	
  mp::mpdmc(mp_dpe(1), mpt1);
  mp_complex::mpangx(qa, mpt1, mp::mppic, res, mpfun::MPNW);
  return res;
}

mp_real atan2(const mp_real& qa, const mp_real& qb)
{
  mp_real res;
	
  mp_complex::mpangx(qb, qa, mp::mppic, res, mpfun::MPNW);
  return res;
}

mp_real cos(const mp_real& qa)
{
  mp_real res, mpt1;
	
  mp_complex::mpcssx(qa, mp::mppic, res, mpt1, mpfun::MPNW);
  return res;
}

mp_real cosh(const mp_real& qa)
{
  mp_real res, mpt1;
	
  mp::mpcshx(qa, mp::mppic, mp::mpl02, res, mpt1, mpfun::MPNW);
  return res;
}

mp_real exp(const mp_real& qa)
{
  mp_real res;
	
  mp::mpexpx(qa, mp::mppic, mp::mpl02, res, mpfun::MPNW);
  return res;
}

mp_real log(const mp_real& qa)
{
  mp_real res;
	
  mp::mplogx(qa, mp::mppic, mp::mpl02, res, mpfun::MPNW);
  return res;
}

mp_real log10(const mp_real& qa)
{
  mp_real res, mpt1;
	
  mp::mplogx(qa, mp::mppic, mp::mpl02, mpt1, mpfun::MPNW);
  mp::mpdivx(mpt1, mp::mpl10, res, mpfun::MPNW);
  return res;    
}

void mpcsshf(const mp_real& qa, mp_real& qb, mp_real& qc)
{
  mp::mpcshx(qa, mp::mppic, mp::mpl02, qb, qc, mpfun::MPNW);
}

void mpcssnf(const mp_real& qa, mp_real& qb, mp_real& qc)
{
  mp_complex::mpcssx(qa, mp::mppic, qb, qc, mpfun::MPNW);
}

mp_real mpnrtf(const mp_real& qa, int ib)
{
  mp_real res;
  
  mp::mpnrtx(qa, ib, res, mpfun::MPNW);
  return res;
}

mp_real mp_rand()
{
  mp_real res;
	
  mp::mprand(res, mpfun::MPNW);
  return res;
}

mp_int nint(const mp_real& qa)
{
  mp_int res;
	
  mp::mpnint(qa, res, mpfun::MPNW);
  return res;
}


mp_real sin(const mp_real& qa)
{
  mp_real res, mpt1;
	
  mp_complex::mpcssx(qa, mp::mppic, mpt1, res, mpfun::MPNW);
  return res;
}

mp_real sinh(const mp_real& qa)
{
  mp_real res, mpt1;
	
  mp::mpcshx(qa, mp::mppic, mp::mpl02, mpt1, res, mpfun::MPNW);
  return res;
}

mp_real sqrt(const mp_real& qa)
{
  mp_real res;
	
  mp::mpsqrx(qa, res, mpfun::MPNW);
  return res;
}

mp_real tan(const mp_real& qa)
{
  mp_real res, mpt1, mpt2;
	
  mp_complex::mpcssx(qa, mp::mppic, mpt1, mpt2, mpfun::MPNW);
  mp::mpdivx(mpt1, mpt2, res, mpfun::MPNW);
  return res;
}

mp_real tanh(const mp_real& qa)
{
  mp_real res, mpt1, mpt2;
	
  mp::mpcshx(qa, mp::mppic, mp::mppic, mpt1, mpt2, mpfun::MPNW);
  mp::mpdivx(mpt1, mpt2, res, mpfun::MPNW);
  return res;
}

