#include "mpreal.h"
#include <iostream.h>
#include <time.h>


int idivm(mp_int& n, int *ipt, int kp)
{
  
  //!     Returns the first prime divisor of the multiprecision N.        Subroutine
  //!     PRIMES must have been previously called to initialize the array IPT.
  
  mp_int k;
  int i;
  
  //!     Check 3 separately (must be congruent to 2 mod 3).
  int id = 3;
  k = n/id;
  int l = integer(n - k*id);
  if (l == 0) return id;
  if (l == 1) return -3;
  
  for (i = 1; i < kp; i++) 
    {
      id = ipt[i-1];
      k = n/id;
      l = integer(n-k*id);
      if (l == 0) return id;
    }
  return 0;
}

void ranprm(mp_int& ip, int *ipt, int kp, int nd4)
{
  
  //!     Finds a large prime number.  Subroutine PRIMES must have
  //been previously 
  //!     called to initialize the array IP and determine KP.
  
  mp_int id, iq, ix, iy, iz, k, n, n1;
  const int it = 10;
  int ii, jj, ik;
  
  int nt = 20*nd4;
  id = power(mp_int(10), nd4)/2;
  
  for (int j = 1; j <= nt; j++) 
    { 
      //!     Generate a trial prime N that is odd.
      //      n1 = 2*mp_int(id*mp_rand());
      n1 = id*mp_rand();
      n1 = 2*n1;
      n = n1+1;
         
      //!     Perform quick-and-dirty check for primality.
      int kd = idivm(n, ipt, kp);
      if (kd != 0) continue;
      
      //!     Quick test failed to find any divisors.  Proceed with
      //advanced test.
      iq = n1;
      
      for (ik = 0; ik <= 30; ik++) 
	{
	  k = iq/2;
	  if (iq != 2*k) break;
	  iq = k;
	}
      //!     Now N = 1 + Q * 2^IK, where Q is odd.
      
      bool doublebreak=false;
      for (ii = 1; ii <= it; ii++) 
	{
	  ix = n*mp_rand();
	  iy = 1;
	  iz = ix;	  
	  while(true)
	    {
	      k = iq/2;
	      if (iq != 2*k) iy = (iz*iy)%n;
	      iq = k;
	      if(iq ==0 )
		break;
	      iz = (iz*iz)%n;
	    }
	  
	  
	  //!     Now Y = X^Q (mod N).
	  jj = 0;
	  if (iy == 1) continue;
	  
	  //!     Check primality conditions.
	  while(true)
	    {
	      jj++;
	      if (jj > ik) {doublebreak=true;break;}
	      if (iy == n1) break;
	      if (iy == 1) {doublebreak=true;break;}
	      iy = (iy*iy)%n;
	    }
	  if(doublebreak) break;
	}
      
      //!     N is probably prime.
  
      if(doublebreak)
	continue;
      
      ip = n; return;
	

      //!     N is definitely not prime.

    }
  
  cout << "Failed to find a prime.  No. trials = " << nt << '\n'; // here
  exit(1);
 
}


int idiv(int n)
{
  //!     Returns the first prime divisor of N (single precision).
  
  static const int id[8] = {1, 7, 11, 13, 17, 19, 23, 29};
  int kd, i, k, l, m;
  
  if (n%3 == 0) 
     return 3;
   
  if (n%5 == 0) 
    return 5;
 
  kd = 0;
  if (n == 29) return kd;
  k = 2;
  l = 0;
  m = int(sqrt(double(n)));
 
while (true)
  {
    for (i = k-1; i < 8; i++) 
      {
	kd = id[i] + l;
	if (n%kd == 0) return kd;
      }
    
    kd = 0;
    l += 30;
    if (l > m) return kd;
    k = 1;
  }
}

void primes(int *ipt, int kp)
{
  
  //!     Places the first KP prime numbers in IPT.
  static const int ipi[9] = {3, 5, 7, 11, 13, 17, 19, 23, 29};
  
  int i;
  for (i = 0; i < 9; i++)
    ipt[i] = ipi[i];
  
  int k = 9;
  
  for (i = 31; i <= 1000000; i += 2) 
    {
      if (idiv(i) == 0) 
	{
	  ipt[k] = i;
	  k++;
	  if (k == kp) return;
	}
    }
  
  //cout << "Table IPT not filled: " << k<< '\n';
  exit(-1);
}


void expm(const mp_int& ia, const mp_int& ib, const mp_int& ic,
	  mp_int& ip)  
{
  //!     Computes        IP = IA ** IB (mod IC)  for integers IA, IB, IC.
  
  mp_int k, l, l2;
  
  k = ia;
  l = ib;
  ip = 1;
  int j = 0;
  
  
  //!     Perform binary algorithm for exponentiation modulo IC.
 do
 {
  j++;
  l2 = l/2;
  if (l != 2*l2) ip = (k*ip)%ic;
  k = (k*k)%ic;
  l = l2;
 }
 while (l != 0);
 
}


int main()
{
  mp_int  ia, ib, id, ie, ip, iq, ix, iy, iz;
  time_t  tm0, tm1, tm2, tt0, tt1;
  const int n1 = 10, n2 = 10, kp = 200;
  int     ipt[kp];
  
  //!  Initialize.
  time(&tt0);
  int nd = mp::mpipl-15;
  int nd4 = nd/4;
  primes(ipt, kp);
  
  //!     Begin outer loop.
  for (int k = 1; k <= n2; k++) 
    {
      time(&tm0);
       cout << "Outer trial " << k << '\n';
      
		//!     Step 1:  obtain two large primes, each congruent to 2
      //mod 3, with a size
      //!     somewhat less than half the machine precision level.
      ranprm(ip, ipt, kp, nd4);
    
      do
      {
	ranprm(iq, ipt, kp, nd4);
      }
      while (ip == iq);
      /*if (idb >= 1) {
	cout << "p = " << ip;
	cout << "q = " << iq;
	}*/
      
      //!    Step 2:  Compute        A = P * Q  and  B = (P-1) * (Q-1).
      ia = ip*iq;
      ib = (ip-1)*(iq-1);
      
      //!    Step 3:  Set E = 3 and compute D = E inv (mod B)        = B -
			 //[B/3]
      ie = 3;
      id = ib - ib/3;
      /*if (idb >= 1) {
	cout << "a = " << ia;
	cout << "b = " << ib;
	cout << "d = " << id;
	cout << "e = " << ie;
	}*/
      time(&tm1);
      
      //!    Step 4: Encrypt and decrypt a series of numbers chosen
      //at random.
      for (int j = 1; j <= n1; j++) 
	{
	// if (idb >= 1) cout << "Inner trial " << j << '\n';
	  ix = ia*mp_rand();
	  expm(ix, ie, ia, iy);
	  expm(iy, id, ia, iz);
	  if (ix != iz) 
	    {
	      cout << "*** Test failed. " << j << '\n';
	      cout << "x = " << ix;
	      cout << "y = " << iy;
	      cout << "z = " << iz;
	    exit(1);
	  }
	}
      time(&tm2);
      //if (idb >= 1) cout << "CPU times: " << difftime(tm1, tm0) << difftime(tm2, tm1) << '\n';
	}
  
  time(&tt1);
  cout << "Test Passed.\n";
	//              cout << Total CPU time = " << difftime(tt1, tt0) << '\n';
  
  return 0;
}





